! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: ebkgmap_attcommon
!
! Authors:  Masaaki Sakano - SSC/LUX
!
! Description: Shared routines for ebkgmap package, related to Attitude and sky coordinate stuffs.
! 
! Comments: This file is aimed to be imported to ssclib or attcalc2 at some point.
! 
! ToDo lists: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module ebkgmap_attcommon

  use dal
  use cal
  use errorHandling

  use binned_att	! in attcalc2

  ! Followings in ssclib
  use ssc_misc
  use cal_aux
  use dal_aux

  implicit none 

  private 

  public getBinnedScAtt

  interface getBinnedScAtt
    module procedure getBinnedScAttCommon, getBinnedScAttWithImage
  end interface

contains 

  !! see ./eootemap_mod.f90
  !---------------------------------------------------------------------
  subroutine getBinnedScAttCommon(attStyleKeyWord&
     , binnedScAtt & !, numAttBins&
     , medianScAttitude, binnedAttFileName)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! get Attitude data.

    character(*), intent(in) :: attStyleKeyWord
    type(attBinType), pointer :: binnedScAtt(:)	! type def in attcalc2/binned_att.
    ! integer, intent(out) :: numAttBins
    type(SpacecraftAttitudeType), intent(in), optional :: medianScAttitude
       ! Mandatory if attStyleKeyWord=='template'
    character(*), intent(in), optional :: binnedAttFileName
       ! Mandatory if attStyleKeyWord=='binnedset'

    character(256) :: binnedAttSetName = ''
    character(20) :: binnedAttTabName = ''
    logical :: noWeightColumn, noColonFound

    !-------------------------------------------------------------------
    ! Construct the list of attitudes:

    select case(attStyleKeyWord)
    case('binnedset')
      call splitSetTabName(binnedAttFileName, binnedAttSetName, binnedAttTabName &
         , noColonFound)	! in ssclib/dal_aux.
      if (noColonFound) then
        write(errStr, '(a, a, a, a, a)') "No colon+block name was &
          &found after the set name (", trim(binnedAttFileName) &
          , "). I'm going to assume that you meant &
          &to access the first block in the set, which has EXTNAME of (" &
          , trim(binnedAttTabName), "). If that's what you wanted, &
          &then there is no cause for this message."
         call message(appMsg, Noisy)
      end if
      if (trim(binnedAttTabName) == '') then
        write(errStr, '(a, a, a)') "No extention is found in File (" &
           , trim(binnedAttFileName), ")."
        call error('noExtentionTabNameFound')
      end if
      call readBinnedAttitude(binnedAttSetName, binnedAttTabName, binnedScAtt, noWeightColumn)
      ! in attcalc2/binned_att.
      ! numAttBins = size(binnedScAtt)

    case('template')
      ! numAttBins = 1
      allocate(binnedScAtt(1))
      call initializeBinnedAtt(binnedScAtt)	! in attcalc2/binned_att.
      binnedScAtt(1)%isGood = .true.
      binnedScAtt(1)%gtiDuration = 1.0d0 ! dummy value.
      binnedScAtt(1)%att = medianScAttitude

    case default
      call warning ('badAttitudeStyle','attStyleKeyWord=('//trim(attStyleKeyWord)//')')
call error('badAttitudeStyle',"")
    end select

  end subroutine getBinnedScAttCommon


  !---------------------------------------------------------------------
  subroutine getBinnedScAttWithImage(attStyleKeyWord&
     , binnedScAtt & !, numAttBins&
     , inputSetNameToGetCal, binnedAttFileName)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! get Attitude data, when the template FITS (eg. image) is given (but no cal info is given).

    character(*), intent(in) :: attStyleKeyWord, inputSetNameToGetCal
    type(attBinType), pointer :: binnedScAtt(:)	! type def in attcalc2/binned_att.
    ! integer, intent(out) :: numAttBins
    character(*), intent(in), optional :: binnedAttFileName
       ! Mandatory if attStyleKeyWord=='binnedset'

    type(DataSetT) :: set
    real(double) :: dummyExpStartTimeStamp
    type(SpacecraftAttitudeType) :: medianScAttitude
    integer(int32) :: dummyInstrumentId
    type(SpacecraftAttitudeType) :: dummyInstrumAttitude

    !-------------------------------------------------------------------
    ! Construct the list of attitudes:

    select case(attStyleKeyWord)
    case('template')
      set = dataSet(inputSetNameToGetCal, READ)
      call getCalInfo(set, dummyInstrumentId&
         , dummyExpStartTimeStamp, medianScAttitude, dummyInstrumAttitude)
         ! in ssclib/cal_aux.
      call release(set)

      call getBinnedScAttCommon(attStyleKeyWord&
         , binnedScAtt & !, numAttBins&
         , medianScAttitude, binnedAttFileName)
    case default	! ie., case('binnedset')
      call getBinnedScAttCommon(attStyleKeyWord&
         , binnedScAtt & !, numAttBins&
         , binnedAttFileName=binnedAttFileName)
    end select

  end subroutine getBinnedScAttWithImage

end module ebkgmap_attcommon
