! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: eootepileupmask_aux
!
! Authors:  Masaaki Sakano - SSC/LUX
!
! Description: Routines for eootepileupmask (they are compiled here mainly for the convenience for testing).
! 
! Comments: 
! 
! ToDo lists: 
!   * 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eootepileupmask_aux

  ! use dal
  !use types
  use errorHandling
  use cal

  use binned_att	! for initializeBinnedAtt(); defined in attcalc2
  use cube_aux		! for type(keyInfoType) etc; defined in eimchip2sky_aux
  use eimchip2sky_aux	! for rebinChipImagesToSky; defined in eimchip2sky

  ! Followings in ssclib
  ! use array_utils
  use coordinate
  ! use ssc_misc
  ! use cal_aux
  ! use dal_aux
  use epic_aux		! for getCcdNodeNum()
  use wcs_aux		! for type(WcsT)

  use epileupmask_shared
  ! use ebkgmap_skyframe 
  ! use ebkgmap_attcommon

  implicit none 

  private 

  public allocateEooteDiagnOutputT
  public deallocateEooteDiagnOutputT
  public initialiseEooteDiagnOutputT
  public addEooteDiagnOutputT
  public getOoteMaskProjected
  public makeOoteRgaSpikeMask
  public calcRgaSpikeMaskProjected
  public readEvent2RawForEootepileup
  public readEvent2SkyForEootepileup

  integer(int8), parameter :: &
       ChipGroupForRgaMos1(3,2)&	! Works only for MOS1
       = reshape( (/ 2,3,4, 7,6,5 /), (/ 3,2 /) )&
         ! ==> (1,:) = | 2 7 |
         !     (2,:) = | 3 6 |
         !     (3,:) = | 4 5 |
     , ChipGroupForRgaMos2(2,3)&	! Works only for MOS2
       = reshape( (/ 2,7, 3,6, 4,5 /), (/ 2,3 /) )&
         ! ==> (1,:) = | 2 3 4 |
         !     (2,:) = | 7 6 5 |
     , ChipGroupForRgaMos1Chip1(2) = (/ 3, 6 /) &	! only for MOS1
     , ChipGroupForRgaMos2Chip1(4) = (/ 2, 4, 5, 7 /)	! only for MOS2
   ! See Figure 1 or Sec. 5.2.3.5 "Spatial coordinate systems" in cal document.
   ! See the comments in calcRgaSpikeMaskProjected() for how it is used.

  type, public :: eooteDiagnOutputT
    logical :: withRateImage  = .false. &
             , withPileupMask = .false.	! make sure to make them .false. once the pointer arrays are deallocated.
    real(single), pointer :: rateImage(:,:) => null() &
                           , maskPileupImage(:,:) => null()
  end type eooteDiagnOutputT

  interface allocateEooteDiagnOutputT
    module procedure allocateEooteDiagnOutputTSgl, allocateEooteDiagnOutputTAry
  end interface

  interface deallocateEooteDiagnOutputT
    module procedure deallocateEooteDiagnOutputTSgl, deallocateEooteDiagnOutputTAry
  end interface

  interface makeOoteRgaSpikeMask
    module procedure makeOoteRgaSpikeMaskManA, makeOoteRgaSpikeMaskInfC
  end interface

  interface readEvent2RawForEootepileup
    module procedure readEvent2RawForEootepileupManA, readEvent2RawForEootepileupInfC
  end interface

  interface readEvent2SkyForEootepileup
    module procedure readEvent2SkyForEootepileupManA, readEvent2SkyForEootepileupInfC
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine allocateEooteDiagnOutputTSgl(eooteDiagn, withRate, withPileup, xiSize, yiSize)
  !---------------------------------------------------------------------
  ! Allocate and initialise type(eooteDiagnOutputT).

    implicit none 

    type(eooteDiagnOutputT), intent(inout) :: eooteDiagn
    logical, intent(in) :: withRate, withPileup
    integer, intent(in), optional :: xiSize, yiSize

    !-------------------------------------------------------------------
    ! Start:

    eooteDiagn%withRateImage = withRate
    if (withRate) then
      if (present(xiSize) .and. present(yiSize)) then
        allocate(eooteDiagn%rateImage(xiSize,yiSize))
      else
        call error('BUGallocateEooteDiagnOutputT','Contact the code developer (withRate).')
      end if
    end if

    eooteDiagn%withPileupMask = withPileup
    if (withPileup) then
      if (present(xiSize) .and. present(yiSize)) then
        allocate(eooteDiagn%maskPileupImage(xiSize,yiSize))
      else
        call error('BUGallocateEooteDiagnOutputT','Contact the code developer (withPileup).')
      end if
    end if

    call initialiseEooteDiagnOutputT(eooteDiagn)

  end subroutine allocateEooteDiagnOutputTSgl

  !---------------------------------------------------------------------
  subroutine allocateEooteDiagnOutputTAry(eooteDiagn, withRate, withPileup, xiSize, yiSize)
  !---------------------------------------------------------------------
  ! Allocate and initialise type(eooteDiagnOutputT).

    implicit none 

    type(eooteDiagnOutputT), intent(inout) :: eooteDiagn(:)
    logical, intent(in) :: withRate, withPileup
    integer, intent(in), optional :: xiSize, yiSize

    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    do i=1, size(eooteDiagn)
      eooteDiagn(i)%withRateImage = withRate
      if (withRate) then
        if (present(xiSize) .and. present(yiSize)) then
          allocate(eooteDiagn(i)%rateImage(xiSize,yiSize))
        else
          call error('BUGallocateEooteDiagnOutputT','Contact the code developer (withRate).')
        end if
      end if

      eooteDiagn(i)%withPileupMask = withPileup
      if (withPileup) then
        if (present(xiSize) .and. present(yiSize)) then
          allocate(eooteDiagn(i)%maskPileupImage(xiSize,yiSize))
        else
          call error('BUGallocateEooteDiagnOutputT','Contact the code developer (withPileup).')
        end if
      end if

      call initialiseEooteDiagnOutputT(eooteDiagn(i))
    end do

  end subroutine allocateEooteDiagnOutputTAry

  !---------------------------------------------------------------------
  subroutine initialiseEooteDiagnOutputT(eooteDiagn)
  !---------------------------------------------------------------------
  ! Initialise type(eooteDiagnOutputT).

    implicit none 

    type(eooteDiagnOutputT), intent(inout) :: eooteDiagn

    !-------------------------------------------------------------------
    ! Start:

    if (eooteDiagn%withRateImage)  eooteDiagn%rateImage(:,:) = 0.0
    if (eooteDiagn%withPileupMask) eooteDiagn%maskPileupImage(:,:) = 0.0

  end subroutine initialiseEooteDiagnOutputT

  !---------------------------------------------------------------------
  subroutine addEooteDiagnOutputT(eooteDiagnOrig, eooteDiagnToAdd)
  !---------------------------------------------------------------------
  ! Perform:  eooteDiagnOrig += eooteDiagnToAdd

    implicit none 

    type(eooteDiagnOutputT), intent(inout) :: eooteDiagnOrig
    type(eooteDiagnOutputT), intent(in)    :: eooteDiagnToAdd

    !-------------------------------------------------------------------
    ! Start:

    if (eooteDiagnOrig%withRateImage .and. eooteDiagnToAdd%withRateImage) &
       eooteDiagnOrig%rateImage = &
       eooteDiagnOrig%rateImage + &
       eooteDiagnToAdd%rateImage
    if (eooteDiagnOrig%withPileupMask .and. eooteDiagnToAdd%withPileupMask) &
       eooteDiagnOrig%maskPileupImage = &
       eooteDiagnOrig%maskPileupImage + &
       eooteDiagnToAdd%maskPileupImage

  end subroutine addEooteDiagnOutputT

  !---------------------------------------------------------------------
  subroutine deallocateEooteDiagnOutputTSgl(eooteDiagn)
  !---------------------------------------------------------------------
  ! Deallocate type(eooteDiagnOutputT).

    implicit none 

    type(eooteDiagnOutputT), intent(inout) :: eooteDiagn

    !-------------------------------------------------------------------
    ! Start:

    if (eooteDiagn%withRateImage) then
      eooteDiagn%withRateImage = .false.
      deallocate(eooteDiagn%rateImage)
    end if

    if (eooteDiagn%withPileupMask) then
      eooteDiagn%withPileupMask = .false.
      deallocate(eooteDiagn%maskPileupImage)
    end if

  end subroutine deallocateEooteDiagnOutputTSgl

  !---------------------------------------------------------------------
  subroutine deallocateEooteDiagnOutputTAry(eooteDiagn)
  !---------------------------------------------------------------------
  ! Deallocate type(eooteDiagnOutputT).

    implicit none 

    type(eooteDiagnOutputT), intent(inout) :: eooteDiagn(:)
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    do i=1, size(eooteDiagn)
      if (eooteDiagn(i)%withRateImage) then
        eooteDiagn(i)%withRateImage = .false.
        deallocate(eooteDiagn(i)%rateImage)
      end if

      if (eooteDiagn(i)%withPileupMask) then
        eooteDiagn(i)%withPileupMask = .false.
        deallocate(eooteDiagn(i)%maskPileupImage)
      end if
    end do

  end subroutine deallocateEooteDiagnOutputTAry


  !---------------------------------------------------------------------
  function getOoteMaskProjected(maskPileupCube&
     , chipXSize &	!, numCubePlanes &
     , flagsIfRunCubePlane, isForOote) result(maskCubeXOnly)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! return the projected mask (compressed for RAWY direction) for OOTE.
    !-- Note --
    ! The returned pointer variable value must be deallocated after use.

    logical, pointer :: maskPileupCube(:,:,:)
    integer(int16), intent(in) :: chipXSize	! , chipYSize
    logical, intent(in), optional :: flagsIfRunCubePlane(size(maskPileupCube,1)) &	! intent(in).  Not processes for chip/node, for which this is FALSE.
       , isForOote	! Global variable ForOote can be passed here.

    integer(int16) :: xi		! xi is a local var.
    integer(int32) :: numCubePlanes	! == size(maskPileupCube,1)
    integer(int32) :: curCubePlane	! curCubePlane is a local var.
    logical, pointer :: maskCubeXOnly(:,:)
    logical :: localFlagsIfRunCubePlane(size(maskPileupCube,1)), localIsForOote

    !-------------------------------------------------------------------
    ! Start:

    numCubePlanes = size(maskPileupCube,1)

    localFlagsIfRunCubePlane(:) = .true.
    if (present(flagsIfRunCubePlane)) localFlagsIfRunCubePlane = flagsIfRunCubePlane
    localIsForOote = .true.
    if (present(isForOote)) localIsForOote = isForOote

    allocate(maskCubeXOnly(numCubePlanes, chipXSize))	! No Y-axis
    maskCubeXOnly = .false.

    !.................................................................
    ! Make 1-dimensional (X-axis only) mask array for OOT.
    
    if (LocalIsForOote) then
      do curCubePlane = 1, numCubePlanes
        if (.not. localFlagsIfRunCubePlane(curCubePlane)) then
          write(errStr, '(a,i2,a)') "Skip examining the OOT for the case on Chip/Node=", curCubePlane, "."
          call message(appMsg, Noisy)
          cycle
        end if

        do xi = 1, chipXSize
          ! All the rows (ie., same RAWY) which contain one or more piled-up
          ! event have to be masked for Out-Of-Time (OOT) events.
          ! Here maskCubeXOnly is one-dimensional array (for each chip),
          ! which will be SPREAD over Y-axis below.
          if (any(maskPileupCube(curCubePlane,xi,:)))&
            !maskCubeRealProjectedY(curCubePlane,xi) = 1.0
            maskCubeXOnly(curCubePlane,xi) = .true.
        end do	! loop over xi (RAWX)
      end do
    end if

  end function getOoteMaskProjected


  !---------------------------------------------------------------------
  subroutine makeOoteRgaSpikeMaskManA(maskPileupCube&
     , chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices&
     , maskCube, flagsIfRunCubePlane, isForOote, isForRgaSpike)
  !---------------------------------------------------------------------
  ! Make sure to deallocate maskCube() after use.

    implicit none 

    !-- Description --
    ! Reads a mask file that masks the piled-up pixels,
    ! and then creates and returns a mask that masks both or either of 
    ! out-of-time events (OOTE) and/or RGA-spikes, as well as
    ! the piled-up pixels themselves.
    !-- Note --
    !

    logical,      pointer :: maskPileupCube(:,:,:)	! intent(in)
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    integer,      pointer :: cubePlaneIndices(:)	! intent(in)
    logical,      pointer :: maskCube(:,:,:)	! intent(out), should not be allocated before the call.
    logical, intent(in), optional :: flagsIfRunCubePlane(size(maskPileupCube,1)) &	! Not processes for chip/node, for which this is FALSE.
       , isForOote, isForRgaSpike	! Global variable ForOote can be passed here.  In default, both of them are TRUE.

    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared

    !-------------------------------------------------------------------
    ! Start:

    infoCcdChips%chipXSize     = chipXSize        
    infoCcdChips%chipYSize     = chipYSize        
    infoCcdChips%instrumentId  = instrumentId
    infoCcdChips%numCubePlanes = numCubePlanes
    call allocateInfoCcdChipsT(infoCcdChips, cubePlaneIndicesPtr=cubePlaneIndices)

     call makeOoteRgaSpikeMaskInfC(maskPileupCube&
        , infoCcdChips &
        , maskCube, flagsIfRunCubePlane, isForOote, isForRgaSpike)

  end subroutine makeOoteRgaSpikeMaskManA


  !---------------------------------------------------------------------
  subroutine makeOoteRgaSpikeMaskInfC(maskPileupCube&
     , infoCcdChips &
     ! , chipXSize, chipYSize&
     ! , instrumentId, numCubePlanes, cubePlaneIndices&
     , maskCube, flagsIfRunCubePlane, isForOote, isForRgaSpike)
  !---------------------------------------------------------------------
  ! Make sure to deallocate maskCube() after use.

    implicit none 

    !-- Description --
    ! Reads a mask file that masks the piled-up pixels,
    ! and then creates and returns a mask that masks both or either of 
    ! out-of-time events (OOTE) and/or RGA-spikes, as well as
    ! the piled-up pixels themselves.
    !-- Note --
    !

    logical,      pointer :: maskPileupCube(:,:,:)	! intent(in)
    type(infoCcdChipsT), intent(in) :: infoCcdChips	! Defined in epileupmask_shared
    integer(int16) :: chipXSize, chipYSize
    ! integer(int32) :: instrumentId, numCubePlanes
    ! integer,      pointer :: cubePlaneIndices(:)	! 
    logical,      pointer :: maskCube(:,:,:)	! intent(out), should not be allocated before the call.
    logical, intent(in), optional :: flagsIfRunCubePlane(size(maskPileupCube,1)) &	! Not processes for chip/node, for which this is FALSE.
       , isForOote, isForRgaSpike	! Global variable ForOote can be passed here.  In default, both of them are TRUE.

    logical :: localIsForOote, localIsForRgaSpike
    logical, pointer :: maskCubeXOnly(:,:), maskCubeYOnly(:,:)&
       , maskCubeProjectedS(:,:)

    !-------------------------------------------------------------------
    ! Start:

     chipXSize        = infoCcdChips%chipXSize
     chipYSize        = infoCcdChips%chipYSize 
     ! instrumentId     = infoCcdChips%instrumentId
     ! numCubePlanes    = infoCcdChips%numCubePlanes
     ! cubePlaneIndices => infoCcdChips%cubePlaneIndices

    !.................................................................
    ! Make 1-dimensional (X-axis only) mask array for OOT.
    
    localIsForOote = .true.
    if (present(isForOote)) localIsForOote = isForOote
    localIsForRgaSpike = .true.
    if (present(isForRgaSpike)) localIsForRgaSpike = isForRgaSpike

    if (localIsForOote) then
       maskCubeXOnly => getOoteMaskProjected(maskPileupCube, chipXSize &
          , flagsIfRunCubePlane=flagsIfRunCubePlane, isForOote=localIsForOote)	! defined in eootepileupmask_aux
          !, numCubePlanes, flagsIfRunCubePlane=flagsIfRunCubePlane)
    end if

    !.................................................................
    ! Make 1-dimensional mask arrays for RGA spikes
    
    ! --- NOTE ---
    ! RGA spikes appear along the X_CamCoord1 (or 2) direction for MOS1 and 2.
    ! This corresponds to RAWX axes for chip1 but to RAWY axes for other chips.
    ! Therefore, chip1 and other chips are treated separately here.

    if (localIsForRgaSpike) then
      if (infoCcdChips%instrumentId == EPN) then
        localIsForRgaSpike = .false.	! This is important because otherwise this code would try to deallocate pointers maskCubeYOnly() and maskCubeProjectedS() at the end of this subroutine, which would cause SegFault.
        ! allocate(maskCubeYOnly(     numCubePlanes, chipYSize)&	! No X-axis
        !        , maskCubeProjectedS(numCubePlanes, chipXSize))	! No Y-axis
        ! maskCubeYOnly = .false.
        ! maskCubeProjectedS = .false.
        call warning('noRgaSpikeForEpn' &
           , 'forrgaspike=yes is specified, however the current instrument pn&
             & does not show the RGA Spike feature, hence ignored.')
      else
        call calcRgaSpikeMaskProjected(maskPileupCube&
           ! , chipXSize, chipYSize&
           , infoCcdChips%instrumentId & !, numCubePlanes
           , infoCcdChips%cubePlaneIndices&
           , maskCubeProjectedS, maskCubeYOnly, flagsIfRunCubePlane)	! defined in eootepileupmask_aux
      end if
    end if		! if (localIsForRgaSpike) then

    !.................................................................
    ! Make the final mask.
    
    allocate(maskCube(infoCcdChips%numCubePlanes, chipXSize, chipYSize))

    !! Initialisation (true at piled-up area)
    !!  => No.  maskPileupCube includes all the CCD-chips regardless
    !!    its exposure.  So, even if a chip has too small exposure
    !!    and hence is ignored to be processed for OOTE/RGA,
    !!    maskPileupCube does not know it, and hence if we make 
    !!    the piled-up area TRUE, those pixels remain TRUE
    !!    regardless its exposure.  That is not desirable.
    !!    Anyhow, when OOTE and/or RGA-spikes are considered,
    !!    those piled-up pixels will be automatically masked.
    !!    Hence it does not have to (and should not) be initialised
    !!    at this stage.
    !!
    ! where (maskPileupCube)
    !   maskCube = .true.
    ! elsewhere
      maskCube = .false.
    ! end where

    if (localIsForOote) then
      ! if (localIsForRgaSpike) then	! Both ForOote and RgaSpike
      !   where ( spread(maskCubeXOnly, 3, chipYSize)&
      !      .or. spread(maskCubeYOnly, 2, chipXSize)&
      !      .or. spread(maskCubeProjectedS, 3, chipYSize))
      !     maskCube = .true.
      !   end where
      ! else			! ForOote only (no RgaSpike)
        where ( spread(maskCubeXOnly, 3, chipYSize))
          maskCube = .true.
        end where
      ! end if	! if (localIsForRgaSpike) then
    end if	! if (localIsForOote) then

    ! else	! if (localIsForOote) then
      if (localIsForRgaSpike) then	! No ForOote, but RgaSpike
        where ( spread(maskCubeYOnly, 2, chipXSize)&
           .or. spread(maskCubeProjectedS, 3, chipYSize))
          maskCube = .true.
        end where
      ! else			! Neither ForOote nor RgaSpike
      !   ! Do nothing
      end if
    ! end if	! if (localIsForOote) then

    if (localIsForOote) then
      deallocate(maskCubeXOnly)
    end if
    if (localIsForRgaSpike) then
      deallocate(maskCubeYOnly, maskCubeProjectedS)
    end if

  end subroutine makeOoteRgaSpikeMaskInfC


  !---------------------------------------------------------------------
  subroutine makeOoteRgaSpikeMask_old(maskPileupCube&
     , chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices&
     , maskCubeReal, flagsIfRunCubePlane, isForOote, isForRgaSpike)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! read mask and further mask for either DETX (RGA-spikes) or RAWY (OOTE) direction if any of the column is above the threshold.
    !-- Note --
    ! Use global variables (command-line args) ForOote and ForRgaSpike .

    logical,      pointer :: maskPileupCube(:,:,:)
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    integer,      pointer :: cubePlaneIndices(:)
    real(single), pointer :: maskCubeReal(:,:,:)
    logical, intent(in), optional :: flagsIfRunCubePlane(size(maskPileupCube,1)) &	! Not processes for chip/node, for which this is FALSE.
       , isForOote, isForRgaSpike	! Global variable ForOote can be passed here.

    logical :: localIsForOote, localIsForRgaSpike
    logical, pointer :: maskCubeXOnly(:,:), maskCubeYOnly(:,:)&
       , maskCubeProjectedS(:,:)

    !-------------------------------------------------------------------
    ! Start:

    !.................................................................
    ! Make 1-dimensional (X-axis only) mask array for OOT.
    
    localIsForOote = .true.
    if (present(isForOote)) localIsForOote = isForOote
    localIsForRgaSpike = .true.
    if (present(isForRgaSpike)) localIsForRgaSpike = isForRgaSpike

    if (localIsForOote) then
       maskCubeXOnly => getOoteMaskProjected(maskPileupCube, chipXSize &
          , flagsIfRunCubePlane=flagsIfRunCubePlane, isForOote=localIsForOote)	! defined in eootepileupmask_aux
          !, numCubePlanes, flagsIfRunCubePlane=flagsIfRunCubePlane)
    end if

    !.................................................................
    ! Make 1-dimensional mask arrays for RGA spikes
    
    ! --- NOTE ---
    ! RGA spikes appear along the X_CamCoord1 (or 2) direction for MOS1 and 2.
    ! This corresponds to RAWX axes for chip1 but to RAWY axes for other chips.
    ! Therefore, chip1 and other chips are treated separately here.

    if (localIsForRgaSpike) then
      if (instrumentId == EPN) then
        localIsForRgaSpike = .false.	! This is important because otherwise this code would try to deallocate pointers maskCubeYOnly() and maskCubeProjectedS() at the end of this subroutine, which would cause SegFault.
        ! allocate(maskCubeYOnly(     numCubePlanes, chipYSize)&	! No X-axis
        !        , maskCubeProjectedS(numCubePlanes, chipXSize))	! No Y-axis
        ! maskCubeYOnly = .false.
        ! maskCubeProjectedS = .false.
        call warning('noRgaSpikeForEpn' &
           , 'forrgaspike=yes is specified, however the current instrument pn&
             & does not show the RGA Spike feature, hence ignored.')
      else
        call calcRgaSpikeMaskProjected(maskPileupCube&
           ! , chipXSize, chipYSize&
           , instrumentId & !, numCubePlanes
           , cubePlaneIndices&
           , maskCubeProjectedS, maskCubeYOnly, flagsIfRunCubePlane)	! defined in eootepileupmask_aux
      end if
    end if		! if (localIsForRgaSpike) then

    !.................................................................
    ! Make the final mask.
    
    allocate(maskCubeReal(numCubePlanes, chipXSize, chipYSize))

    ! Initialisation (true at piled-up area)
    where (maskPileupCube)
      maskCubeReal = 1.0
    elsewhere
      maskCubeReal = 0.0
    end where

    if (localIsForOote) then
      if (localIsForRgaSpike) then	! Both ForOote and RgaSpike
        where ( spread(maskCubeXOnly, 3, chipYSize)&
           .or. spread(maskCubeYOnly, 2, chipXSize)&
           .or. spread(maskCubeProjectedS, 3, chipYSize))
          maskCubeReal = 1.0
        elsewhere
          maskCubeReal = 0.0
        end where
      else			! ForOote only (no RgaSpike)
        where ( spread(maskCubeXOnly, 3, chipYSize))
          maskCubeReal = 1.0
        elsewhere
          maskCubeReal = 0.0
        end where
      end if	! if (localIsForRgaSpike) then

    else	! if (localIsForOote) then
      if (localIsForRgaSpike) then	! No ForOote, but RgaSpike
        where ( spread(maskCubeYOnly, 2, chipXSize)&
           .or. spread(maskCubeProjectedS, 3, chipYSize))
          maskCubeReal = 1.0
        elsewhere
          maskCubeReal = 0.0
        end where
      else			! Neither ForOote nor RgaSpike
        ! Do nothing
      end if
    end if	! if (localIsForOote) then

    if (localIsForOote) then
      deallocate(maskCubeXOnly)
    end if
    if (localIsForRgaSpike) then
      deallocate(maskCubeYOnly, maskCubeProjectedS)
    end if

  end subroutine makeOoteRgaSpikeMask_old


  !---------------------------------------------------------------------
  subroutine calcRgaSpikeMaskProjected(maskPileupCube&
     !, chipXSize, chipYSize&
     , instrumentId & !, numCubePlanes
     , cubePlaneIndices&
     , maskCubeXOnly, maskCubeYOnly, flagsIfRunCubePlane)
  !---------------------------------------------------------------------

  ! TODO: Tests where there is a bright source on peripheral chips are needed!!!!!!!!!!!

    implicit none 

    !-- Description --
    ! return the projected masks, one compressed for RAWX direction for chip 1,
    ! and the other compressed for RAWY direction for othe chips, indicating
    ! the possible affected field by RGA spikes.
    !-- Note --
    ! RGA spikes appear along the Y_SacCoord1 direction, which is
    ! X_CamCoord1 (or 2) and Y_CamCoord1 (or 2) for MOS1 and 2, respectively,
    ! namely Chip 3,1,6 for MOS1 and chip 1 and chips 4/2 or 5/7 for MOS 2.
    ! This corresponds to RAWX axe for chip1 but to RAWY axe for the other chips
    ! for MOS1, and RAWY for chip1 and RAWX for the others for MOS2.
    ! Therefore, chip1 and other chips are treated separately here.
    !
    ! When there is a bright source in other chips, I (Masa) am guessing
    ! the RGA spikes would not appear, however it is not confirmed
    ! by the experts.
    !
    ! NOTE!!
    !  Nodes in MOSs are not considered, namely numNodes==1 is assumed,
    ! and hence cubePlaneIndices(index) is assumed to give simply the chip number.
    !  maskCubeYOnly and maskCubeXOnly have to be deallocated
    !  by the caller before and after use.

    logical, pointer :: maskPileupCube(:,:,:)	! intent(in)
    ! integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId !, numCubePlanes
    integer, pointer :: cubePlaneIndices(:)	! intent(in)
    logical, pointer :: maskCubeXOnly(:,:), maskCubeYOnly(:,:)	! intent(out)
    logical, intent(in), optional :: flagsIfRunCubePlane(size(maskPileupCube,1))	! Not processes for chip/node, for which this is FALSE.

    integer(int16) :: chipXSize, chipYSize
    integer(int32) :: numCubePlanes, iCubePlanes
    integer(int16) :: chipXSizeHalf, chipYSizeHalf	! local var.
    integer(int16) :: xi, yi !, refXRaw, refYRaw	! local var.
    integer(int16) :: numChips, iChipGroup, indexChip, i
    real(double) :: xDet, yDet, xRaw, yRaw !, refXDet, refYDet
    logical(bool) :: isOffChip
    ! integer(int8)  :: ccdNum		! for detToRaw() in ssclib/coordinate
    integer(int32) :: chipFrom, chipTo	! for CAL_setState()
    integer(int32) :: instrumentIdCal, ccdNumCal
    ! integer :: k
    logical        :: chipsWithRga(EMOS_CCD_LAST,EMOS_CCD_LAST) &	! True if Chips (From,To) are used to mask the RGA feature.
       , localFlagsIfRunCubePlane(size(maskPileupCube,1))
    character(64)  :: strChips
logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    chipsWithRga(:,:) = .false.

    localFlagsIfRunCubePlane(:) = .true.
    if (present(flagsIfRunCubePlane)) localFlagsIfRunCubePlane = flagsIfRunCubePlane

    numCubePlanes = size(maskPileupCube, 1)
    chipXSize     = size(maskPileupCube, 2)
    chipYSize     = size(maskPileupCube, 3)
    allocate(maskCubeYOnly(numCubePlanes, chipYSize)&	! No X-axis
           , maskCubeXOnly(numCubePlanes, chipXSize))	! No Y-axis
    maskCubeYOnly = .false.
    maskCubeXOnly = .false.

    chipXSizeHalf = chipXSize/2	! int16
    chipYSizeHalf = chipYSize/2	! int16

    if (.not. (instrumentId == EMOS1 .or. instrumentId == EMOS2)) then
      return
    else
      if (chipXSize .ne. chipYSize) call error('unequalChipSizes','')
    end if

    call CAL_getState(instrument=instrumentIdCal, ccdChipId=ccdNumCal)

    !.................................................................
    ! Make 1-dimensional mask arrays for RGA spikes

    ! if ((instrumentId == EMOS1 .or. instrumentId == EMOS2)) then
    if (instrumentId == EMOS1) then
      !.................................................................
      ! EMOS1

      !--- RGA spikes to/from the surrounding chips (except for Chip 1). ---
       call message(appMsg, Noisy, 'Masking RGA spikes to/from the surrounding chips (except for Chip 1). (MOS1)')
      numChips = size(      ChipGroupForRgaMos1, 2)
      do iChipGroup=1, size(ChipGroupForRgaMos1, 1)	! Loop over a "group" of chips, such as (2, 7) -- see definition of ChipGroupForRgaMos1.

        do indexChip=1, numChips
          chipFrom = ChipGroupForRgaMos1(iChipGroup, indexChip)	! Bright source in this chip
          iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipFrom))	! getCcdNodeNum() defined in ssclib/epic_aux
          write(errStr, '(a,i2,a)') "Skip processing the case from Chip=", chipFrom, "."
          if (iCubePlanes < 1) then	! ==0, meaning the CCD/NODE is not defined/used.  See the comments in subroutine countNumberOfFrames() in epileupmask_shared.f90
            call message(appMsg, Noisy)
            cycle
          else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
            call message(appMsg, Noisy)
            cycle
          end if

          call CAL_setState(instrument=instrumentId, ccdChipId=chipFrom)
          do i=1, numChips
            chipTo = ChipGroupForRgaMos1(iChipGroup, i)	! RGA spike appearing in this chip
            iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipTo))
            write(errStr, '(a,i2,a)') "Skip processing the case to Chip=", chipTo, "."
            if (iCubePlanes < 1) then	! ==0, see above.
              call message(appMsg, Noisy)
              cycle
            else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
              call message(appMsg, Noisy)
              cycle
            end if

            if (chipFrom == chipTo) then	! RGA-spike in the "self"-chip
              ! if (i == indexChip) then	! : the same meaning as above.
              do xi = 1, chipXSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),xi,:))) then
                  maskCubeXOnly(cubePlaneIndices(chipTo), xi) = .true.
                  chipsWithRga(chipFrom,chipTo) = .true.
                end if
              end do	!  loop over xi (RAWX)
            else			! RGA-spike in the surrounding chips
              do xi = 1, chipXSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),xi,:))) then
                  call CAL_setState(ccdChipId=chipFrom)
                  call rawToDet(xi, chipYSizeHalf, xDet, yDet)	! in ssclib/coordinate
                  call CAL_setState(ccdChipId=chipTo)
                  call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip)
                  ! call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)

                  if (.not. isOffChip) then
                    write(errStr, '(a,a,i2,"=>",i2,a,i3,",",i3,a,f10.5,",",f10.5,").")') &
                       "The coordinates fall inside the chip (isOffChip==.false.), "&
                       //"probably because either axis value is at the very edge.: "&
                       , "chipFrom/To=", chipFrom, chipTo &
                       ,  ", (xi,yi)=(", xi, chipYSizeHalf &
                       , "), (xRaw,yRaw)=(", xRaw, yRaw
                    call message(appMsg, Noisy)
                  end if

                  if (xRaw < 0.5) then
                    write(errStr, '(a,": chipFrom=",i2," chipto=",i2, " xRaw=",f10.5)') &
                       "xRaw is negative.", chipFrom, chipTo, xRaw
                    call message(appMsg, Noisy)
                    ! call error('xRawNegativeInPileupmaskDetToRaw')	! Should never happen.  Just a safety net.
                  else if (xRaw >= chipXSize+0.5) then
                    write(errStr, '(a,": chipFrom=",i2,", chipto=",i2, ", xRaw=",f10.5)') &
                       "xRaw too large", chipFrom, chipTo, xRaw
                    call message(appMsg, Noisy)
                    ! call warning('xRawTooBigInPileupmaskDetToRaw')
                  else
                    maskCubeXOnly(cubePlaneIndices(chipTo), nint(xRaw)) = .true.
                    chipsWithRga(chipFrom,chipTo) = .true.
                  end if

                end if
              end do	!  loop over xi (RAWX)
            end if	! if (i == indexChip) then

          end do	!  do i=1, numChips		(for chipTo)
        end do		! do indexChip=1, numChips	(for chipFrom)
      end do		! do iChipGroup=1, size(ChipGroupForRgaMos1, 1)


      !---- RGA spikes to/from Chip 1. ----

      numChips = size(      ChipGroupForRgaMos1Chip1, 1)

      do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)
        ! RGA spikes from Chip 1. -----
        !   nb., this routine is similar to the above, except that
        !    the RGA spike to chip1 (the same chip) is not handled
        !    and that the coordinates of X/Y are different.
          call message(appMsg, Noisy, 'Masking RGA spikes from Chip 1. (MOS1)')
          chipFrom = 1
if (testFlag) then
print *,'DEBUG:index? cubePlaneIndices(chipFrom=',chipFrom,')=',cubePlaneIndices(chipFrom)
print *,'DEBUG:index? cubePlaneIndices(:)=',cubePlaneIndices(:)
end if
          iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipFrom))	! getCcdNodeNum() defined in ssclib/epic_aux
          write(errStr, '(a,i2,a)') "Skip processing the case from Chip=", chipFrom, "."
          if (iCubePlanes < 1) then	! ==0, see above.
            call message(appMsg, Noisy)
            cycle
          else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
            call message(appMsg, Noisy)
            cycle
          end if

          call CAL_setState(instrument=instrumentId, ccdChipId=chipFrom)
          do i=1, numChips
            chipTo = ChipGroupForRgaMos1Chip1(i)	! RGA spike appearing in this chip
            iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipTo))
            write(errStr, '(a,i2,a)') "Skip processing the case to Chip=", chipTo, "."
            if (iCubePlanes < 1) then	! ==0, see above.
              call message(appMsg, Noisy)
              cycle
            else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
              call message(appMsg, Noisy)
              cycle
            end if

              do yi = 1, chipYSize	! in Chip1 (chipFrom)
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),:,yi))) then
                  call CAL_setState(ccdChipId=chipFrom)
                  call rawToDet(chipXSizeHalf, yi, xDet, yDet)	! in ssclib/coordinate
                  ! call rawToDet(1_int16, yi, xDet, yDet)	! in ssclib/coordinate
                  ! call rawToDet(refXRaw, yi, xDet, yDet)	! in ssclib/coordinate
                  call CAL_setState(ccdChipId=chipTo)
                  call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip)
                  ! call detToRaw(refXDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
! if (isOffChip) then
!    !print *,'=== isOffChip is True!'
!    write (*,'("ccdNum=",i1,", (xDet,yDet)=(",f7.1,",",f7.1,"), (xRaw,yRaw)=(",f5.1,",",f5.1,").")')&
!         ccdNum, xDet,yDet,xRaw,yRaw
! else
!    write (*,'("ccdNum=",i1,", (xDet,yDet)=(",f7.1,",",f7.1,"), (xRaw,yRaw)=(",f5.1,",",f5.1,").")')&
!         ccdNum, xDet,yDet,xRaw,yRaw
! end if
                  if (.not. isOffChip) then
                    write(errStr, '(a,a,i2,"=>",i2,a,i3,",",i3,a,f10.5,",",f10.5,").")') &
                       "The coordinates fall inside the chip (isOffChip==.false.), "&
                       //"probably because either axis value is at the very edge.: "&
                       , "chipFrom/To=", chipFrom, chipTo &
                       ,  ", (xi,yi)=(", chipXSizeHalf, yi &
                       , "), (xRaw,yRaw)=(", xRaw, yRaw
                    call message(appMsg, Noisy)
                  end if

                  if (xRaw < 0.5) then
                    write(errStr, '(a,": chipFrom=",i2," chipto=",i2, " xRaw=",f10.5)') &
                       "xRaw is negative", chipFrom, chipTo, xRaw
                    call message(appMsg, Noisy)
                    ! call error('xRawNegativeInPileupmaskDetToRaw')	! Should never happen.  Just a safety net.
                  else if (xRaw >= chipXSize+0.5) then
                    write(errStr, '(a,": chipFrom=",i2,", chipto=",i2, ", xRaw=",f10.5)') &
                       "xRaw too large", chipFrom, chipTo, xRaw
                    call message(appMsg, Noisy)
                    ! call warning('xRawTooBigInPileupmaskDetToRaw')
                  else
                    maskCubeXOnly(cubePlaneIndices(chipTo), nint(xRaw)) = .true.
                    chipsWithRga(chipFrom,chipTo) = .true.
                  end if
                end if
              end do	!  loop over yi (RAWY)
          end do	!  do i=1, numChips		(for chipTo)
      end do	! do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)

      do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)
        ! RGA spikes to Chip 1. -----
        !   nb., this routine is quite different from the above (X <=> Y etc).
        !        Those from chip1 to chip1 are NOT handled here (see below).
        call message(appMsg, Noisy, 'Masking RGA spikes to Chip 1. (MOS1)')
        do indexChip=1, numChips
          chipFrom = ChipGroupForRgaMos1Chip1(indexChip)
          iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipFrom))	! getCcdNodeNum() defined in ssclib/epic_aux
          write(errStr, '(a,i2,a)') "Skip processing the case from Chip=", chipFrom, "."
          if (iCubePlanes < 1) then	! ==0, see above.
            call message(appMsg, Noisy)
            cycle
          else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
            call message(appMsg, Noisy)
            cycle
          end if

          call CAL_setState(instrument=instrumentId, ccdChipId=chipFrom)
            chipTo = 1
            iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipTo))
            write(errStr, '(a,i2,a)') "Skip processing the case to Chip=", chipTo, "."
            if (iCubePlanes < 1) then	! ==0, see above.
              call message(appMsg, Noisy)
              cycle
            else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
              call message(appMsg, Noisy)
              cycle
            end if

              do xi = 1, chipXSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),xi,:))) then
                  call CAL_setState(ccdChipId=chipFrom)
                  call rawToDet(xi, chipYSizeHalf, xDet, yDet)	! in ssclib/coordinate
                  call CAL_setState(ccdChipId=chipTo)
                  call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip)
                  ! call rawToDet(xi, int(1, kind=int16), xDet, yDet)	! in ssclib/coordinate
                  ! call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
                  ! maskCubeYOnly(cubePlaneIndices(chipTo), nint(yRaw)) = .true.
                  if (.not. isOffChip) then
                    write(errStr, '(a,a,i2,"=>",i2,a,i3,",",i3,a,f10.5,",",f10.5,").")') &
                       "The coordinates fall inside the chip (isOffChip==.false.), "&
                       //"probably because either axis value is at the very edge.: "&
                       , "chipFrom/To=", chipFrom, chipTo &
                       ,  ", (xi,yi)=(", xi, chipYSizeHalf &
                       , "), (xRaw,yRaw)=(", xRaw, yRaw
                    call message(appMsg, Noisy)
                  end if

                  if (yRaw < 0.5) then
                    write(errStr, '(a,": chipFrom=",i2," chipto=",i2, " yRaw=",f10.5)') &
                       "yRaw is negative", chipFrom, chipTo, yRaw
                    call message(appMsg, Noisy)
                    ! call error('yRawNegativeInPileupmaskDetToRaw')	! Should never happen.  Just a safety net.
                  else if (yRaw >= chipYSize+0.5) then
                    write(errStr, '(a,": chipFrom=",i2,", chipto=",i2, ", yRaw=",f10.5)') &
                       "yRaw too large", chipFrom, chipTo, yRaw
                    call message(appMsg, Noisy)
                  else
                    maskCubeYOnly(cubePlaneIndices(chipTo), nint(yRaw)) = .true.	! chipTo=1
                    chipsWithRga(chipFrom,chipTo) = .true.
                  end if
                end if
              end do	!  loop over xi (RAWX)
        end do		! do indexChip=1, numChips	(for chipFrom)
      end do	! do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)


      do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)
        ! --- RGA spikes from Chip1 to Chip 1. ---
        call message(appMsg, Noisy, 'Masking RGA spikes form Chip 1 to Chip 1. (MOS1)')
          chipFrom = 1
          iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipFrom))	! getCcdNodeNum() defined in ssclib/epic_aux
          write(errStr, '(a,i2,a)') "Skip processing the case from Chip=", chipFrom, "."
          if (iCubePlanes < 1) then	! ==0, see above.
            call message(appMsg, Noisy)
            cycle
          else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
            call message(appMsg, Noisy)
            cycle
          end if

            chipTo = 1
            iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipTo))
            write(errStr, '(a,i2,a)') "Skip processing the case to Chip=", chipTo, "."
            if (iCubePlanes < 1) then	! ==0, see above.
              call message(appMsg, Noisy)
              cycle
            else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
              call message(appMsg, Noisy)
              cycle
            end if

              do yi = 1, chipYSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),:,yi))) then
                  maskCubeYOnly(cubePlaneIndices(chipTo), yi) = .true.
                  chipsWithRga(chipFrom,chipTo) = .true.
                end if
              end do	!  loop over yi (RAWY)
      end do	! do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)


    else if (instrumentId == EMOS2) then
    !.................................................................
    ! EMOS2

      !--- RGA spikes to/from the surrounding chips (except for Chip 1). ---
       call message(appMsg, Noisy, 'Masking RGA spikes to/from the surrounding chips (except for Chip 1). (MOS2)')
      numChips = size(      ChipGroupForRgaMos2, 2)
      do iChipGroup=1, size(ChipGroupForRgaMos2, 1)	! Loop over a "group" of chips, such as (2, 3, 4) -- see definition of ChipGroupForRgaMos2.

        do indexChip=1, numChips
          chipFrom = ChipGroupForRgaMos2(iChipGroup, indexChip)	! Bright source in this chip
          iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipFrom))	! getCcdNodeNum() defined in ssclib/epic_aux
          write(errStr, '(a,i2,a)') "Skip processing the case from Chip=", chipFrom, "."
          if (iCubePlanes < 1) then	! ==0, meaning the CCD/NODE is not defined/used.  See the comments in subroutine countNumberOfFrames() in epileupmask_shared.f90
            call message(appMsg, Noisy)
            cycle
          else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
            call message(appMsg, Noisy)
            cycle
          end if

          call CAL_setState(instrument=instrumentId, ccdChipId=chipFrom)
          do i=1, numChips
            chipTo = ChipGroupForRgaMos2(iChipGroup, i)	! RGA spike appearing in this chip
            iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipTo))
            write(errStr, '(a,i2,a)') "Skip processing the case to Chip=", chipTo, "."
            if (iCubePlanes < 1) then	! ==0, see above.
              call message(appMsg, Noisy)
              cycle
            else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
              call message(appMsg, Noisy)
              cycle
            end if

            if (chipFrom == chipTo) then	! RGA-spike in the "self"-chip
              ! if (i == indexChip) then	! : the same meaning as above.
              do yi = 1, chipYSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),:,yi))) then
                  maskCubeYOnly(cubePlaneIndices(chipTo), yi) = .true.
                  chipsWithRga(chipFrom,chipTo) = .true.
                end if
              end do	!  loop over yi (RAWY)
            else			! RGA-spike in the surrounding chips
              do yi = 1, chipYSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),:,yi))) then
                  call CAL_setState(ccdChipId=chipFrom)
                  call rawToDet(chipXSizeHalf, yi, xDet, yDet)	! in ssclib/coordinate
                  call CAL_setState(ccdChipId=chipTo)
                  call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip)
                  ! call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)

                  if (.not. isOffChip) then
                    write(errStr, '(a,a,i2,"=>",i2,a,i3,",",i3,a,f10.5,",",f10.5,").")') &
                       "The coordinates fall inside the chip (isOffChip==.false.), "&
                       //"probably because either axis value is at the very edge.: "&
                       , "chipFrom/To=", chipFrom, chipTo &
                       ,  ", (xi,yi)=(", chipXSizeHalf, yi &
                       , "), (xRaw,yRaw)=(", xRaw, yRaw
                    call message(appMsg, Noisy)
                  end if

                  if ((yRaw < 0.5) .or. (yRaw >= chipYSize+0.5)) then	! to avboid nint(yRaw) falls out of range of the index.
                  ! if ((yRaw <= 0) .or. (yRaw >= chipYSize+0.5)) then
                    ! Do nothing...
                    ! NOTE: In MOS2, one in chip-pairs of (2,3,4) and (5,6,7)
                    !  often falls out of range of the RGA spike from
                    !  another chip.
                    !  In that case, it just skips processing.
                  else
                    maskCubeYOnly(cubePlaneIndices(chipTo), nint(yRaw)) = .true.
                    chipsWithRga(chipFrom,chipTo) = .true.
                  end if
                end if
              end do	!  loop over yi (RAWY)
            end if	! if (i == indexChip) then

          end do	!  do i=1, numChips		(for chipTo)
        end do		! do indexChip=1, numChips	(for chipFrom)
      end do		! do iChipGroup=1, size(ChipGroupForRgaMos2, 1)


      !---- RGA spikes to/from Chip 1. ----

      numChips = size(      ChipGroupForRgaMos2Chip1, 1)

      do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)
        ! RGA spikes from Chip 1. -----
        !   nb., this routine is similar to the above, except that
        !    the RGA spike to chip1 (the same chip) is not handled
        !    and that the coordinates of X/Y are different.
          call message(appMsg, Noisy, 'Masking RGA spikes from Chip 1 (MOS2).')
          chipFrom = 1
          iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipFrom))	! getCcdNodeNum() defined in ssclib/epic_aux
          write(errStr, '(a,i2,a)') "Skip processing the case from Chip=", chipFrom, "."
          if (iCubePlanes < 1) then	! ==0, see above.
            call message(appMsg, Noisy)
            cycle
          else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
            call message(appMsg, Noisy)
            cycle
          end if

          call CAL_setState(instrument=instrumentId, ccdChipId=chipFrom)
          ! do i=1, size(ChipGroupForRgaMos2Chip1)
          do i=1, numChips
            chipTo = ChipGroupForRgaMos2Chip1(i)	! RGA spike appearing in this chip
            iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipTo))
            write(errStr, '(a,i2,a)') "Skip processing the case to Chip=", chipTo, "."
            if (iCubePlanes < 1) then	! ==0, see above.
              call message(appMsg, Noisy)
              cycle
            else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
              call message(appMsg, Noisy)
              cycle
            end if

              do xi = 1, chipXSize	! in Chip1 (chipFrom)
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),xi,:))) then
                  call CAL_setState(ccdChipId=chipFrom)
                  call rawToDet(xi, chipYSizeHalf, xDet, yDet)	! in ssclib/coordinate
                  call CAL_setState(ccdChipId=chipTo)
                  call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip)

                  if (.not. isOffChip) then
                    write(errStr, '(a,a,i2,"=>",i2,a,i3,",",i3,a,f10.5,",",f10.5,").")') &
                       "The coordinates fall inside the chip (isOffChip==.false.), "&
                       //"probably because either axis value is at the very edge.: "&
                       , "chipFrom/To=", chipFrom, chipTo &
                       ,  ", (xi,yi)=(", xi, chipYSizeHalf &
                       , "), (xRaw,yRaw)=(", xRaw, yRaw
                    call message(appMsg, Noisy)
                  end if

                  if ((yRaw < 0.5) .or. (yRaw >= chipYSize+0.5)) then
                    ! Do nothing...
                    ! NOTE: In MOS2, either in chip-pairs of (4,5) and (2,7) is
                    !  guaranteed to be out of range of the RGA spike from Chip 1.
                    !  In that case, it just skips processing.
                  else
                    maskCubeYOnly(cubePlaneIndices(chipTo), nint(yRaw)) = .true.
                    chipsWithRga(chipFrom,chipTo) = .true.
                  end if
                end if
              end do	!  loop over xi (RAWX)
          end do	!  do i=1, size(ChipGroupForRgaMos2Chip1) (for chipTo)
      end do	! do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)

      do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)
        ! RGA spikes to Chip 1. -----
        !   nb., this routine is quite different from the above (X <=> Y etc).
        !        Those from chip1 to chip1 are NOT handled here (see below).
        call message(appMsg, Noisy, 'Masking RGA spikes to Chip 1. (MOS2)')
        do indexChip=1, numChips
          chipFrom = ChipGroupForRgaMos2Chip1(indexChip)
          iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipFrom))	! getCcdNodeNum() defined in ssclib/epic_aux
          write(errStr, '(a,i2,a)') "Skip processing the case from Chip=", chipFrom, "."
          if (iCubePlanes < 1) then	! ==0, see above.
            call message(appMsg, Noisy)
            cycle
          else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
            call message(appMsg, Noisy)
            cycle
          end if

          call CAL_setState(instrument=instrumentId, ccdChipId=chipFrom)
            chipTo = 1
            iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipTo))
            write(errStr, '(a,i2,a)') "Skip processing the case to Chip=", chipTo, "."
            if (iCubePlanes < 1) then	! ==0, see above.
              call message(appMsg, Noisy)
              cycle
            else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
              call message(appMsg, Noisy)
              cycle
            end if

              do yi = 1, chipYSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),:,yi))) then
                  call CAL_setState(ccdChipId=chipFrom)
                  call rawToDet(chipXSizeHalf, yi, xDet, yDet)	! in ssclib/coordinate
                  call CAL_setState(ccdChipId=chipTo)
                  call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip)

                  if (.not. isOffChip) then
                    write(errStr, '(a,a,i2,"=>",i2,a,i3,",",i3,a,f10.5,",",f10.5,").")') &
                       "The coordinates fall inside the chip (isOffChip==.false.), "&
                       //"probably because either axis value is at the very edge.: "&
                       , "chipFrom/To=", chipFrom, chipTo &
                       ,  ", (xi,yi)=(", chipXSizeHalf, yi &
                       , "), (xRaw,yRaw)=(", xRaw, yRaw
                    call message(appMsg, Noisy)
                  end if

                  if ((xRaw < 0.5) .or. (xRaw >= chipXSize+0.5)) then	! to avboid nint(xRaw) falls out of range of the index.
                  ! if ((xRaw <= 0) .or. (xRaw >= chipXSize+0.5)) then
                    ! Do nothing...
                    ! NOTE: In MOS2, Chip 1 is guaranteed to be out of range
                    !  of the RGA spike from either in chip-pairs of
                    !  (4,5) and (2,7).
                    !  In that case, it just skips processing.
                  else
                    maskCubeXOnly(cubePlaneIndices(chipTo), nint(xRaw)) = .true.	! chipTo=1
                    chipsWithRga(chipFrom,chipTo) = .true.
                  end if
                end if
              end do	!  loop over yi (RAWY)
        end do		! do indexChip=1, numChips	(for chipFrom)
      end do	! do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)


      do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)
        ! --- RGA spikes from Chip1 to Chip 1. ---
        call message(appMsg, Noisy, 'Masking RGA spikes form Chip 1 to Chip 1. (MOS2)')
          chipFrom = 1
          iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipFrom))	! getCcdNodeNum() defined in ssclib/epic_aux
          write(errStr, '(a,i2,a)') "Skip processing the case from Chip=", chipFrom, "."
          if (iCubePlanes < 1) then	! ==0, see above.
            call message(appMsg, Noisy)
            cycle
          else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
            call message(appMsg, Noisy)
            cycle
          end if

            chipTo = 1
            iCubePlanes = cubePlaneIndices(getCcdNodeNum(chipTo))
            write(errStr, '(a,i2,a)') "Skip processing the case to Chip=", chipTo, "."
            if (iCubePlanes < 1) then	! ==0, see above.
              call message(appMsg, Noisy)
              cycle
            else if (.not. localFlagsIfRunCubePlane(iCubePlanes)) then
              call message(appMsg, Noisy)
              cycle
            end if

              do xi = 1, chipXSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),xi,:))) then
                  maskCubeXOnly(cubePlaneIndices(chipTo), xi) = .true.
                  chipsWithRga(chipFrom,chipTo) = .true.
                end if
              end do	!  loop over xi (RAWX)
      end do	! do iChipGroup=1, 1	! Dummy loop (to allow "cycle"/exit)


    else
      ! Do nothing
    end if		! if (instrumentId == EMOS1) then

    call CAL_setState(instrument=instrumentIdCal, ccdChipId=ccdNumCal)

    if (verbosity(AppMsg) >= Noisy) then
      do i=1,size(chipsWithRga,1)
        strChips = ''
        call joinChipsToStr(chipsWithRga(i,:), strChips)	! Get a string of " 3, 9,12" etc.
        if (trim(strChips) .ne. '') then
          write(errStr, '(a,i2,a)') "RGA spikes from Chip=(", i &
             , ") are masked in Chips=("//trim(strChips)//")."
          call message(appMsg, Noisy)
        end if
      end do
    end if

  end subroutine calcRgaSpikeMaskProjected


  !---------------------------------------------------------------------
  subroutine joinChipsToStr(aryIndex, strJoined)
  !---------------------------------------------------------------------
  ! Returns a string of " 3, 9,12" etc.

    implicit none 

    logical,      intent(in)  :: aryIndex(:)
    character(*), intent(out) :: strJoined

    character(16) :: strTmp
    logical :: isFirst
    integer :: i

    !-------------------------------------------------------------------
    ! Start:

    isFirst = .true.
    strJoined = ''

    do i=1, size(aryIndex)
      if (aryIndex(i)) then
        if (isFirst) then
          write(strJoined, '(i2)') i
          isFirst = .false.
        else
          strTmp = ''
          write(strTmp, '(",",i2)') i
          strJoined = trim(strJoined)//trim(strTmp)
        end if
      end if
    end do

  end subroutine joinChipsToStr


  !---------------------------------------------------------------------
  subroutine calcRgaSpikeMaskProjected_org(maskPileupCube&
     , chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices&
     , maskCubeXOnly, maskCubeYOnly)
  !---------------------------------------------------------------------

  ! TODO: Tests where there is a bright source on peripheral chips are needed!!!!!!!!!!!

    implicit none 

    !-- Description --
    ! return the projected masks, one compressed for RAWX direction for chip 1,
    ! and the other compressed for RAWY direction for othe chips, indicating
    ! the possible affected field by RGA spikes.
    !-- Note --
    ! RGA spikes appear along the X_CamCoord1 (or 2) direction for MOS1 and 2.
    ! This corresponds to RAWX axe for chip1 but to RAWY axe for the other chips.
    ! Therefore, chip1 and other chips are treated separately here.
    ! 
    ! NOTE!! the cal must be set up by the calling routine for instrument.
    !  Nodes in MOSs are not considered.
    !  maskCubeYOnly and maskCubeXOnly have to be deallocated by the caller after use.

    logical, pointer :: maskPileupCube(:,:,:)
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    integer, pointer :: cubePlaneIndices(:)	! intent(in)
    logical, pointer :: maskCubeXOnly(:,:), maskCubeYOnly(:,:)	! intent(out)

    integer(int16) :: chipXSizeHalf, chipYSizeHalf	! local var.
    integer(int16) :: xi, yi !, refXRaw, refYRaw	! local var.
    integer(int16) :: numChips, iChipGroup, indexChip, i
    real(double) :: xDet, yDet, xRaw, yRaw !, refXDet, refYDet
    logical(bool) :: isOffChip
    ! integer(int8)  :: ccdNum		! for detToRaw() in ssclib/coordinate
    integer(int32) :: chipFrom, chipTo	! for CAL_setState()
    integer(int32) :: instrumentIdCal, ccdNumCal

  ! Originally defined as a global parameter at the top of this file.
  integer(int8), parameter :: ChipGroupForRgaSpike(3,2)&
       = reshape( (/ 2,3,4, 7,6,5 /), (/ 3,2 /) )&
         ! ==> (1,:) = | 2 7 |
         !     (2,:) = | 3 6 |
         !     (3,:) = | 4 5 |
     , ChipGroupForRgaSpikeChip1(2) = (/ 3, 6 /)
   ! See Figure 1 or Sec. 5.2.3.5 "Spatial coordinate systems" in cal document.
   ! See the comments in calcRgaSpikeMaskProjected() for how it is used.


    !-------------------------------------------------------------------
    ! Start:

    allocate(maskCubeYOnly(numCubePlanes, chipYSize)&	! No X-axis
           , maskCubeXOnly(numCubePlanes, chipXSize))	! No Y-axis
    maskCubeYOnly = .false.
    maskCubeXOnly = .false.

    chipXSizeHalf = int(chipXSize)
    chipYSizeHalf = int(chipYSize)

    if (chipXSize .ne. chipYSize) call error('unequalChipSizes')

    call CAL_getState(instrument=instrumentIdCal, ccdChipId=ccdNumCal)

    !.................................................................
    ! Make 1-dimensional mask arrays for RGA spikes
    
    ! --- NOTE ---
    ! RGA spikes appear along the X_CamCoord1 (or 2) direction for MOS1 and 2.
    ! This corresponds to RAWX axes for chip1 but to RAWY axes for other chips.
    ! Therefore, chip1 and other chips are treated separately here.

    if ((instrumentId == EMOS1 .or. instrumentId == EMOS2)) then
      ! For the sake of the speed, ideally, the judgement of instrumentId
      ! should be carried out of the call and the flag should be received
      ! as an argument of this subroutine.

      ! RGA spikes to/from the surrounding chips (except for Chip 1).
       call message(appMsg, Noisy, 'Masking RGA spikes to/from the surrounding chips (except for Chip 1).')
      numChips = size(      ChipGroupForRgaSpike, 2)
      do iChipGroup=1, size(ChipGroupForRgaSpike, 1)	! Loop over a "group" of chips, such as (2, 7) -- see definition of ChipGroupForRgaSpike.

        do indexChip=1, numChips
          ! if (indexChip < 1) cycle
          chipFrom = ChipGroupForRgaSpike(iChipGroup, indexChip)	! Bright source in this chip

          call CAL_setState(instrument=instrumentId, ccdChipId=chipFrom)
          do i=1, numChips
            chipTo = ChipGroupForRgaSpike(iChipGroup, i)	! RGA spike appearing in this chip

            if (chipFrom == chipTo) then	! RGA-spike in the "self"-chip
              ! if (i == indexChip) then	! : the same meaning as above.
              do xi = 1, chipXSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),xi,:))) then
                  maskCubeXOnly(cubePlaneIndices(chipTo), xi) = .true.
                end if
              end do	!  loop over xi (RAWX)
            else			! RGA-spike in the surrounding chips
              do xi = 1, chipXSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),xi,:))) then
                  call CAL_setState(ccdChipId=chipFrom)
                  call rawToDet(xi, chipYSize/2_int16, xDet, yDet)	! in ssclib/coordinate
                  call CAL_setState(ccdChipId=chipTo)
                  call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip)
                  ! call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
                  if (.not. isOffChip) then
                    write(errStr, '(a,a,i2,"=>",i2,a,i3,",",i3,a,f10.5,",",f10.5,").")') &
                       "The coordinates fall inside the chip (isOffChip==.false.), "&
                       //"probably because either axis value is at the very edge.: "&
                       , "chipFrom/To=", chipFrom, chipTo &
                       ,  ", (xi,yi)=(", xi, chipYSize/2_int16 &
                       , "), (xRaw,yRaw)=(", xRaw, yRaw
                    call message(appMsg, Noisy)
                  end if
                  if (xRaw <= 0) then
                    write(errStr, '("chipFrom=",i2," chipto=",i2, " xRaw=",f10.5)') &
                       chipFrom, chipTo, xRaw
                    call error('xRawNegativeInPileupmaskDetToRaw')	! Should never happen.  Just a safety net.
                  else
                    maskCubeXOnly(cubePlaneIndices(chipTo), nint(xRaw)) = .true.
                  end if
                end if
              end do	!  loop over xi (RAWX)
            end if	! if (i == indexChip) then

          end do	!  do i=1, numChips		(for chipTo)
        end do		! do indexChip=1, numChips	(for chipFrom)
      end do		! do iChipGroup=1, size(ChipGroupForRgaSpike, 1)


      !---- RGA spikes to/from Chip 1. ----

      numChips = size(      ChipGroupForRgaSpikeChip1, 1)

        ! RGA spikes from Chip 1.
        !   nb., this routine is similar to the above, except that
        !    the RGA spike to chip1 (the same chip) is not handled
        !    and that the coordinates of X/Y are different.
          call message(appMsg, Noisy, 'Masking RGA spikes from Chip 1.')
          chipFrom = 1
          call CAL_setState(instrument=instrumentId, ccdChipId=chipFrom)
          do i=1, numChips
            chipTo = ChipGroupForRgaSpikeChip1(i)	! RGA spike appearing in this chip

!             ! Checking if the chipTo is in plus or minus direction
!             ! from chipFrom along X (in chip1) axis.
!               refXRaw = chipXSizeHalf+chipXSize
!               call rawToDet(refXRaw, chipYSizeHalf, xDet, yDet)	! in ssclib/coordinate
!               call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
!               refXDet = xDet
!             if (ccdNum == chipTo) then
!                 ! chipTo is in plus direction.
! !print *,ccdNum,xDet,yDet,refXDet
!             else
!               ! refXRaw = chipXSizeHalf-chipXSize
!               !refXRaw = 1
!               refXRaw = chipXSizeHalf
! !print *, '!! Test'
! !print *,ccdNum,xDet,yDet,refXDet
!               call rawToDet(refXRaw, chipYSizeHalf, xDet, yDet)	! in ssclib/coordinate
!               !call rawToDet(int(refXRaw/2,kind=int16), int(chipYSizeHalf/2,kind=int16), xDet, yDet)	! in ssclib/coordinate
!               call rawToDet(int(2,kind=int16), int(2,kind=int16), xDet, yDet)	! in ssclib/coordinate
! !call rawToDet(int(-10, kind=int16), chipYSizeHalf, xDet, yDet)	! in ssclib/coordinate
!               !refXDet = xDet-chipXSizeHalf*80
!               refXDet = xDet
!               !call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
! print *,'detToRaw --- maybe wrong'
!               call detToRaw(refXDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
! ! print *,ccdNum,xDet,yDet,refXDet
!               if (ccdNum == chipTo) then
!                 ! chipTo is in minus direction.
!               else
! if (isOffChip) print *,'===Offchip......'
!                 write(errStr, '(a,i1,a)') "chipTo(",chipTo,") is not found."
!                 call error('unexpectedChipConfiguration')
!               end if
!             end if

              do yi = 1, chipYSize	! in Chip1 (chipFrom)
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),:,yi))) then
                  call CAL_setState(ccdChipId=chipFrom)
                  call rawToDet(chipXSize/2_int16, yi, xDet, yDet)	! in ssclib/coordinate
                  ! call rawToDet(1_int16, yi, xDet, yDet)	! in ssclib/coordinate
                  ! call rawToDet(refXRaw, yi, xDet, yDet)	! in ssclib/coordinate
                  call CAL_setState(ccdChipId=chipTo)
                  call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip)
                  ! call detToRaw(refXDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
                  ! ! call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
! if (isOffChip) then
!    !print *,'=== isOffChip is True!'
!    write (*,'("ccdNum=",i1,", (xDet,yDet)=(",f7.1,",",f7.1,"), (xRaw,yRaw)=(",f5.1,",",f5.1,").")')&
!         ccdNum, xDet,yDet,xRaw,yRaw
! else
!    write (*,'("ccdNum=",i1,", (xDet,yDet)=(",f7.1,",",f7.1,"), (xRaw,yRaw)=(",f5.1,",",f5.1,").")')&
!         ccdNum, xDet,yDet,xRaw,yRaw
! end if
                  if (.not. isOffChip) then
                    write(errStr, '(a,a,i2,"=>",i2,a,i3,",",i3,a,f10.5,",",f10.5,").")') &
                       "The coordinates fall inside the chip (isOffChip==.false.), "&
                       //"probably because either axis value is at the very edge.: "&
                       , "chipFrom/To=", chipFrom, chipTo &
                       ,  ", (xi,yi)=(", chipXSize/2_int16, yi &
                       , "), (xRaw,yRaw)=(", xRaw, yRaw
                    call message(appMsg, Noisy)
                  end if
                  if (xRaw <= 0) then
                    write(errStr, '("chipFrom=",i2,"chipto=",i2, " xRaw=",f10.5)') &
                       chipFrom, chipTo, xRaw
                    call error('xRawNegativeInPileupmaskDetToRaw')	! Should never happen.  Just a safety net.
                  else
                    maskCubeXOnly(cubePlaneIndices(chipTo), nint(xRaw)) = .true.
                  end if
                end if
              end do	!  loop over yi (RAWY)
          end do	!  do i=1, numChips		(for chipTo)

        ! RGA spikes to Chip 1.
        !   nb., this routine is quite different from the above (X <=> Y etc).
        !        Those from chip1 to chip1 are NOT handled here (see below).
        call message(appMsg, Noisy, 'Masking RGA spikes to Chip 1.')
        do indexChip=1, numChips
          chipFrom = ChipGroupForRgaSpikeChip1(indexChip)
          call CAL_setState(instrument=instrumentId, ccdChipId=chipFrom)
            chipTo = 1
              do xi = 1, chipXSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),xi,:))) then
                  call CAL_setState(ccdChipId=chipFrom)
                  call rawToDet(xi, chipYSize/2_int16, xDet, yDet)	! in ssclib/coordinate
                  call CAL_setState(ccdChipId=chipTo)
                  call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip)
                  ! call rawToDet(xi, int(1, kind=int16), xDet, yDet)	! in ssclib/coordinate
                  ! call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
                  ! maskCubeYOnly(cubePlaneIndices(chipTo), nint(yRaw)) = .true.
                  if (.not. isOffChip) then
                    write(errStr, '(a,a,i2,"=>",i2,a,i3,",",i3,a,f10.5,",",f10.5,").")') &
                       "The coordinates fall inside the chip (isOffChip==.false.), "&
                       //"probably because either axis value is at the very edge.: "&
                       , "chipFrom/To=", chipFrom, chipTo &
                       ,  ", (xi,yi)=(", xi, chipYSize/2_int16 &
                       , "), (xRaw,yRaw)=(", xRaw, yRaw
                    call message(appMsg, Noisy)
                  end if
                  if (yRaw <= 0) then
                    write(errStr, '("chipFrom=",i2,"chipto=",i2, " yRaw=",f10.5)') &
                       chipFrom, chipTo, yRaw
                    call error('yRawNegativeInPileupmaskDetToRaw')	! Should never happen.  Just a safety net.
                  else
                    maskCubeYOnly(cubePlaneIndices(chipTo), nint(yRaw)) = .true.	! chipTo=1
                  end if
                end if
              end do	!  loop over xi (RAWX)
        end do		! do indexChip=1, numChips	(for chipFrom)

        ! RGA spikes from Chip1 to Chip 1.
        call message(appMsg, Noisy, 'Masking RGA spikes within Chip 1.')
          chipFrom = 1
            chipTo = 1
              do yi = 1, chipYSize
                if (any(maskPileupCube(cubePlaneIndices(chipFrom),:,yi))) then
                  maskCubeYOnly(cubePlaneIndices(chipTo), yi) = .true.
                end if
              end do	!  loop over yi (RAWY)

    end if		! if ((instrumentId == EMOS1 .or. ...)) then

    call CAL_setState(instrument=instrumentIdCal, ccdChipId=ccdNumCal)

  end subroutine calcRgaSpikeMaskProjected_org


  !---------------------------------------------------------------------
  subroutine readEvent2RawForEootepileupManA(rateThre&
     ! , flagWithApproxPatFrac&
     , chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     , numFrames, maskCube&
     , evCol & ! , expStartTimeStamp&
     ! , rawKeyInfo, wcs, att, outputDiagn&
     , startTime, finisTime, rateCube, minFrameToProcess&
     , isForOote, isForRgaSpike)
  !---------------------------------------------------------------------
  ! Make sure to deallocate, after use, cubePlaneIndices, pattFracSingle,
  !  maskImage, numFrames, rawKeyInfo, outputDiagn, rateCube

    implicit none 

    !-- Description --
    ! read data from the input event for the chip-cube output.
    !-- Note --
    ! numFrames specified here is probably binnedAttsNumFrames(:,attBinNum)

    real(single), intent(in) :: rateThre
    ! logical, intent(in) :: flagWithApproxPatFrac
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    integer, pointer :: cubePlaneIndices(:)
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    real(single), pointer :: numFrames(:)	! Constraints: numFrames(numCubePlanes)
    logical, pointer :: maskCube(:,:,:)	! intent(out)
    type(evlistColumns), intent(in) :: evCol
    real(double), intent(in), optional :: startTime, finisTime
     ! if this is <= 0, then the values defined in the event list are used.
    real(single), pointer, optional :: rateCube(:,:,:) !, maskPileupImage(:,:)	! intent(out)
    integer(int32), intent(in), optional :: minFrameToProcess
    logical, intent(in), optional :: isForOote, isForRgaSpike	! Global variable ForOote can be passed here.

    real(single), pointer :: rateCubeLocal(:,:,:)
    ! real(single), pointer :: maskCubeReal(:,:,:)
    logical, pointer :: maskPileupCube(:,:,:)
    logical, allocatable :: flagsIfRunCubePlane(:)
    integer :: i
logical, parameter :: testFlag = .false.
 
    !-------------------------------------------------------------------
    ! Start:

    call readEvent2RawForEpileup(rateThre&
       ! , flagWithApproxPatFrac&
       , chipXSize, chipYSize&
       , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
       , numFrames, rateCubeLocal, maskPileupCube&
       , evCol, startTime, finisTime)	! defined in epileupmask_shared.f90

!!! Equivalent to the above !!!
!     call readEvent2RateCubeForEpileup(! flagWithApproxPatFrac&
!          chipXSize, chipYSize&
!        , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
!        , numFrames, rateCubeLocal&
!        , evCol, startTime, finisTime)
!     ! call readEvent2RawForEpileup(rateThre& ...)	!! would make unnecessary maskCube.
! 
!     allocate(maskPileupCube(numCubePlanes, chipXSize, chipYSize))
!     allocate(flagsIfRunCubePlane(numCubePlanes))
!     where (rateCubeLocal >= rateThre)
!       maskPileupCube = .true.
!     elsewhere
!       maskPileupCube = .false.
!     end where

    allocate(flagsIfRunCubePlane(numCubePlanes))

if (testFlag) &
print *,'DEBUG: numCubePlanes=',numCubePlanes, ' size(numFrames)=',size(numFrames),' numFrames=',numFrames
    flagsIfRunCubePlane(:) = .true.
    if (present(minFrameToProcess)) then
      flagsIfRunCubePlane(:) = .false.
      do i=1, numCubePlanes
      ! do i=1, min(size(numFrames), size(flagsIfRunCubePlane))
        if (numFrames(i) < minFrameToProcess) then
          write(errStr, '(a,i2,a,f10.2,a,i8,a)') 'Will skip processing for Chip '&
             , i, ' because the number of frames in the current attitude bin (' &
             , numFrames(i), ') is smaller than the threshold (' &
             , minFrameToProcess, ').'
          call message(AppMsg, Verbose)
        else
          flagsIfRunCubePlane(i) = .true.
        end if
      end do
    end if

    call makeOoteRgaSpikeMask(maskPileupCube&
     , chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices&
     , maskCube, flagsIfRunCubePlane=flagsIfRunCubePlane &
     , isForOote=isForOote, isForRgaSpike=isForRgaSpike)	! defined in eootepileupmask_aux
! rateCubeLocal = maskCubeReal	! Debug only!!!

    ! call convertMask2Real(maskCube, maskCubeReal)	! defined in epileupmask_shared

    if (present(rateCube)) then
      rateCube => rateCubeLocal
    else
      deallocate(rateCubeLocal)
    end if

  end subroutine readEvent2RawForEootepileupManA


  !---------------------------------------------------------------------
  subroutine readEvent2RawForEootepileupInfC(rateThre&
     , infoCcdChips, infoFrameAtt &
     ! , flagWithApproxPatFrac&
     ! , chipXSize, chipYSize&
     ! , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     ! , numFrames &
     , maskCube&
     , evCol & ! , expStartTimeStamp&
     ! , rawKeyInfo, wcs, att, outputDiagn&
     , startTime, finisTime, rateCube, minFrameToProcess&
     , isForOote, isForRgaSpike)
  !---------------------------------------------------------------------
  ! Make sure to deallocate, after use, cubePlaneIndices, pattFracSingle,
  !  maskImage, numFrames, rawKeyInfo, outputDiagn, rateCube

    implicit none 

    !-- Description --
    ! read data from the input event for the chip-cube output.
    !-- Note --
    ! numFrames specified here is probably binnedAttsNumFrames(:,attBinNum)

    real(single), intent(in) :: rateThre
    type(infoCcdChipsT), intent(in) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT), intent(in) :: infoFrameAtt	! Defined in epileupmask_shared
    ! logical, intent(in) :: flagWithApproxPatFrac
    ! integer(int16) :: chipXSize, chipYSize
    ! integer(int32) :: instrumentId, numCubePlanes
    ! integer, pointer :: cubePlaneIndices(:)
    ! real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    real(single), pointer :: numFrames(:)	! Constraints: numFrames(numCubePlanes)
    logical, pointer :: maskCube(:,:,:)	! intent(out)
    type(evlistColumns), intent(in) :: evCol
    real(double), intent(in), optional :: startTime, finisTime
     ! if this is <= 0, then the values defined in the event list are used.
    real(single), pointer, optional :: rateCube(:,:,:) !, maskPileupImage(:,:)	! intent(out)
    integer(int32), intent(in), optional :: minFrameToProcess
    logical, intent(in), optional :: isForOote, isForRgaSpike	! Global variable ForOote can be passed here.

    real(single), pointer :: rateCubeLocal(:,:,:)
    ! real(single), pointer :: maskCubeReal(:,:,:)
    logical, pointer :: maskPileupCube(:,:,:)
    logical, allocatable :: flagsIfRunCubePlane(:)
    integer :: i
logical, parameter :: testFlag = .false.
 
    !-------------------------------------------------------------------
    ! Start:

    ! instrumentId  = infoCcdChips%instrumentId
    ! chipXSize     = infoCcdChips%chipXSize
    ! chipYSize     = infoCcdChips%chipYSize
    ! numCubePlanes = infoCcdChips%numCubePlanes
    ! cubePlaneIndices  => infoCcdChips%cubePlaneIndices
    ! pattFracSingle => infoCcdChips%pattFracSingle

    numFrames         => infoFrameAtt%numTotFrames

    call readEvent2RawForEpileup(rateThre&
       , infoCcdChips, infoFrameAtt &
       , rateCubeLocal, maskPileupCube&
       , evCol, startTime, finisTime)
    ! call readEvent2RawForEpileup(rateThre&
    !    ! , flagWithApproxPatFrac&
    !    , chipXSize, chipYSize&
    !    , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
    !    , numFrames, rateCubeLocal, maskPileupCube&
    !    , evCol, startTime, finisTime)	! defined in epileupmask_shared.f90

!!! Equivalent to the above !!!
!     call readEvent2RateCubeForEpileup(! flagWithApproxPatFrac&
!          chipXSize, chipYSize&
!        , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
!        , numFrames, rateCubeLocal&
!        , evCol, startTime, finisTime)
!     ! call readEvent2RawForEpileup(rateThre& ...)	!! would make unnecessary maskCube.
! 
!     allocate(maskPileupCube(numCubePlanes, chipXSize, chipYSize))
!     allocate(flagsIfRunCubePlane(numCubePlanes))
!     where (rateCubeLocal >= rateThre)
!       maskPileupCube = .true.
!     elsewhere
!       maskPileupCube = .false.
!     end where

    allocate(flagsIfRunCubePlane(infoCcdChips%numCubePlanes))

if (testFlag) &
print *,'DEBUG: numCubePlanes=',infoCcdChips%numCubePlanes, ' size(numFrames)=',size(numFrames),' numFrames=',numFrames
    flagsIfRunCubePlane(:) = .true.
    if (present(minFrameToProcess)) then
      flagsIfRunCubePlane(:) = .false.
      do i=1, infoCcdChips%numCubePlanes
      ! do i=1, min(size(numFrames), size(flagsIfRunCubePlane))
        if (numFrames(i) < minFrameToProcess) then
          write(errStr, '(a,i2,a,f10.2,a,i8,a)') 'Will skip processing for Chip '&
             , i, ' because the number of frames in the current attitude bin (' &
             , numFrames(i), ') is smaller than the threshold (' &
             , minFrameToProcess, ').'
          call message(AppMsg, Verbose)
        else
          flagsIfRunCubePlane(i) = .true.
        end if
      end do
    end if

    call makeOoteRgaSpikeMask(maskPileupCube&
       , infoCcdChips &
       , maskCube, flagsIfRunCubePlane=flagsIfRunCubePlane &
       , isForOote=isForOote, isForRgaSpike=isForRgaSpike)	! defined in eootepileupmask_aux
    ! call makeOoteRgaSpikeMask(maskPileupCube&
    !  , chipXSize, chipYSize&
    !  , instrumentId, numCubePlanes, cubePlaneIndices&
    !  , maskCube, flagsIfRunCubePlane=flagsIfRunCubePlane &
    !  , isForOote=isForOote, isForRgaSpike=isForRgaSpike)	! defined in eootepileupmask_aux
! rateCubeLocal = maskCubeReal	! Debug only!!!

    ! call convertMask2Real(maskCube, maskCubeReal)	! defined in epileupmask_shared

    if (present(rateCube)) then
      rateCube => rateCubeLocal
    else
      deallocate(rateCubeLocal)
    end if

  end subroutine readEvent2RawForEootepileupInfC


  !---------------------------------------------------------------------
  subroutine readEvent2SkyForEootepileupManA(rateThre&
     ! , flagWithApproxPatFrac&
     , chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     , numFrames, maskImage&
     , evCol, expStartTimeStamp&
     , rawKeyInfo, wcs, att, outputDiagn&
     , startTime, finisTime, rateCubeRet, minFrameToProcess&
     , isForOote, isForRgaSpike)
  !---------------------------------------------------------------------
  ! Make sure to deallocate, after use, cubePlaneIndices, pattFracSingle,
  !  maskImage, numFrames, rawKeyInfo, outputDiagn, rateCubeRet

    implicit none 

    !-- Description --
    ! read data from the input event for the sky-coord output.
    !-- Note --
    ! numFrames specified here is probably binnedAttsNumFrames(:,attBinNum)

    real(single), intent(in) :: rateThre
    ! logical, intent(in) :: flagWithApproxPatFrac
    ! integer, intent(in) :: xiSize, yiSize
    integer(int16), intent(in) :: chipXSize, chipYSize
    ! integer(int16) :: xi		! xi is a local var.
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    !integer(int32) :: curCubePlane	! curCubePlane is a local var.
    integer, pointer :: cubePlaneIndices(:)
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    ! real(single), pointer :: rateImage(:,:)
    real(single), pointer :: maskImage(:,:), numFrames(:)	! Constraints: numFrames(numCubePlanes)
    ! real(single), pointer :: maskCubeReal(:,:,:)
    ! logical, allocatable :: puMaskCube(:,:,:)
    type(evlistColumns), intent(in) :: evCol
    ! type(TableT), intent(in), optional :: eventTab
    real(double), intent(in) :: expStartTimeStamp
    type(keyInfoType), pointer :: rawKeyInfo(:)	! intent(in)
    type(WcsT), intent(in) :: wcs	! type def in ssclib/wcs_aux
    type(SpacecraftAttitudeType), intent(in) :: att	! == binnedScAtt(attBinNum)%att
    type(eooteDiagnOutputT), intent(in) :: outputDiagn
    ! logical, intent(in), optional :: withRateImage
    real(double), intent(in), optional :: startTime, finisTime
     ! if this is <= 0, then the values defined in the event list are used.
    real(single), pointer, optional :: rateCubeRet(:,:,:) !, maskPileupImage(:,:)	! intent(out)
    integer(int32), intent(in), optional :: minFrameToProcess
    logical, intent(in), optional :: isForOote, isForRgaSpike	! Global variable ForOote can be passed here.

    ! real(single), pointer :: rateCube(:,:,:)
    ! ! integer :: numEvents, row
    ! ! real,           pointer :: evlistSkyX(:), evlistSkyY(:)
    ! ! integer(int32), allocatable :: eventCountCube(:,:,:)
    ! real(single), allocatable :: dummyAttWeights(:,:)
    ! logical :: withRateImageLocal
    ! real(single), pointer :: maskCubeReal(:,:,:)
    ! logical, pointer :: maskPileupCube(:,:,:), maskCube(:,:,:)
    ! real(single), pointer :: maskPileupCubeReal(:,:,:)
    ! logical, allocatable :: flagsIfRunCubePlane(:)
    ! integer :: i

    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT) :: infoFrameAtt	! Defined in epileupmask_shared

logical, parameter :: testFlag = .false.
 
    !-------------------------------------------------------------------
    ! Start:

    infoCcdChips%instrumentId  = instrumentId  
    infoCcdChips%chipXSize     = chipXSize     
    infoCcdChips%chipYSize     = chipYSize     
    infoCcdChips%numCubePlanes = numCubePlanes 
    call allocateInfoCcdChipsT(infoCcdChips &
       , rawKeyInfoPtr=rawKeyInfo &
       , cubePlaneIndicesPtr=cubePlaneIndices &
       , pattFracSinglePtr=pattFracSingle)

    infoFrameAtt%expStartTimeStamp = expStartTimeStamp
    call allocateInfoFrameAttT(infoFrameAtt &
       , binnedScAttSize=1 &
       , numTotFramesPtr=numFrames)
    infoFrameAtt%binnedScAtt(1)%isGood = .true.
    infoFrameAtt%binnedScAtt(1)%gtiDuration = 1.0d0 ! dummy value.
    infoFrameAtt%binnedScAtt(1)%att = att

    call readEvent2SkyForEootepileupInfC(rateThre&
       , infoCcdChips, infoFrameAtt &
       , maskImage &
       , evCol &
       , 1 &	! ==attBinNum
       , wcs &
       , outputDiagn&
       , startTime, finisTime, rateCubeRet, minFrameToProcess&
       , isForOote, isForRgaSpike)

    ! infoFrameAtt%isAllocatedNumTotFrames = .false.	! %numTotFrames must not be deallocated here, as it is linked to the external pointer.
    call deallocateInfoFrameAttT(infoFrameAtt, doNumTotFrames=.false.)

  end subroutine readEvent2SkyForEootepileupManA

  !---------------------------------------------------------------------
  subroutine readEvent2SkyForEootepileupInfC(rateThre&
     , infoCcdChips, infoFrameAtt &
     ! , flagWithApproxPatFrac &
     ! , chipXSize, chipYSize&
     ! , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     ! , numFrames &
     , maskImage &
     , evCol &
     ! , expStartTimeStamp&
     , attBinNum &
     ! , rawKeyInfo &
     , wcs &
     ! , att &
     , outputDiagn&
     , startTime, finisTime, rateCubeRet, minFrameToProcess&
     , isForOote, isForRgaSpike)
  !---------------------------------------------------------------------
  ! Make sure to deallocate, after use, cubePlaneIndices, pattFracSingle,
  !  maskImage, numFrames, rawKeyInfo, outputDiagn, rateCubeRet

    implicit none 

    !-- Description --
    ! read data from the input event for the sky-coord output.
    !-- Note --
    ! numFrames specified here is probably binnedAttsNumFrames(:,attBinNum)

    real(single), intent(in) :: rateThre
    type(infoCcdChipsT), intent(in) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT), intent(in) :: infoFrameAtt	! Defined in epileupmask_shared
    ! logical, intent(in) :: flagWithApproxPatFrac
    ! integer, intent(in) :: xiSize, yiSize
    ! integer(int16) :: chipXSize, chipYSize
    ! integer(int16) :: xi		! xi is a local var.
    ! integer(int32) :: instrumentId, numCubePlanes
    !integer(int32) :: curCubePlane	! curCubePlane is a local var.
    ! integer, pointer :: cubePlaneIndices(:)
    ! real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    ! real(single), pointer :: rateImage(:,:)
    real(single), pointer :: maskImage(:,:), numFrames(:)	! Constraints: numFrames(numCubePlanes)
    ! real(single), pointer :: maskCubeReal(:,:,:)
    ! logical, allocatable :: puMaskCube(:,:,:)
    type(evlistColumns), intent(in) :: evCol
    integer, intent(in) :: attBinNum	! Index for binnedScAtt
    ! type(TableT), intent(in), optional :: eventTab
    real(double) :: expStartTimeStamp
    type(keyInfoType), pointer :: rawKeyInfo(:)	! intent(in)
    type(WcsT), intent(in) :: wcs	! type def in ssclib/wcs_aux
    type(SpacecraftAttitudeType) :: att	! == binnedScAtt(attBinNum)%att
    type(eooteDiagnOutputT), intent(in) :: outputDiagn
    ! logical, intent(in), optional :: withRateImage
    real(double), intent(in), optional :: startTime, finisTime
     ! if this is <= 0, then the values defined in the event list are used.
    real(single), pointer, optional :: rateCubeRet(:,:,:) !, maskPileupImage(:,:)	! intent(out)
    integer(int32), intent(in), optional :: minFrameToProcess
    logical, intent(in), optional :: isForOote, isForRgaSpike	! Global variable ForOote can be passed here.

    real(single), pointer :: rateCube(:,:,:)
    ! integer :: numEvents, row
    ! real,           pointer :: evlistSkyX(:), evlistSkyY(:)
    ! integer(int32), allocatable :: eventCountCube(:,:,:)
    real(single), allocatable :: dummyAttWeights(:,:)
    logical :: withRateImageLocal
    real(single), pointer :: maskCubeReal(:,:,:)
    logical, pointer :: maskPileupCube(:,:,:), maskCube(:,:,:)
    real(single), pointer :: maskPileupCubeReal(:,:,:)
    logical, allocatable :: flagsIfRunCubePlane(:)
    integer :: i
logical, parameter :: testFlag = .false.
 
    !-------------------------------------------------------------------
    ! Start:

    ! instrumentId  = infoCcdChips%instrumentId
    ! chipXSize     = infoCcdChips%chipXSize
    ! chipYSize     = infoCcdChips%chipYSize
    ! numCubePlanes = infoCcdChips%numCubePlanes
    ! cubePlaneIndices => infoCcdChips%cubePlaneIndices
    ! pattFracSingle   => infoCcdChips%pattFracSingle
    rawKeyInfo       => infoCcdChips%rawKeyInfo

    numFrames         => infoFrameAtt%numTotFrames
    expStartTimeStamp =  infoFrameAtt%expStartTimeStamp
    att = infoFrameAtt%binnedScAtt(attBinNum)%att

    withRateImageLocal = outputDiagn%withRateImage

    call readEvent2RawForEpileup(rateThre&
       , infoCcdChips, infoFrameAtt &
       , rateCube, maskPileupCube&
       , evCol, startTime, finisTime)
    ! call readEvent2RawForEpileup(rateThre&
    !    ! , flagWithApproxPatFrac&
    !    , chipXSize, chipYSize&
    !    , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
    !    , numFrames, rateCube, maskPileupCube&
    !    , evCol, startTime, finisTime)	! defined in epileupmask_shared.f90

!!! Equivalent to the above !!!
!     call readEvent2RateCubeForEpileup(! flagWithApproxPatFrac&
!          chipXSize, chipYSize&
!        , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
!        , numFrames, rateCube&
!        , evCol, startTime, finisTime)
!     ! call readEvent2RawForEpileup(rateThre& ...)	!! would make unnecessary maskCube.
! 
!     allocate(maskPileupCube(numCubePlanes, chipXSize, chipYSize))
!     allocate(flagsIfRunCubePlane(numCubePlanes))
!     where (rateCube >= rateThre)
!       maskPileupCube = .true.
!     elsewhere
!       maskPileupCube = .false.
!     end where

    allocate(flagsIfRunCubePlane(infoCcdChips%numCubePlanes))

if (testFlag) &
print *,'DEBUG: numCubePlanes=',infoCcdChips%numCubePlanes, ' size(numFrames)=',size(numFrames),' numFrames=',numFrames
    flagsIfRunCubePlane(:) = .true.
    if (present(minFrameToProcess)) then
      flagsIfRunCubePlane(:) = .false.
      do i=1, infoCcdChips%numCubePlanes
      ! do i=1, min(size(numFrames), size(flagsIfRunCubePlane))
        if (numFrames(i) < minFrameToProcess) then
          maskPileupCube(i,:,:) = .false.	! Pile-up on the chip is ignored.
          write(errStr, '(a,i2,a,f10.2,a,i8,a)') 'Will skip processing for Chip '&
             , i, ' because the number of frames in the current attitude bin (' &
             , numFrames(i), ') is smaller than the threshold (' &
             , minFrameToProcess, ').'
          call message(AppMsg, Verbose)
        else
          flagsIfRunCubePlane(i) = .true.
        end if
      end do
    end if

    call makeOoteRgaSpikeMask(maskPileupCube&
       , infoCcdChips &
       , maskCube, flagsIfRunCubePlane=flagsIfRunCubePlane &
       , isForOote=isForOote, isForRgaSpike=isForRgaSpike)	! defined in eootepileupmask_aux
    ! call makeOoteRgaSpikeMask(maskPileupCube&
    !  , chipXSize, chipYSize&
    !  , instrumentId, numCubePlanes, cubePlaneIndices&
    !  , maskCube, flagsIfRunCubePlane=flagsIfRunCubePlane &
    !  , isForOote=isForOote, isForRgaSpike=isForRgaSpike)	! defined in eootepileupmask_aux
! rateCube = maskCubeReal	! Debug only!!!

    call convertMask2Real(maskCube, maskCubeReal)	! defined in epileupmask_shared
    deallocate(maskCube)

    allocate(dummyAttWeights(infoCcdChips%numCubePlanes, 1))
    dummyAttWeights = 1.0

    call message(appMsg, Verbose, 'Converting mask images into Sky coordinates.')
if (testFlag) then
print *,'DEBUG: maskCubeReal=',sum(maskCubeReal),maxval(maskCubeReal),minval(maskCubeReal)
print *,'DEBUG: size(maskImage,*)=',size(maskImage,1),size(maskImage,2)
end if
    call rebinChipImagesToSky(maskCubeReal&
       , rawKeyInfo%ccdNum, wcs, (/att/)&	! ==(/binnedScAtt(attBinNum)%att/)&
       ! , dummyAttWeights, expStartTimeStamp, 'rate'&
       , dummyAttWeights, expStartTimeStamp, 'count'&
       , .false., maskImage)
       !, doApprox=.true., outImage=maskImage)
     ! in eimchip2sky/eimchip2sky_aux.
     ! Note: Here only one bin in attitude is used, so (/att/) does the job.
     !  rebinChipImagesToSky() is meant to "rebin" the input over the given
     !  attitude.  In this case only we need is just the simple sky-coordinate
     !  conversion, but not further rebinning.
     ! Note: In rebinChipImagesToSky() in eimchip2sky_aux, the variable
     !  "binnedScAtt" is different from the same name ones used elsewhere!
     !  There, it actually refers to binnedScAtt(:)%att in the ordinary notation.

    !.................................................................
    ! Creates diagnostic outputs if wanted (plus, some clean-up):

    if (outputDiagn%withPileupMask) then
      call convertMask2Real(maskPileupCube, maskPileupCubeReal)	! defined in epileupmask_shared
      ! allocate(maskPileupCubeReal(numCubePlanes, chipXSize, chipYSize))
      ! where (maskPileupCube)
      !   maskPileupCubeReal = 1.0
      ! elsewhere
      !   maskPileupCubeReal = 0.0
      ! end where

      call rebinChipImagesToSky(maskPileupCubeReal&
         , rawKeyInfo%ccdNum, wcs, (/att/)&	! ==(/binnedScAtt(attBinNum)%att/)&
         , dummyAttWeights, expStartTimeStamp, 'rate'&
         , .false., outputDiagn%maskPileupImage)

      deallocate(maskPileupCubeReal)
    end if

    deallocate(maskPileupCube)

    if (withRateImageLocal) then
      call message(appMsg, Verbose, 'Converting rate images into Sky coordinates.')
      call rebinChipImagesToSky(rateCube&
         , rawKeyInfo%ccdNum, wcs, (/att/)&	! ==(/binnedScAtt(attBinNum)%att/)&
         , dummyAttWeights, expStartTimeStamp, 'rate'&
         , .false., outputDiagn%rateImage)
    end if

    deallocate(dummyAttWeights)

    if (present(rateCubeRet)) then
      rateCubeRet => rateCube
    else
      deallocate(rateCube)
    end if

      ! How does this call work to make a mask? Ok, we have sent to
      ! rebinChipImagesToSky a cube in which a value of 0 represents a
      ! CCD column for which the pileup criterion has not been exceeded
      ! at any pixel during the present attitude interval, and a value
      ! of 1 represents the contrary. This cube is reprojected to sky
      ! coordinates and added to outImage. At the end of the process
      ! outImage will therefore have 0-valued and non-0-valued pixels.
      ! The 0 values represent those places on the sky where the camera
      ! has never had a piled-up column; >0 values represent the contrary.
      ! These >0 areas of outImage are translated into a logical mask
      ! of the piled-up areas during the final call to saveImage.

  end subroutine readEvent2SkyForEootepileupInfC


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module eootepileupmask_aux

