! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: eootepileupmask_mod
!
! Authors:  Ian Stewart - SSC/LUX
!        :  Masaaki Sakano - SSC/LUX
!
! Description: This task produces a mask file, which masks the out-of-time
!  event (OOTE) streaks and RGA spikes, caused by the pile-up pixels.
!  To estimate the pile-up, it calculates a single-pixel event count rate
!  per frame for each pixel, normalised to 1.49 keV Al-K line based on
!  the grade branching ratio, which is a function of energy, and judges
!  whether the calculated rate exceeds the given threshold.
!
! Comments: The code is largely based on epileupmask_mod.f90 .
! 
! TODO: 
!  - Consider the attitude if specified, even if OutputStyle=raw.  There is no good reason not to...
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eootepileupmask_mod

  use dal
  use param  
  use errorHandling
  use cal
  ! use CalTypes	! used in epileupmask_shared

!  use temp_ssclib

  use binned_att	! in attcalc2
  use cube_aux		! in eimchip2sky_aux
  use eimchip2sky_aux	! in eimchip2sky

  ! Followings in ssclib
  use array_utils
  use cal_aux
use coordinate ! only needed for applyAttcalcFudge
  use dal_aux
  ! use dss_aux
  ! use intervals_aux	! for gtiVecType
  use save_image
  use ssc_misc
  use wcs_aux

  use eootepileupmask_aux
  use epileupmask_shared
  ! use ebkgmap_skyframe	! called from epileupmask_shared
  ! use ebkgmap_attcommon	! called from epileupmask_shared

  implicit none 

  private 

  public eootepileupmask

  ! Command-line parameters:
  character(256) ::&
    InputStyle='',&		! not used as a command-line argument.
    EventSetName='',&
    ! ImageSetName='',&		! not used as a command-line argument.
    OutputStyle='',&
    TemplateImageSetName='',&
    MaskSetName='',&
    PileupMaskSetRoot='',&
    AttitudeStyle='',&
    BinnedAttSetTabName='',&
    RateSetRoot=''
  real(single) ::&
    RateThreshold=0.0,&
    ! EvLow=0.0,&			! not used as a command-line argument.
    ! EvHigh=0.0,&		! not used as a command-line argument.
    AvgCntToProcess=0.0
  logical ::&
    ! IsSingleEventOnly = .false.,&	! not used as a command-line argument.
    WithRateImage = .false.,&
    WithApproxPatFrac = .true.,&
    WithPileupMask = .false.,&
    ForOote = .true.,&
    ForRgaSpike = .false.,&
    IsInverted = .true.

contains 

  !---------------------------------------------------------------------
  subroutine eootepileupmask
  !---------------------------------------------------------------------

    implicit none 

    character(256), parameter :: temporaryImageFileName = 'eootepileupmask_tmpimage0001.ds'

    type(DataSetT) :: inputSet
    character(20) :: messageInputFile
    ! integer(int32) :: instrumentId, numCubePlanes
    integer(int32) :: minFrameToProcess
    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT) :: infoFrameAtt	! Defined in epileupmask_shared
    integer(int8) :: numCcds, numNodes !, ccdNum , nodeNum
    real(single), pointer :: rateCube(:,:,:)	! , maskPileupImage(:,:)
    logical, pointer :: maskCube(:,:,:) => null()
    real(single), pointer :: maskImageRealTmp(:,:) => null()
         ! The mask defined here is real-valued, which makes it easier
         ! to rebin it to sky coordinates (if specified so).
    integer(int8), pointer :: maskImageInt(:,:) => null(), maskImageIntTmp(:,:) => null()
    integer :: xiSize, yiSize, numProcessedFrames
    character(256) :: outputRefSetName

    type(WcsT) :: wcs	! type def in ssclib/wcs_aux
    integer :: numAttBins, attBinNum

    type(TableT) :: eventTab
    type(evlistColumns) :: evCol	! typedef in epileupmask_shared
    real(double) :: startTime, finisTime
    logical :: isCycle, isExisting
    type(eooteDiagnOutputT), allocatable :: outputDiagn(:)	! type def in eootepileupmask_aux
    character(280) :: tmpFilename
logical, parameter :: isDebug = .false.
! logical, parameter :: isDebug = .true.

    !-------------------------------------------------------------------
    ! Start:

!call testssclibraw2det()
!return

    rateCube => null()

    call readParameters
    minFrameToProcess = ceiling(AvgCntToProcess/RateThreshold)

    ! Setting up inputSet etc.
    !   Get outputRefSetName, messageInputFile, inputSet, rateImage(:,:)
    call initialSetUpForEpileup(InputStyle, EventSetName, TemplateImageSetName&
       , outputRefSetName, inputSet, messageInputFile) ! , rateImage)	! TemplateImageSetName is a dummy argument here.
       ! rateImage is only set if (InputStyle=='image'), which is never the case as it is not supported in this task.

    ! Setting up for instument-related stuffs.
    !   Get instrumentId, chip(X|Y)Size, numCcds, numNodes
    call instSetUpForEpileup(inputSet, infoCcdChips, withTlminmax=.true.)
    numCcds      = infoCcdChips%numCcds
    numNodes     = infoCcdChips%numNodes

    !...................................................................
    ! Get Attitude and Sky coordinates frame infomation.
    if ((OutputStyle .eq. 'sky') .or. (InputStyle .eq. 'image')) then

      !   Get wcs, (x|y)iSize, binnedScAtt(:), expStartTimeStamp, medianScAttitude
      call getAttSkyCoordForEpileup(TemplateImageSetName, AttitudeStyle, BinnedAttSetTabName&
         , wcs, xiSize, yiSize&
         , infoFrameAtt )
      numAttBins = size(infoFrameAtt%binnedScAtt)

!*** NOTE! the only things that the template image supplies are (i) wcs, (ii) image pixel dimensions, (iii) median attitude and (iv) exposure start time stamp.  It would be nice to have some facility to input the last of these in other ways.

      allocate(outputDiagn(numAttBins))
      ! call allocateEooteDiagnOutputT(outputDiagn,    WithRateImage, WithPileupMask, xiSize, yiSize)
      ! call allocateEooteDiagnOutputT(outputDiagnTmp, WithRateImage, WithPileupMask, xiSize, yiSize)

      if (OutputStyle .eq. 'sky') then
        allocate(maskImageInt(xiSize,yiSize)&	!, maskImageReal(xiSize,yiSize)
           , maskImageIntTmp(xiSize,yiSize), maskImageRealTmp(xiSize,yiSize))
        ! maskImageReal = 0
        maskImageInt = 0
        maskImageIntTmp  = 0
        maskImageRealTmp = 0.0
      end if
      ! if (InputStyle .ne. 'image') then	! Unnecessary because (InputStyle == 'evlist') always in this code.
        ! If yes, it is already set in initialSetUpForEpileup() in epileupmask_shared
        ! allocate(rateImage(xiSize,yiSize))	!, rateImageTmp(xiSize,yiSize))
        ! rateImage = 0
      ! end if

    end if	! if ((OutputStyle .eq. 'sky') .or. (InputStyle .eq. 'image')) then


    !...................................................................
    ! Count number of frames of each chip.
    !   Get rawKeyInfo(:), numTotFrames(:), cubePlaneIndices(:), numCubePlanes,
    !       binnedAttsNumFrames(1:numCubePlanes, 1:numAttBins)
    !  See comments in countNumberOfFrames() in epileupmask_shared.f90 for detail.
    call countNumberOfFrames(inputSet, InputStyle, OutputStyle&
     , infoCcdChips, infoFrameAtt &
     , messageInputFile )

if (isDebug .and. trim(OutputStyle)=='sky') then
print *,'DEBUG: finished countNumberOfFrames()'
do attBinNum = 1, size(infoFrameAtt%binnedAttsNumFrames,1)
!do attBinNum = 1, numAttBins
print *,'DEBUG: binnedAttsNumFrames(',attBinNum,')=',infoFrameAtt%binnedAttsNumFrames(attBinNum,:)
end do
print *,'DEBUG: After countNumberOfFrames()'
end if


    ! Get the Array of fraction of GRADE_SINGLE events as a function of PI.
    !   Get pattFracSingle
    call getPatSingleFractionArrayTmpl(infoCcdChips)


    !...................................................................
    ! MAIN: Now calculate rateImage/rateCube, maskImage/maskCube

    call message(appMsg, Verbose, 'Reading data from the '&
         //trim(messageInputFile)//'.')

    ! Get evCol, eventTab (as well as some checks).
    !   n.b. The latter is required just to release() later on.
    call setupToReadDataForEpileup(InputStyle, OutputStyle&
       , inputSet, infoCcdChips &
       , evCol, eventTab)
    ! call setupToReadDataForEpileup(InputStyle, OutputStyle&
    !    ! , inputSet, chipXSize, chipYSize, maxCcdNodeNum&
    !    , inputSet, chipXSize, chipYSize & !, size(cubePlaneIndices)&
    !    , cubePlaneIndices& 
    !    , evCol, eventTab)

    !...................................................................
    ! MAIN select case:
    ! numFrames => numTotFrames
    select case(InputStyle)
    case('evlist')

      select case (OutputStyle)
      case('sky')
        ! rateImage = 0.0
        maskImageInt = 0
        ! maskImageReal = 0.0
        write(errStr, '(a,i8,a,i8,a)') 'Will skip any attitude bin, containing less than'&
           , minFrameToProcess, ' frames, out of', numAttBins, ' attitude bins.'
        call message(AppMsg, Verbose)

        numProcessedFrames = 0
        do attBinNum = 1, numAttBins

          call calcStartFinisTime(infoFrameAtt, attBinNum &
             , evCol, minFrameToProcess &
             , startTime, finisTime, isCycle)
          if (isCycle) cycle

          numProcessedFrames = numProcessedFrames + 1
          ! rateImageTmp = 0
          call allocateEooteDiagnOutputT(outputDiagn(attBinNum) &
             , WithRateImage, WithPileupMask, xiSize, yiSize)	! If the frame is skipped, outputDiagn(attBinNum) is not allocated or initialised.	! defined in eootepileupmask_aux
          call readEvent2SkyForEootepileup(RateThreshold&
             , infoCcdChips, infoFrameAtt &
             ! , WithApproxPatFrac&
             , maskImageRealTmp&
             , evCol, attBinNum &
             , wcs, outputDiagn(attBinNum)&
             , startTime=startTime, finisTime=finisTime, rateCubeRet=rateCube &
             , minFrameToProcess=minFrameToProcess &
             , isForOote=ForOote, isForRgaSpike=ForRgaSpike)
             ! Get rateImageTmp, maskImageRealTmp

!------- DEBUG ---------
if (isDebug) then
        call saveImage(maskImageRealTmp, TemplateImageSetName, 'maskImageRealTmp.ds'&
           , trim(temporaryImageFileName), wcs=wcs) ! in ssclib/save_image.
end if
!------- DEBUG (up to here) ---------

          ! Now, how shold we calculate the output logical mask?
          ! If we want the mask based on the averaged rate value, then
          !    maskImageReal = maskImageReal + maskImageRealTmp
          ! here (ie., for each attitude bin(=interval)) and 
          ! calculating the mask file based on the final maskImageReal
          ! will do the job.  However, if a pixel at a certain time interval
          ! exceeds the given threshold, namely piled-up, a result containing
          ! that pixel must be unreliable in a conservative approach.
          ! Indeed if the attitude is fluctuated enough and if a source
          ! flux exceeds just above the threshold, then the resultant
          ! `averaged' rate image may not give any masked pixel,
          ! which is undesirable.  Therefore, the latter conservative approach
          ! is accepted here.  The task will mask any pixel
          ! that exceeds the threshold at any time interval.
          maskImageIntTmp = 0
          where(maskImageRealTmp > 0.3) maskImageIntTmp = 1
           ! Here, if the threshold is 0, instead of 0.3, it will give the mask with a bit wider coverage.
          maskImageInt = maskImageInt + maskImageIntTmp

          if (verbosity(appMsg) >= Verbose) then			! Should be outside of the loop, ideally..
            if (any(maskImageIntTmp > 0)) then
              write(errStr, '(a, i4, ".")') &
                 "Found piled-up pixel(s) in the attitude bin", attBinNum 
              ! call message(appMsg, Verbose, 'Found piled-up pixel(s).')
            else
              write(errStr, '(a, i4, ".")') &
                 "Found NO piled-up pixel in the attitude bin", attBinNum 
              ! call message(appMsg, Verbose, 'Found NO piled-up pixel.')
            end if
            call message(appMsg, Verbose)
          end if
        end do	! do attBinNum = 1, numAttBins ! for case('evlist'),OutputStyle==case('sky')

        if (numProcessedFrames < 1) then
          deallocate(maskImageInt, maskImageIntTmp, maskImageRealTmp)
          call deallocateEooteDiagnOutputT(outputDiagn(:))	! defined in eootepileupmask_aux
          call warning('noFrameProcessed' &
             , 'No frame has been processed, hence no output is created.')

          ! Final clean up before the end of this task.
          call deallocateInfoCcdChipsT(infoCcdChips)
          call deallocateInfoFrameAttT(infoFrameAtt)

          inquire(file=trim(MaskSetName), exist=isExisting)
          if (isExisting) call error('noOutputMade'&
             , 'The specified mask file ('//trim(MaskSetName)&
             //') already exists, but is not overwritten.')
          return	!! END of the task!
        end if

        !! If you want to calculate the output mask based on the average, this is the way...
        ! maskImageInt = 0
        ! where(maskImageReal > 0.3) maskImageInt = 1
        ! if (IsInverted) then
        !   where(maskImageInt > 0)
        !     maskImageInt = 0
        !   elsewhere
        !     maskImageInt = 1
        !   end where
        ! else
        where(maskImageInt > 1) maskImageInt = 1	! (... > 0) is OK, too.
        ! end if

      case('raw')
          call readEvent2RawForEootepileup(RateThreshold&
             , infoCcdChips, infoFrameAtt &
             ! , WithApproxPatFrac&
             , maskCube, evCol &
             , rateCube=rateCube &
             , isForOote=ForOote, isForRgaSpike=ForRgaSpike)

      case default
         call error('badOutputStyleB3',"Should not happen.  Contact the code developer.")
      end select	! select case (styleOut) in case(styleIn)=('image')

      call release(eventTab)

    ! case('image')	! select case(styleIn)
    !   select case (OutputStyle)
    !   case('sky')
    !     call readImage2SkyForEpileup(RateThreshold&
    !        , IsSingleEventOnly, WithApproxPatFrac&
    !        , xiSize, yiSize&
    !        , instrumentId, cubePlaneIndices, pattFracSingle&
    !        , numFrames, rateImage, maskImageInt&
    !        , expStartTimeStamp, wcs, medianScAttitude&
    !        , (EvLow+EvHigh)*500.0)	! The Last-argument == E/2*1000 (eV)
    !     allocate(rateCube(0,0,0))
    ! 
    !   case('raw')
    !     allocate(rateCube(numCubePlanes, chipXSize, chipYSize))
    !     allocate(maskCube(numCubePlanes, chipXSize, chipYSize))
    !     call readImage2RawForEpileup(RateThreshold&
    !        , IsSingleEventOnly, WithApproxPatFrac&
    !        , xiSize, yiSize&
    !        , instrumentId, cubePlaneIndices, pattFracSingle&
    !        , numFrames, rateImage, rateCube, maskCube&
    !        , expStartTimeStamp, wcs, medianScAttitude&
    !        , (EvLow+EvHigh)*500.0)	! The Last-argument == E/2*1000 (eV)
    ! 
    !   case default
    !      call error('badOutputStyleB3_2',"Should not happen.  Contact the code developer.")
    !   end select	! select case (styleOut) in case(styleIn)=('image')

    case default
      call error('badInputStyleB2',"Should not happen.  Contact the code developer.")
    end select	! select case(styleIn)

    call release(inputSet)


    !...................................................................
    ! Output:

    select case(OutputStyle)
    case('raw')

      if (IsInverted) then
        call invertMask(maskCube)	! ssclib/array_utils
      end if

      call message(appMsg, Verbose, 'Writing mask cube.')
      call writeCubeData(trim(MaskSetName), trim(outputRefSetName), maskCube&
        , infoCcdChips%rawKeyInfo)
        ! , rawKeyInfo(:numCubePlanes))
      ! in eimchip2sky/cube_aux.
      ! call message(appMsg, Verbose, 'Mask cube written.')

      if (WithRateImage) then
        call message(appMsg, Verbose, 'Writing rate cube.')
        call writeCubeData(trim(RateSetRoot)//'.ds', trim(outputRefSetName), rateCube&
           , infoCcdChips%rawKeyInfo)
           ! , rawKeyInfo(:numCubePlanes))
        ! in eimchip2sky/cube_aux.
      end if

    case('sky')	! from ./eootemap_mod.f90

      if (IsInverted) then
        call invertMask(maskImageInt)	! ssclib/array_utils
      end if

!       allocate(binnedAttWeights(numCubePlanes, numAttBins))
! 
!       !.................................................................
!       ! Construct the attitude vs. image table of weights:
! 
!       call message(AppMsg, Verbose, "Calculating attitude weights.")
! 
!       do imageNum = 1, numCubePlanes
!         call gtiWeightBinnedAttitude(gtiVec(imageNum)%gtis, binnedScAtt)
!         ! in attcalc2/binned_att.
! 
!         binnedAttWeights(imageNum,:) = binnedScAtt%gtiDuration
!       end do ! imageNum
! 
! !********** normalize?????

      call   message(appMsg, Verbose, 'Writing mask image.')
      call   saveImage(maskImageInt, TemplateImageSetName, MaskSetName&
           , trim(temporaryImageFileName), wcs=wcs) ! in ssclib/save_image.

      if (WithRateImage) then
        call message(appMsg, Verbose, 'Writing rate image.')
        tmpFilename = ''
        do attBinNum = 1, numAttBins
          if (.not. outputDiagn(attBinNum)%withRateImage)  cycle
          write(tmpFilename,'(a,i4.4,a)') trim(RateSetRoot), attBinNum, '.ds'
          call saveImage(outputDiagn(attBinNum)%rateImage &
             , trim(TemplateImageSetName), trim(tmpFilename) &
             , trim(temporaryImageFileName), wcs=wcs) ! in ssclib/save_image.
        end do	! do attBinNum = 1, numAttBins
      end if

      if (WithPileupMask) then
        call message(appMsg, Verbose, 'Writing pile-up mask images.')
        tmpFilename = ''
        do attBinNum = 1, numAttBins
          if (.not. outputDiagn(attBinNum)%withPileupMask)  cycle
          write(tmpFilename,'(a,i04.4,a)') trim(PileupMaskSetRoot), attBinNum, '.ds'
          call saveImage(outputDiagn(attBinNum)%maskPileupImage &
             , trim(TemplateImageSetName), trim(tmpFilename) &
             , trim(temporaryImageFileName), wcs=wcs) ! in ssclib/save_image.
          call message(appMsg, Verbose, ' File ('//trim(tmpFilename)//') written.')
        end do	! do attBinNum = 1, numAttBins
      end if

if (isDebug) then
        call message(appMsg, Verbose, 'Writing (ratecube.ds)')
        call writeCubeData('ratecube.ds', TemplateImageSetName, rateCube&
           , infoCcdChips%rawKeyInfo)
           ! , rawKeyInfo(:numCubePlanes))
end if

      deallocate(maskImageInt, maskImageIntTmp, maskImageRealTmp)
!!! For DEBUG
!call writeCubeData('rate_debug.ds', outputRefSetName, rateCube&
!   , rawKeyInfo(:numCubePlanes))
!!! Note: In order to activate the above, you need to add rateCube argument in the above call readEvent2SkyForEpileup(RateThreshold ...)

    case default
      call error('badOutputStyleB2_2',"Should not happen.  Contact the code developer.")
    end select

!print *,'DEBUG:deallocating rawKeyInfo, rateCube'
    !deallocate(rawKeyInfo, rateCube)	! rateCube may not be allocated...?
    !deallocate(rawKeyInfo)
!print *,'deallocating rateCube'
    deallocate(rateCube)
!print *,'deallocated rawKeyInfo, rateCube'
    if ((OutputStyle .eq. 'sky') .or. (InputStyle .eq. 'image')) then
      ! deallocate(binnedScAtt, binnedAttsNumFrames)
    end if

! print *,'deallocating outputDiagn'
    if (allocated(outputDiagn)) call deallocateEooteDiagnOutputT(outputDiagn)	! n.b., automatically judged if each of them is allocated or not.	! defined in eootepileupmask_aux
! print *,'outputDiagn deallocated.'
    ! call deallocateEooteDiagnOutputT(outputDiagn)	! n.b., automatically judged if they are allocated or not.
    ! call deallocateEooteDiagnOutputT(outputDiagnTmp)	! n.b., automatically judged if they are allocated or not.

    call deallocateInfoCcdChipsT(infoCcdChips)
    call deallocateInfoFrameAttT(infoFrameAtt)

    ! puProbCube = 1.0 - exp(-puProbCube)

    ! if (withPuProbCube)&
    !   call writeCubeData(puProbCubeSetName, outputRefSetName, puProbCube&
    !     , rawKeyInfo(:numCubePlanes))
    !   ! in eimchip2sky/cube_aux.

    ! allocate(puMaskCube(numCubePlanes, chipXSize, chipYSize))
    ! puMaskCube = (puProbCube < pileupMaskProbability)

  end subroutine eootepileupmask


  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    RateThreshold       = realParameter('ratethreshold')
    MaskSetName         = stringParameter('maskset')

    InputStyle          = 'evlist'	! Parameter
    !select case(InputStyle)
    !case('evlist')
      EventSetName        = stringParameter('eventset')
    !end select

    OutputStyle         = stringParameter('outputstyle')
    select case(OutputStyle)
    case('sky')
      TemplateImageSetName     = stringParameter('templateset')
        if (trim(TemplateImageSetName) == '') call error('nullTemplateImageSetName'&
           , 'templateset must be specified when outputstyle=sky.')

      AttitudeStyle = stringParameter('attstyle')
      select case(AttitudeStyle)
      case('binnedset')
        BinnedAttSetTabName = stringParameter('binnedattset')
        AvgCntToProcess = realParameter('avcnttoprocess')

      case('template')
      case default
        call error('badAttitudeStyle',"")
      end select

    case('raw')
      ! Do nothing.
    case default
      call error('badOutputStyle',"Value of --outputstyle must be either sky or raw.")
    end select

    WithRateImage   = booleanParameter('withrateimage')
    if (WithRateImage) then
      RateSetRoot     = stringParameter('ratesetroot')
    end if
    WithPileupMask  = booleanParameter('withpileupmask')
    if (WithPileupMask) then
      PileupMaskSetRoot = stringParameter('pileupmasksetroot')
    end if

    ForOote         = booleanParameter('foroote')
    ForRgaSpike     = booleanParameter('forrgaspike')

    IsInverted      = booleanParameter('isinverted')

    WithApproxPatFrac = booleanParameter('withapproxpatfrac')
    if (.not. WithApproxPatFrac) then
      call warning('withapproxpatfracNotSupported' &
         , 'withapproxpatfrac is not yet supported and is forced to be TRUE.')
      WithApproxPatFrac = .true.
    end if
    applyAttcalcFudge = booleanParameter('withboresightfudge')	! Global variable defined and used in ssclib/coordinate

  end subroutine readParameters




  !---------------------------------------------------------------------
  subroutine testssclibraw2det()
  !---------------------------------------------------------------------

    implicit none 

    integer(int16), parameter :: xRawIn=2, yRawIn=2
    real(double),        parameter :: det_unit   = 0.05d0	! in emevents

    ! real(double) :: xCam, yCam, xRaw, yRaw, xChi, yChi, xDet, yDet
    real(double) :: xRaw, yRaw, xDet, yDet
    integer(int16), pointer :: xDetInt(:), yDetInt(:)
    integer(int16), pointer :: xRawInAry(:), yRawInAry(:)
    logical(bool) :: isOffChip
    integer(int8)  :: ccdNum		! for detToRaw() in ssclib/coordinate
    real(double) :: pi, conv, step, focalLength			! in emevents
    integer(int16) :: arx(1)
    integer(int16), pointer :: xInt16s(:), yInt16s(:)
    real(double),   pointer :: xDoubles(:), yDoubles(:), zDoubles(:)
    integer(int8), dimension(:), pointer          :: ccds

    !-------------------------------------------------------------------
    ! Start:

    allocate(xInt16s(1), yInt16s(1), xDoubles(1), yDoubles(1), zDoubles(1))
    arx = (/ int(2, kind=int16) /)
print *, "== Cal routines check == "
    call CAL_setState(EMOS1, ccdChipId=1, randomize=.false.)	! EMOS1 == 4
    call CAL_chipCoordToCamCoord1(arx, arx, xDoubles, yDoubles, zDoubles)
print *, " CamX/Y(chip1) = ", xDoubles(1), yDoubles(1), zDoubles(1)

    call CAL_setState(EMOS1, ccdChipId=2, randomize=.false.)	! EMOS1 == 4
    call CAL_chipCoordToCamCoord1(arx, arx, xDoubles, yDoubles, zDoubles)
print *, " CamX/Y(chip2) = ", xDoubles(1), yDoubles(1), zDoubles(1)

    call CAL_setState(EMOS1, ccdChipId=1, randomize=.false.)	! EMOS1 == 4
    call CAL_camCoord1ToChipCoord((/-11.937869d0/), (/-11.937869d0/), (/0.0d0/)&
                                      , xInt16s, yInt16s)
print *, " ChiX/Y(chip1) = ", xInt16s, yInt16s
    call CAL_camCoord1ToChipCoord((/-11.937869d0/), (/-11.937869d0/), (/0.0d0/)&
                                      , xInt16s, yInt16s, ccds)
print *, " ChiX/Y(ccd:1) = ", xInt16s, yInt16s, ccds

    call CAL_setState(EMOS1, ccdChipId=2, randomize=.false.)	! EMOS1 == 4
    call CAL_camCoord1ToChipCoord((/-11.937869d0/), (/-11.937869d0/), (/0.0d0/)&
                                      , xInt16s, yInt16s)
print *, " ChiX/Y(chip2) = ", xInt16s, yInt16s
    call CAL_camCoord1ToChipCoord((/-11.937869d0/), (/-11.937869d0/), (/0.0d0/)&
                                      , xInt16s, yInt16s, ccds)
print *, " ChiX/Y(ccd:2) = ", xInt16s, yInt16s, ccds

    !function ims_getFocalLength() result(focalLengthMm)
    call CAL_getMiscellaneousData('FOCAL_LENGTH', focalLength)
print *, "Focal length(just-CAL): ", focalLength

    !! ssclib/coordinate
focalLength = ims_getFocalLength()
print *, "Focal length: ", focalLength
    call CAL_setState(EMOS1, ccdChipId=1, randomize=.false.)	! EMOS1 == 4
    call rawToDet(xRawIn, yRawIn, xDet, yDet)
    call detToRaw(xDet, yDet, xRaw, yRaw, isOffChip, ccdNum)
print *,'Result1=',ccdNum,xRaw,yRaw,xDet,yDet

    !! emevents test
    allocate(xRawInAry(1), yRawInAry(1), xDetInt(1), yDetInt(1))
    xRawInAry(1)=xRawIn
    yRawInAry(1)=yRawIn
    !! in emevents
    focalLength = CAL_getMiscellaneousDataValue("FOCAL_LENGTH")
print *, "Focal length: ", focalLength
    pi = acos(0.d0) * 2.d0
    step = 3600.d0 / det_unit
    conv = step * 180.d0 / pi
    call getDetCoords(xRawInAry, yRawInAry, xDetInt, yDetInt, conv, focalLength)
print *,'Result2=',0,0,0,xDetInt(1),yDetInt(1)

  end subroutine testssclibraw2det



  !! from emevents
  subroutine getDetCoords(xin, yin, xout, yout, conv, focalLength)

! xin,yin   : RAWX/RAWY coordinates
! xout,yout : DETX/DETY coordinates
! conv      : conversion factor from radians to detector coords
! focalLength: Focal length of the telescope
! printon   : sets verbosity on

  integer(kind=int16), parameter :: det_offset = 0	! from emevents main
  integer(kind=int32), parameter :: detSysType = DETECTOR_OPTICSCENTER_SYSTEM
    integer(kind=int16), dimension(:), pointer :: xin, yin, xout, yout
    real(kind=double), intent(in)              :: conv, focalLength

    real(double), dimension(:), pointer :: xmilli, ymilli, zmilli

    call CAL_rawXY2mm(xin,yin,xmilli,ymilli,zmilli,detSysType)

! nint does not work with kind=int16 on NAGware
! Z axis of CAMCOORD2 points toward the mirror

    zmilli = focalLength - zmilli		! Masa: before, zmilli==0
    xmilli = det_offset + xmilli*conv/zmilli	! Masa: After, xmilli[ch]
    ymilli = det_offset + ymilli*conv/zmilli
    xout = int(xmilli+sign(0.5d0,xmilli), kind=int16)
    yout = int(ymilli+sign(0.5d0,ymilli), kind=int16)
    call CAL_releaseMemory(xmilli, ymilli, zmilli)

  end subroutine getDetCoords



  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module eootepileupmask_mod

