! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: eoptloadmask_mod
!
! Authors:  Masaaki Sakano - SSC/LUX
!          
! Description: This task reads external (optical) catalogues and generate a mask file that masks the areas, which are affected with optical loading.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eoptloadmask_mod

  use dal
  use param  
  use errorHandling
  use dataFilter        ! in selectlib.

  ! Followings in ssclib
  use array_utils
  use constants		! for DegToRad etc
  use coordinate	! for raDecToPixels()
  use dal_aux
  use geometric_types	! for type(Point2dT)
  use save_image
  use ssc_misc
  use wcs_aux

  implicit none 

  private 

  public eoptloadmask

  ! Command-line parameters:
  character(1024), allocatable ::&
    SelectionExpressions(:)
  character(256), allocatable :: OptSrcListSetTabNames(:)
  character(256) ::&
    MaskSetName='',&
    TemplateImageSetName='',&
    TemporaryImageSetName=''
    ! ColnameFile=''

    ! EventSetName='',&
  character(64) ::&
    RadiusStyle='',&
    ColnameStyle=''
  character(64), allocatable ::&
    ColNamesRa(:),&
    ColNamesDec(:),&
    ColNamesRadius(:)
  real(single) ::&
    OptLoadRadius=-99.0	! in arcsec
  logical ::&
    WithExpressions=.false.,& 
    IsInverted=.true.

  type :: SrcPosT
    real(double) ::&
      raDeg,&
      decDeg,&
      radiusArcsec
  end type SrcPosT

contains 

  !---------------------------------------------------------------------
  subroutine eoptloadmask
  !---------------------------------------------------------------------

    implicit none 

    integer :: numOptSrcList, numRows
    character(256), allocatable :: srcListSetNames(:), srcListTabNames(:)
    character(256) :: tempSrcSetName
    logical, allocatable  :: maskImage(:,:)
    integer(int8), allocatable  :: maskImageInt(:,:)
    integer, allocatable  :: numOptSrcRows(:), numOptSrcRowsFiltered(:)
    type(DataSetT) :: srcSet
    type(TableT)   :: srcTab
    type(Point2dT), allocatable :: indexForMaskXY(:,:)	! type def in ssclib/geometric_types, Real(single)
    type(WcsT) :: wcs		! type def in ssclib/wcs_aux
    type(SrcPosT), allocatable  :: srcPos(:,:)	! type def in this code.
    integer :: i, j
    integer(int32) :: numSrcTooFar
    integer(int16) :: xiSize, yiSize
    real(double) :: pix2arcsecX2, pix2arcsecY2
    real(single) :: xPixel, yPixel
    logical :: isExisting

    ! logical, parameter :: isTest = .true.

    !-------------------------------------------------------------------
    ! Start:

    call readParameters()
    tempSrcSetName = trim(TemporaryImageSetName)	! Use TemporaryImageSetName) for the Temporary source-list filename, as they are not used simultaneously.

    call message(AppMsg, Verbose, 'Checking the maximum number of rows.')

    numOptSrcList = size(OptSrcListSetTabNames)
    allocate(numOptSrcRows(numOptSrcList), numOptSrcRowsFiltered(numOptSrcList)  &
       , srcListSetNames(numOptSrcList), srcListTabNames(numOptSrcList))
    numOptSrcRows(:) = -99

    do i=1, numOptSrcList
      write(errStr, '(a)') " Checking the number of rows in "//trim(OptSrcListSetTabNames(i)) 
      call message(AppMsg, Noisy)
      call splitSetTabName(trim(OptSrcListSetTabNames(i)), srcListSetNames(i), srcListTabNames(i))	! in ssclib/dal_aux
      srcSet = dataSet(srcListSetNames(i), READ)
      srcTab = table(srcSet, srcListTabNames(i), READ)
      numOptSrcRows(i) = numberOfRows(srcTab)
      write(errStr, '(a, i8)') " The number of rows = ", numOptSrcRows(i)
      call message(AppMsg, Noisy)
      call release(srcSet)
    end do

    allocate(srcPos(numOptSrcList, maxval(numOptSrcRows)))	! can accommodate all the rows in the input source list.
    write(errStr, '(a, i8)') " The maximum number of rows = ", maxval(numOptSrcRows)
    call message(AppMsg, Noisy)

    call message(AppMsg, Verbose, 'Filtering source lists.')

    do i=1, numOptSrcList
      call message(AppMsg, Verbose, 'Processing (' &
         //trim(srcListSetNames(i))//':'//trim(srcListTabNames(i))//')...')
      call message(AppMsg, Noisy, 'Creating temporary source list set ('//trim(tempSrcSetName)//').')
      srcSet = clone(srcListSetNames(i), tempSrcSetName, TEMP)
      srcTab = table(srcSet, srcListTabNames(i), MODIFY)

      numRows = numberOfRows(srcTab)
      write(errStr, '(a, i8)') " Original NumberOfRows = ", numRows
      call message(AppMsg, Verbose)
      if (trim(SelectionExpressions(i)) == '') then
        numOptSrcRowsFiltered(i) = numRows
        call message(AppMsg, Verbose, ' No filtering is applied.')
      else
        call message(AppMsg, Noisy, ' Applying filtering condition: "'//trim(SelectionExpressions(i))//'"')
        call initFilter(SelectionExpressions(i))
        numOptSrcRowsFiltered(i) = doFilter(srcTab)
        write(errStr, '(a, i8)') " Filtered NumberOfRows = ", numOptSrcRowsFiltered(i)
        call message(AppMsg, Verbose)
      end if

      if (numOptSrcRowsFiltered(i) <= 0) then
        call warning('noSourcesSelected'&
           , 'No sources passed the specified filtering expression.  No output is made.')
        call release(srcSet)
        deallocate(maskImage, maskImageInt)	! allocatable
        deallocate(OptSrcListSetTabNames, ColNamesRa, ColNamesDec, ColNamesRadius, SelectionExpressions)	! Global allocatable

        ! Checking if the output file already exists.
        ! To avoid the possibility of abnormal exit in this condition,
        ! make sure the output maskfile does not exist priot to run this task.
        inquire(file=trim(MaskSetName), exist=isExisting)
        if (isExisting) call error('noOutputMade'&
           , 'The specified mask file already exists, but is not overwritten.')
        return
      end if

      ! call initializeSrcDataT(srcPos(i,:)) ! in ssclib/source_cutouts
      ! srcPos(i,:)%isValid = .true. ! default
      ! srcPos(i,:)%region%unit = 'arcsec'	! in ssclib/ellipse
      ! srcPos(i,:)%region%isCircle = .true.

      call message(AppMsg, Noisy, 'Reading ('//trim(ColNamesRa(i))//')...')
      ! call readColDataToFixed(srcTab, trim(ColNamesRa(i)),  srcPos(i,:)%raDeg)	! => causes an error of mismatch of the Array size if selectionExpression is applied and valid.
      call readColDataToFixed(srcTab, trim(ColNamesRa(i)),  srcPos(i,1:numOptSrcRowsFiltered(i))%raDeg)
      call message(AppMsg, Noisy, 'Reading ('//trim(ColNamesDec(i))//')...')
      call readColDataToFixed(srcTab, trim(ColNamesDec(i)), srcPos(i,1:numOptSrcRowsFiltered(i))%decDeg)

      if (trim(RadiusStyle) .ne. 'column') then
      ! if (trim(ColNamesRadius(i)) == '') then
        srcPos(i,:)%radiusArcsec = OptLoadRadius
      else
        call message(AppMsg, Noisy, 'Reading ('//trim(ColNamesRadius(i))//')...')
        call readColDataToFixed(srcTab, trim(ColNamesRadius(i)), srcPos(i,1:numOptSrcRowsFiltered(i))%radiusArcsec)
      end if
      ! ellipse(i,:)%shortSemiAxis = ellipse(i,:)%longSemiAxis

      call release(srcSet)
    end do	! do i=1, numOptSrcList
! print *,'radius=',srcPos(:,:)%radiusArcsec

    ! --------------- Read the template image ---------------

    call message(appMsg, Verbose &
       , 'Reading the WCS information from the template dataset (' &
       //trim(TemplateImageSetName)//').')

    call getArraySize(TemplateImageSetName, xiSize, yiSize)	! ssclib/dal_aux.f90
    call readWcs(TemplateImageSetName, wcs)	! in ssclib/wcs_aux.
    call dumpWcs(wcs, Noisy)

!     imgSet = dataSet(TemplateImageSetName, READ)
!     imArray = array(imgSet, 0)
! integer(int32), pointer :: dimVector(:)
!     dimVector => dimensions(inArray)
!     select case(size(dimVector))
!     case(2)
!       xiSize = dimVector(0)
!       yiSize = dimVector(1)
!     case default 
!       call error('wrongArrayRank', 'The input template image does not look an image.')
!     end select
!     call readWcs(imgSet, imArray, wcs) ! in ssclib/wcs_aux.
!     call release(imgSet)

    ! --------------- MAIN ---------------

    allocate(maskImage(xiSize, yiSize), maskImageInt(xiSize, yiSize) &
           , indexForMaskXY(xiSize, yiSize))
           !, indexForMaskX(xiSize, yiSize), indexForMaskY(xiSize, yiSize))

    do j=1, yiSize
      indexForMaskXY(:,j)%x = (/ (i, i=1,xiSize) /)
    end do
    do i=1, xiSize
      indexForMaskXY(i,:)%y = (/ (j, j=1,yiSize) /)
    end do

    maskImage = .false.
    pix2arcsecX2 = (wcs%x%matrixOnAxis * 3600)**2	! [arcsec/pix]
    pix2arcsecY2 = (wcs%y%matrixOnAxis * 3600)**2	! [arcsec/pix]
    do i=1, numOptSrcList		! Loop over files
      numSrcTooFar = 0
      do j=1, numOptSrcRowsFiltered(i)	! Loop over the sources in each file

        if (angleBetweenCelCoords((/ srcPos(i,j)%raDeg, srcPos(i,j)%decDeg /) &
             , (/ wcs%x%refWorld, wcs%y%refWorld /), isRadian=.false.) >= 90.0) then
          numSrcTooFar = numSrcTooFar + 1
          cycle
        end if

        call raDecToPixels(srcPos(i,j)%raDeg*DegToRad &
                         , srcPos(i,j)%decDeg*DegToRad &
                         , wcs, xPixel, yPixel)	! defined in ssclib/coordinate
        where ((indexForMaskXY%x-xPixel)**2 * pix2arcsecX2 &
             + (indexForMaskXY%y-yPixel)**2 * pix2arcsecY2 &
             < srcPos(i,j)%radiusArcsec**2)
          ! NOTE:
          ! This will check the entire area in the mask for each source.
          ! It is possible to limit the area of search by hand-writing
          ! the loop structure rather than using `where' statement,
          ! and it may be faster, given the affected area per each source
          ! in the realistic situations are likely to be very small,
          ! compared with the entire area (though the hand-written Fortran
          ! routine would not look pretty :p).
          maskImage = .true.
        end where

      end do	! do j=1, numOptSrcRowsFiltered(i)	! Loop over the sources in each file

      if (numSrcTooFar > 0) then
        if (numSrcTooFar == numOptSrcRowsFiltered(i)) then
          call warning('allSourcesTooFar'&
             , 'All the sources in ('//trim(OptSrcListSetTabNames(i)) &
             //') are more than 90 degrees away from the frame and hence ignored.')
        else
          write(errStr, '(a,i8,a,a,a)') &
             "Some sources (Number: ", numSrcTooFar, ") in (" &
             , trim(OptSrcListSetTabNames(i)) &
             , ") are more than 90 degree away from the frame and hence ignored."
          call message(appMsg, Verbose)
        end if
      end if
    end do	! do i=1, numOptSrcList

    deallocate(indexForMaskXY)


    ! --------------- Output ---------------

    if (IsInverted) then
      call invertMask(maskImage)	! ssclib/array_utils
    end if

    maskImageInt = 0
    where(maskImage)
      maskImageInt = 1
    end where

    call message(AppMsg, Verbose, 'Writing mask image.')
    call saveImage(maskImageInt, TemplateImageSetName, MaskSetName&
       , TemporaryImageSetName, wcs=wcs) ! in ssclib/save_image.

    deallocate(maskImage, maskImageInt)	! allocatable
    deallocate(OptSrcListSetTabNames, ColNamesRa, ColNamesDec, ColNamesRadius, SelectionExpressions)	! Global allocatable

  end subroutine eoptloadmask


  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    integer :: i, numImages, numRaDec, numElements

    !-------------------------------------------------------------------
    ! Start:

    numImages = parameterCount('srclisttab')
    if (numImages<=0) call error('noInputCatalogue',"")
    allocate(OptSrcListSetTabNames(numImages) &
       , ColNamesRa(numImages), ColNamesDec(numImages), ColNamesRadius(numImages) &
       , SelectionExpressions(numImages))
    OptSrcListSetTabNames(:) = ''
    SelectionExpressions(:) = ''

    do i = 1, numImages
      OptSrcListSetTabNames(i) = stringParameter('srclisttab', i-1)
    end do

    MaskSetName           = stringParameter('maskset')
    TemplateImageSetName  = stringParameter('templateset')
    TemporaryImageSetName = stringParameter('temporaryset')

    RadiusStyle = stringParameter('radiusstyle')
    select case(trim(RadiusStyle))
    case('fixed')
      OptLoadRadius = realParameter('radius')
    case('column')
      ColNamesRadius(:) = ''
      ! ColnameRadius = stringParameter('colradius')
      numRaDec = parameterCount('colradius')
      do i = 1, numRaDec
        ColNamesRadius(i) = stringParameter('colradius', i-1)
      end do
      ! If only a single element is given, that is applied to the rest, too.
      if (numRaDec == 1 .and. numImages > 1) then
        call message(AppMsg, Verbose &
           , 'Only 1 element "'//trim(ColNamesRadius(1)) &
           //'" for colradius is specified for more than 1 source list.' &
           //'  It is applied to all the input source lists.')
        ColNamesRadius(:) = trim(ColNamesRadius(1))
      end if

    case default
      call error('badRadiusStyle',"The value of the --radiusstyle parameter &
         &is not recognized.")
    end select		! case(trim(RadiusStyle))

    ColnameStyle         = stringParameter('colnamestyle')
    ColNamesRa(:)  = 'RA'
    ColNamesDec(:) = 'DEC'
    select case(trim(ColnameStyle))
    case('default')
      ! Do nothing.

    case('user')
      numRaDec = parameterCount('colra')
      do i = 1, numRaDec
      ! do i = 1, max(numImages, numRaDec)
        ColNamesRa(i)  = stringParameter('colra', i-1)
      end do
      ! If only a single element is given, that is applied to the rest, too.
      if (numRaDec == 1 .and. numImages > 1) then
        call message(AppMsg, Verbose &
           , 'Only 1 element "'//trim(ColNamesRa(1)) &
           //'" for colra is specified for more than 1 source list.' &
           //'  It is applied to all the input source lists.')
        ColNamesRa(:) = trim(ColNamesRa(1))
      end if

      numRaDec = parameterCount('coldec')
      do i = 1, numRaDec
        ColNamesDec(i) = stringParameter('coldec', i-1)
      end do
      ! If only a single element is given, that is applied to the rest, too.
      if (numRaDec == 1 .and. numImages > 1) then
        call message(AppMsg, Verbose &
           , 'Only 1 element "'//trim(ColNamesDec(1)) &
           //'" for coldec is specified for more than 1 source list.' &
           //'  It is applied to all the input source lists.')
        ColNamesDec(:) = trim(ColNamesDec(1))
      end if

    ! case('file')
    !   ColnameFile = stringParameter('colnamefile')
    !   if (trim(RadiusStyle) == 'column') then
    !     ColNamesRadius(:) = 'RADIUS'	! Default
    !     call readColnamesRaDec(trim(ColnameFile), ColNamesRa, ColNamesDec, ColNamesRadius)
    !   else
    !     call readColnamesRaDec(trim(ColnameFile), ColNamesRa, ColNamesDec)
    !   end if

    case default
      call error('badColnameStyle',"The value of the --colnamestyle parameter &
         &is not recognized.")
    end select		! case(trim(ColnameStyle))

    WithExpressions = booleanParameter('withexpressions')
    if (WithExpressions) then
      numElements = parameterCount('expressions')
      SelectionExpressions(:) = ''
      do i = 1, numElements
! print *,'readPara: i=',i,' numElements=',numElements,' numImages=',numImages
        SelectionExpressions(i) = stringParameter('expressions', i-1)
! print *,'SelectionExpressions(i)=('//trim(SelectionExpressions(i))//')'
      end do

      ! If SelectionExpressions(:) has only a single element, that expression is applied to all the files.
      if (numElements == 1 .and. numImages > 1) then
        call message(AppMsg, Noisy &
           , 'Only 1 expressions "'//trim(SelectionExpressions(1)) &
           //'" is specified for more than 1 source list.' &
           //'  It is applied to all the input source lists.')
        SelectionExpressions(:) = trim(SelectionExpressions(1))
        ! do i = 2, numImages
        !   if (trim(SelectionExpressions(i)) == '') then
        !     SelectionExpressions(i) = trim(SelectionExpressions(1))
        !   end if
        ! end do
      end if
    end if

    IsInverted = booleanParameter('isinverted')

  end subroutine readParameters


  !---------------------------------------------------------------------
  subroutine readColnamesRaDec(inFilename, colRaStr, colDecStr, colRadiusStr)
  !---------------------------------------------------------------------
  ! For colnamestyle=file.
  ! The option is taken out, so this subroutine is not used.

    implicit none 

    character(*), intent(in) :: inFilename
    character(*), intent(out) :: colRaStr(:), colDecStr(size(colRaStr))
    character(*), intent(out), optional :: colRadiusStr(size(colRaStr))

    character(2048) :: inStr
    character(64), pointer :: tmpStrAry(:)
    integer :: nunit, ioStatus, iFile
    logical :: doesExist

    !-------------------------------------------------------------------
    ! Start:

    call getFreeIoUnit(nunit)	! defined in ssc_misc

    inquire(file=trim(inFilename), exist=doesExist)
    if (.not. doesExist) then
      call error('notExists', 'File=('//trim(inFilename)//')')
    end if

    call message(AppMsg, Verbose, 'Reading File ('//trim(inFilename)//')')
    open(nunit)
    iFile = 1
    do 
      read(nunit, '(a)', iostat=ioStatus) inStr
      if (ioStatus < 0) then	! EOF
        exit
      else if (ioStatus > 0) then
        call error('IOError', 'while reading File=('//trim(inFilename)//')')
      end if

      call splitStr(inStr, tmpStrAry)	! in ssclib/ssc_misc
      if (size(tmpStrAry) <= 0) then	! Empty line
        deallocate(tmpStrAry)
        cycle
      else if (tmpStrAry(1)(1:1) == '#') then	! Comment line
        deallocate(tmpStrAry)
        cycle
      end if
      
      if (iFile > size(colRaStr)) then
        write(errStr, '(a,i2,a,i2,").",a)') &
           "readColnamesRaDec: Number of valid lines in the given ASCII file is too large (" &
           , iFile &
           , "), as opposed to the number of given catalogue files (" &
           , size(colRaStr)
        call error('inconsistentAsciiInputFile')
      end if
      colRaStr(iFile)  = trim(tmpStrAry(1))
      colDecStr(iFile) = trim(tmpStrAry(2))
      if (present(colRadiusStr)) then
        colRadiusStr(iFile)= trim(tmpStrAry(3))
      end if
      deallocate(tmpStrAry)

      write(errStr, '("(Effective) Line no.(",i2,") is read.")') iFile 
      call message(AppMsg, Noisy)
      iFile = iFile + 1
    end do
    close(nunit)

    call message(AppMsg, Verbose, 'Successfully read File ('//trim(inFilename)//')')

  end subroutine readColnamesRaDec


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module eoptloadmask_mod

