! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: epileupmask_mod
!
! Authors:  Ian Stewart - SSC/LUX
!        :  Masaaki Sakano - SSC/LUX
!          
! Description: This task calculates a single-pixel event count rate per frame for each pixel, normalised to 1.49 keV Al-K line based on the grade branching ratio, which is a function of energy, and outputs a mask cube or image indicating those pixels where the normalised count rate exceeds a set limit, as well as rate cube/image if specified so.
! 
! Comments: 
! 
! ToDo lists: 
!   * write writeCubeDataDouble writeCubeDataInt8 etc in eimchip2sky/src/cube_aux
!   * write generic coordinate transform tool
!   * write dumpArray in ssclib/dump
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module epileupmask_mod

  use dal
  use param  
  use errorHandling
  use cal
  ! use CalTypes	! used in epileupmask_shared

  use binned_att	! in attcalc2
  use cube_aux		! in eimchip2sky_aux
  ! use eimchip2sky_aux	! in eimchip2sky (used in epileupmask_shared)

  ! Followings in ssclib
  use array_utils
  use cal_aux
  use coordinate	! needed for applyAttcalcFudge
  use dal_aux
  ! use intervals_aux	! for gtiVecType
  use save_image
  use ssc_misc
  use wcs_aux		! in ssclib

  use epileupmask_shared
  ! use ebkgmap_skyframe	! called from epileupmask_shared
  ! use ebkgmap_attcommon	! called from epileupmask_shared

  implicit none 

  private 

  public epileupmask

  ! Command-line parameters:
  character(256) ::&
    InputStyle='',&
    EventSetName='',&
    ImageSetName='',&
    OutputStyle='',&
    TemplateImageSetName='',&
    MaskSetName='',&
    AttitudeStyle='',&
    BinnedAttSetTabName='',&
    RateSetRoot=''
  real(single) ::&
    RateThreshold=0.0,&
    EvLow=0.0,&
    EvHigh=0.0,&
    AvgCntToProcess=0.0
  logical ::&
    IsSingleEventOnly = .false.,&
    WithRateImage = .false.,&
    WithApproxPatFrac = .true.

contains 

  !---------------------------------------------------------------------
  subroutine epileupmask
  !---------------------------------------------------------------------

    implicit none 

    character(256), parameter :: temporaryImageFileName = 'epileupmask_tmpimage0001.ds'

    type(DataSetT) :: inputSet
    character(20) :: messageInputFile
    ! integer(int32) :: instrumentId, numCubePlanes
    integer(int32) :: minFrameToProcess
    ! integer :: maxCcdNodeNum
    ! integer, pointer :: cubePlaneIndices(:) => null()
    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT) :: infoFrameAtt	! Defined in epileupmask_shared
    ! integer(int16) :: chipXSize, chipYSize
    integer(int8) :: numCcds, numNodes !, ccdNum , nodeNum
    ! type(keyInfoType), pointer :: rawKeyInfo(:) => null()	! type def in eimchip2sky/cube_aux.
    real(single), pointer :: rateCube(:,:,:) => null()
    logical, pointer :: maskCube(:,:,:)
    ! real(single), allocatable :: puProbCube(:,:,:)
    real(single), pointer :: maskImageRealTmp(:,:) => null() &
       , rateImages(:,:,:) => null(), rateImageTmp(:,:) => null()
	!, maskImageReal(:,:)&
       ! , numFrames(:) => null(), numTotFrames(:) => null()
         ! The mask defined here is real-valued, which makes it easier
         ! to rebin it to sky coordinates (if specified so).
    ! real(single) :: maxNumFrames
    integer(int8), pointer :: maskImageInt(:,:) => null(), maskImageIntTmp(:,:) => null()
    integer :: xiSize, yiSize, numProcessedFrames
    ! integer(int16) :: tlmin, tlmax
    ! real(kind=double), dimension(:), pointer :: pattFracSingle => null()	! grade pattern
    character(256) :: outputRefSetName

    !! For attitude conversion.
    ! type(gtiVecType), pointer :: gtiVec(:)	! type def in ssclib/intervals_aux.
    ! real(single), allocatable :: binnedAttWeights(:,:)
    ! type(attBinType), pointer :: binnedScAtt(:) => null()	! type def is in attcalc2/binned_att.
    ! real(single),     pointer :: binnedAttsNumFrames(:,:) => null()
    ! real(double) :: expStartTimeStamp
    ! type(SpacecraftAttitudeType) :: medianScAttitude
    type(WcsT) :: wcs	! type def in ssclib/wcs_aux
    integer :: numAttBins, attBinNum

    type(TableT) :: eventTab
    type(evlistColumns) :: evCol	! typedef in epileupmask_shared
    real(double) :: startTime, finisTime
    logical :: isCycle, isExisting
    character(280) :: tmpFilename

    !-------------------------------------------------------------------
    ! Start:

    call readParameters
    minFrameToProcess = ceiling(AvgCntToProcess/RateThreshold)

    ! Setting up inputSet etc.
    !   Get outputRefSetName, messageInputFile, inputSet
    !    and possibly rateImages(1,:,:) if InputStyle=='image'
    call initialSetUpForEpileup(InputStyle, EventSetName, ImageSetName&
       , outputRefSetName, inputSet, messageInputFile, rateImageTmp)

    ! Setting up for instument-related stuffs.
    !   Get instrumentId, chip(X|Y)Size, numCcds, numNodes
    call instSetUpForEpileup(inputSet, infoCcdChips, withTlminmax=.false.)
    ! instrumentId = infoCcdChips%instrumentId
    ! chipXSize    = infoCcdChips%chipXSize
    ! chipYSize    = infoCcdChips%chipYSize
    numCcds      = infoCcdChips%numCcds
    numNodes     = infoCcdChips%numNodes
    ! tlmin        = infoCcdChips%tlmin 
    ! tlmax        = infoCcdChips%tlmax
    ! call instSetUpForEpileup(inputSet, instrumentId, chipXSize, chipYSize&
    !    , numCcds, numNodes, tlmin, tlmax)

    !...................................................................
    ! Get Attitude and Sky coordinates frame infomation.
    if ((OutputStyle .eq. 'sky') .or. (InputStyle .eq. 'image')) then

      !   Get wcs, (x|y)iSize, binnedScAtt(:), expStartTimeStamp, medianScAttitude
      call getAttSkyCoordForEpileup(TemplateImageSetName, AttitudeStyle, BinnedAttSetTabName&
         , wcs, xiSize, yiSize&
         , infoFrameAtt )
      ! binnedScAtt => infoFrameAtt%binnedScAtt
      numAttBins = size(infoFrameAtt%binnedScAtt)
      ! expStartTimeStamp = infoFrameAtt%expStartTimeStamp
      ! medianScAttitude  = infoFrameAtt%medianScAttitude
      ! call getAttSkyCoordForEpileup(TemplateImageSetName, AttitudeStyle, BinnedAttSetTabName&
      !    , wcs, xiSize, yiSize&
      !    , binnedScAtt & !, numAttBins
      !    , expStartTimeStamp, medianScAttitude)
      ! numAttBins = size(binnedScAtt)

!*** NOTE! the only things that the template image supplies are (i) wcs, (ii) image pixel dimensions, (iii) median attitude and (iv) exposure start time stamp.  It would be nice to have some facility to input the last of these in other ways.

      if (OutputStyle .eq. 'sky') then
        allocate(maskImageInt(xiSize,yiSize)&	!, maskImageReal(xiSize,yiSize)
           , maskImageIntTmp(xiSize,yiSize), maskImageRealTmp(xiSize,yiSize))
        ! maskImageReal = 0
        maskImageInt = 0
        maskImageIntTmp  = 0
        maskImageRealTmp = 0.0
      end if
      if (InputStyle .eq. 'image') then
        ! rateImageTmp is() already set in initialSetUpForEpileup() in epileupmask_shared
        numAttBins = 1	! No separate attitude bin is used.
        allocate(rateImages(1,size(rateImageTmp,1),size(rateImageTmp,2)))	! No attitude bin is considered.
        rateImages(1,:,:) = rateImageTmp
        deallocate(rateImageTmp)
      else	! namely, if (InputStyle .ne. 'image') then
        allocate(rateImageTmp(xiSize,yiSize) &
               , rateImages(numAttBins,xiSize,yiSize))
        rateImages = 0
      end if
    end if	! if ((OutputStyle .eq. 'sky') .or. (InputStyle .eq. 'image')) then

    !...................................................................
    ! Count number of frames of each chip.
    !   Get rawKeyInfo(:), numTotFrames(:), cubePlaneIndices(:), numCubePlanes,
    !       binnedAttsNumFrames(1:numCubePlanes, 1:numAttBins)
    !  See comments in countNumberOfFrames() in epileupmask_shared.f90 for detail.
    call countNumberOfFrames(inputSet, InputStyle, OutputStyle&
     , infoCcdChips, infoFrameAtt &
     , messageInputFile )
    ! call countNumberOfFrames(inputSet, InputStyle, OutputStyle&
    !    , instrumentId, numCcds, numNodes, binnedScAtt & !, numAttBins
    !    , messageInputFile&
    !    , rawKeyInfo, numTotFrames, cubePlaneIndices&
    !    , numCubePlanes & ! , maxCcdNodeNum
    !    , binnedAttsNumFrames)	! in epileupmask_shared.f90
    !    ! binnedScAtt may be left undefined, depending on Input/OutputStyle.
    ! rawKeyInfo    => infoCcdChips%rawKeyInfo
    ! numTotFrames  => infoFrameAtt%numTotFrames
    ! cubePlaneIndices => infoCcdChips%cubePlaneIndices
    ! numCubePlanes =  infoCcdChips%numCubePlanes
    ! if (infoFrameAtt%isAllocatedBinnedAttsNumFrames) &
    !    binnedAttsNumFrames => infoFrameAtt%binnedAttsNumFrames

    ! Get the Array of fraction of GRADE_SINGLE events as a function of PI.
    !   Get pattFracSingle
    call getPatSingleFractionArrayTmpl(infoCcdChips)
    ! pattFracSingle => infoCcdChips%pattFracSingle
    ! call getPatSingleFractionArrayTmpl(chipXSize, chipYSize, pattFracSingle)
    !    , tlmax=tlmax)


    !...................................................................
    ! MAIN: Now calculate rateImages/rateCube, maskImage/maskCube

    call message(appMsg, Verbose, 'Reading data from the '&
         //trim(messageInputFile)//'.')

    ! Get evCol, eventTab (as well as some checks).
    !   n.b. The latter is required just to release() later on.
    call setupToReadDataForEpileup(InputStyle, OutputStyle&
       , inputSet, infoCcdChips &
       , evCol, eventTab)
    ! call setupToReadDataForEpileup(InputStyle, OutputStyle&
    !    ! , inputSet, chipXSize, chipYSize, maxCcdNodeNum&
    !    , inputSet, chipXSize, chipYSize & !, size(cubePlaneIndices)&
    !    , cubePlaneIndices& 
    !    , evCol, eventTab)

    !...................................................................
    ! MAIN select case:
    ! numFrames => numTotFrames
    select case(InputStyle)
    case('evlist')

      select case (OutputStyle)
      case('sky')
        rateImages = 0.0
        maskImageInt = 0
        ! maskImageReal = 0.0
        write(errStr, '(a, i8, a)') 'Will skip any attitude bin, containing less than'&
           , minFrameToProcess, ' frames.'
        call message(AppMsg, Verbose)

        numProcessedFrames = 0
        do attBinNum = 1, numAttBins

          call calcStartFinisTime(infoFrameAtt, attBinNum &
             , evCol, minFrameToProcess &
             , startTime, finisTime, isCycle)
          if (isCycle) cycle

!           numFrames => binnedAttsNumFrames(1:numCubePlanes,attBinNum)
!           ! numFrames => binnedAttsNumFrames(:,attBinNum)
!           maxNumFrames = maxval(numFrames)
! 
!           ! Calculate conservative start and end times for the applicability of each attitude sample. Note that it isn't necessary to consider the affect of GTIs on this period because we assume that the event list contains only events within these GTIs. These crudely-calculated start and stop times will therefore net the same number of events as a more sophisticated calculation would.
!           ! Note `crudely' means GTI is not taken into account below.
! 
!           if (attBinNum == 1) then
!              startTime = minval(evCol%times)
!           else
!              startTime = binnedScAtt(attBinNum)%binStartTime
!           end if
! 
!           if (attBinNum == numAttBins) then
!              finisTime = maxval(evCol%times)
!           else
!              finisTime = binnedScAtt(attBinNum+1)%binStartTime
!           end if
! 
!           if (maxNumFrames >= minFrameToProcess) then
!             write(errStr, '(a, i4, a, i4, a, f10.1, ".")') 'Processing the attitude bin'&
!                , attBinNum, ' out of', numAttBins&
!                , ', which has the maximum number of frames of', maxNumFrames
!             call message(AppMsg, Verbose)
!             write(errStr, '(a, es21.15, ".")') ' which starts roughly at ',startTime
!             call message(AppMsg, Noisy)
!           else
!             write(errStr, '(a, i4, a, f9.1)') 'Skip the attitude bin', attBinNum&
!                , ' due to the small number of frames with the maximum=', maxNumFrames
!             call message(AppMsg, Verbose)
!             write(errStr, '(a,es21.15," - ",es21.15,".")') ' which starts/ends roughly at ' &
!                , startTime, finisTime
!             call message(AppMsg, Noisy)
!             cycle	! Skip this Frame.  Note outputDiagn(attBinNum) is not allocated/initialised.
!           end if

          numProcessedFrames = numProcessedFrames + 1

          rateImageTmp = 0
          call readEvent2SkyForEpileup(RateThreshold&
             , infoCcdChips, infoFrameAtt &
             , rateImageTmp, maskImageRealTmp&
             , evCol, attBinNum &
             , wcs, WithRateImage&
             , startTime=startTime, finisTime=finisTime)	! , rateCube)
             ! Get rateImageTmp, maskImageRealTmp
          ! call readEvent2SkyForEpileup(RateThreshold&
          !    ! , WithApproxPatFrac&
          !    , chipXSize, chipYSize&
          !    , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
          !    , numFrames, rateImageTmp, maskImageRealTmp&
          !    , evCol, expStartTimeStamp&
          !    , rawKeyInfo, wcs, binnedScAtt(attBinNum)%att, WithRateImage&
          !    , startTime, finisTime)	! , rateCube)
          !    ! Get rateImageTmp, maskImageRealTmp

          ! Now, how shold we calculate the output logical mask?
          ! If we want the mask based on the averaged rate value, then
          !    maskImageReal = maskImageReal + maskImageRealTmp
          ! here (ie., for each attitude bin(=interval)) and 
          ! calculating the mask file based on the final maskImageReal
          ! will do the job.  However, if a pixel at a certain time interval
          ! exceeds the given threshold, namely piled-up, a result containing
          ! that pixel must be unreliable in a conservative approach.
          ! Indeed if the attitude is fluctuated enough and if a source
          ! flux exceeds just above the threshold, then the resultant
          ! `averaged' rate image may not give any masked pixel,
          ! which is undesirable.  Therefore, the latter conservative approach
          ! is accepted here.  The task will mask any pixel
          ! that exceeds the threshold at any time interval.
          maskImageIntTmp = 0
          where(maskImageRealTmp > 0.3) maskImageIntTmp = 1
           ! Here, if the threshold is 0, instead of 0.3, it will give the mask with a bit wider coverage.
          maskImageInt = maskImageInt + maskImageIntTmp

          rateImages(attBinNum,:,:) = rateImageTmp
          ! if (WithRateImage) rateImage = rateImage + rateImageTmp	! Should be outside of the loop, ideally..

          if (verbosity(AppMsg) >= Verbose) then			! Should be outside of the loop, ideally..
            if (any(maskImageIntTmp > 0)) then
              write(errStr, '(a, i4, ".")') &
                 "Found piled-up pixel(s) in the attitude bin", attBinNum 
              ! call message(appMsg, Verbose, 'Found piled-up pixel(s).')
            else
              write(errStr, '(a, i4, ".")') &
                 "Found NO piled-up pixel in the attitude bin", attBinNum 
              ! call message(appMsg, Verbose, 'Found NO piled-up pixel.')
            end if
            call message(appMsg, Verbose)
          end if
        end do	! do attBinNum = 1, numAttBins ! for case('evlist'),OutputStyle==case('sky')

        deallocate(maskImageRealTmp, maskImageIntTmp, rateImageTmp)

        if (numProcessedFrames < 1) then
          deallocate(maskImageInt)
          call warning('noFrameProcessed' &
             , 'No frame has been processed, hence no output is created.')

          ! Final clean up before the end of this task.
          call deallocateInfoCcdChipsT(infoCcdChips)
          call deallocateInfoFrameAttT(infoFrameAtt)

          inquire(file=trim(MaskSetName), exist=isExisting)
          if (isExisting) call error('noOutputMade'&
             , 'The specified mask file ('//trim(MaskSetName)&
             //') already exists, but is not overwritten.')
          return	!! END of the task!
        end if

        !! If you want to calculate the output mask based on the average, this is the way...
        ! maskImageInt = 0
        ! where(maskImageReal > 0.3) maskImageInt = 1

        where(maskImageInt > 1) maskImageInt = 1	! (... > 0) is OK, too.

      case('raw')
          call readEvent2RawForEpileup(RateThreshold&
             , infoCcdChips, infoFrameAtt &
             , rateCube, maskCube&
             , evCol)
          ! call readEvent2RawForEpileup(RateThreshold&
          !    ! , WithApproxPatFrac&
          !    , chipXSize, chipYSize&
          !    , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
          !    , numFrames, rateCube, maskCube&
          !    , evCol)

      case default
         call error('badOutputStyleB4',"Should not happen.  Contact the code developer.")
      end select	! select case (styleOut) in case(styleIn)=('image')

      call release(eventTab)

    case('image')	! select case(styleIn)

      allocate(rateImageTmp(size(rateImages,2), size(rateImages,3)))
      rateImageTmp = rateImages(1,:,:)

      select case (OutputStyle)
      case('sky')
        call readImage2SkyForEpileup(RateThreshold&
           , IsSingleEventOnly, WithApproxPatFrac&
           , xiSize, yiSize&
           , infoCcdChips, infoFrameAtt &
           , rateImageTmp, maskImageInt&
           , wcs &
           , (EvLow+EvHigh)*500.0)	! The Last-argument == E/2*1000 (eV)
        ! call readImage2SkyForEpileup(RateThreshold&
        !    , IsSingleEventOnly, WithApproxPatFrac&
        !    , xiSize, yiSize&
        !    , instrumentId, cubePlaneIndices, pattFracSingle&
        !    ! , numFrames, rateImages(1,:,:), maskImageInt&
        !    , numFrames, rateImageTmp, maskImageInt&
        !    , expStartTimeStamp, wcs, medianScAttitude&
        !    , (EvLow+EvHigh)*500.0)	! The Last-argument == E/2*1000 (eV)
        rateImages(1,:,:) = rateImageTmp
        deallocate(rateImageTmp)

      case('raw')
        allocate(rateCube(infoCcdChips%numCubePlanes, infoCcdChips%chipXSize, infoCcdChips%chipYSize) &
               , maskCube(infoCcdChips%numCubePlanes, infoCcdChips%chipXSize, infoCcdChips%chipYSize))
        call readImage2RawForEpileup(RateThreshold&
           , IsSingleEventOnly, WithApproxPatFrac&
           , xiSize, yiSize&
           , infoCcdChips, infoFrameAtt &
           , rateImageTmp, rateCube, maskCube&
           , wcs &
           , (EvLow+EvHigh)*500.0)	! The Last-argument == E/2*1000 (eV)
        ! call readImage2RawForEpileup(RateThreshold&
        !    , IsSingleEventOnly, WithApproxPatFrac&
        !    , xiSize, yiSize&
        !    , instrumentId, cubePlaneIndices, pattFracSingle&
        !    ! , numFrames, rateImages(1,:,:), rateCube, maskCube&
        !    , numFrames, rateImageTmp, rateCube, maskCube&
        !    , expStartTimeStamp, wcs, medianScAttitude&
        !    , (EvLow+EvHigh)*500.0)	! The Last-argument == E/2*1000 (eV)
        rateImages(1,:,:) = rateImageTmp
        deallocate(rateImageTmp)

      case default
         call error('badOutputStyleB4',"Should not happen.  Contact the code developer.")
      end select	! select case (styleOut) in case(styleIn)=('image')

    case default
      call error('badInputStyleB2',"Should not happen.  Contact the code developer.")
    end select	! select case(styleIn)

    ! deallocate(pattFracSingle)
    call release(inputSet)


    !...................................................................
    ! Output:

    select case(OutputStyle)
    case('raw')

      call message(appMsg, Verbose, 'Writing mask cube.')
      call writeCubeData(trim(MaskSetName), trim(outputRefSetName), maskCube&
        , infoCcdChips%rawKeyInfo)
        ! , rawKeyInfo(:numCubePlanes))
      ! in eimchip2sky/cube_aux.

      if (WithRateImage) then
        call message(appMsg, Verbose, 'Writing rate cube.')
        call writeCubeData(trim(RateSetRoot)//'.ds', trim(outputRefSetName), rateCube&
           , infoCcdChips%rawKeyInfo)
           ! , rawKeyInfo(:numCubePlanes))
        ! in eimchip2sky/cube_aux.
      end if

      deallocate(rateCube)

    case('sky')	! from ./eootemap_mod.f90

!       allocate(binnedAttWeights(numCubePlanes, numAttBins))
! 
!       !.................................................................
!       ! Construct the attitude vs. image table of weights:
! 
!       call message(AppMsg, Verbose, "Calculating attitude weights.")
! 
!       do imageNum = 1, numCubePlanes
!         call gtiWeightBinnedAttitude(gtiVec(imageNum)%gtis, binnedScAtt)
!         ! in attcalc2/binned_att.
! 
!         binnedAttWeights(imageNum,:) = binnedScAtt%gtiDuration
!       end do ! imageNum
! 
! !********** normalize?????

      call   message(appMsg, Verbose, 'Writing mask image.')
      call   saveImage(maskImageInt, trim(TemplateImageSetName), trim(MaskSetName)&
           , trim(temporaryImageFileName), wcs=wcs) ! in ssclib/save_image.

      if (WithRateImage) then
        write(errStr,'("Writing rate images: (number of attitude bins = ",i5,")")') numAttBins
        call message(appMsg, Verbose)
        if (size(rateImages,1) < numAttBins) then
          print *,'DEBUG: size(rateImages,1)=',size(rateImages,1), ', numAttBins =',numAttBins
          call message(appMsg, Verbose)
          ! call error('tooSmall_rateImages','Contact the code developer.')
        end if

        tmpFilename = ''
        do attBinNum = 1, min(size(rateImages,1),numAttBins)	! min() part should not be needed, but just to play safe.
        ! do attBinNum = 1, numAttBins
          ! if (.not. outputDiagn(attBinNum)%withRateImage)  cycle
          write(tmpFilename,'(a,i4.4,a)') trim(RateSetRoot), attBinNum, '.ds'
          call saveImage(rateImages(attBinNum,:,:) &
             , trim(TemplateImageSetName), trim(tmpFilename) &
             , trim(temporaryImageFileName), wcs=wcs) ! in ssclib/save_image.
        end do	! do attBinNum = 1, numAttBins
        ! call saveImage(rateImage, trim(TemplateImageSetName), trim(RateSetRoot)&
        !      , trim(temporaryImageFileName), wcs=wcs) ! in ssclib/save_image.
      end if

!!! For DEBUG
!call writeCubeData('rate_debug.ds', outputRefSetName, rateCube&
!   , rawKeyInfo(:numCubePlanes))
!!! Note: In order to activate the above, you need to add rateCube argument in the above call readEvent2SkyForEpileup(RateThreshold ...)

      deallocate(maskImageInt)

    case default
      call error('badOutputStyleB2',"Should not happen.  Contact the code developer.")
    end select

    call deallocateInfoCcdChipsT(infoCcdChips)
    call deallocateInfoFrameAttT(infoFrameAtt)

    ! puProbCube = 1.0 - exp(-puProbCube)

    ! if (withPuProbCube)&
    !   call writeCubeData(puProbCubeSetName, outputRefSetName, puProbCube&
    !     , rawKeyInfo(:numCubePlanes))
    !   ! in eimchip2sky/cube_aux.

    ! allocate(puMaskCube(numCubePlanes, chipXSize, chipYSize))
    ! puMaskCube = (puProbCube < pileupMaskProbability)

  end subroutine epileupmask


  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

    RateThreshold       = realParameter('ratethreshold')
    MaskSetName         = stringParameter('maskset')

    InputStyle          = stringParameter('inputstyle')
    select case(trim(InputStyle))
    case('evlist')
      EventSetName        = stringParameter('eventset')
    case('image')
      ImageSetName        = stringParameter('imageset')
      TemplateImageSetName = trim(ImageSetName)	! The command-line argument templateset may be specified, but anyway would be ignored.
      EvLow  = realParameter('evlo')
      EvHigh = realParameter('evhi')
      IsSingleEventOnly   = booleanParameter('issingleeventonly')
    case default
      call error('badInputStyle',"Value of --inputstyle must be either evlist or image.")
    end select

    OutputStyle         = stringParameter('outputstyle')
    select case(OutputStyle)
    case('sky')
      TemplateImageSetName     = stringParameter('templateset')
      if (InputStyle .eq. 'image') then
!call error('notSupported','Sorry, InputStyle==image is not supported, yet.')
      ! if ((trim(TemplateImageSetName) .eq. '') .and. (InputStyle .eq. 'image')) then
        if ((trim(TemplateImageSetName) .ne. trim(ImageSetName)) .and. &
            (trim(TemplateImageSetName) .ne. 'DEFAULT') .and. &
            (trim(TemplateImageSetName) .ne. '') ) then
          call warning('TemplateImageSetNameIgnored' &
             , 'templateset ('//trim(TemplateImageSetName)&
             //') is replaced with imageset ('//trim(ImageSetName)//').')
        end if
        TemplateImageSetName = ImageSetName	! regardless of user-specification
        !! If user-specified value is accepted, sky-to-sky conversion will be needed.
      else	! === if (InputStyle .eq. 'evlist') then
        if (trim(TemplateImageSetName) == '') call error('nullTemplateImageSetName'&
           , 'templateset must be specified in this case.')
      end if

    case('raw')
      ! Do nothing.
    case default
      call error('badOutputStyle',"Value of --outputstyle must be either sky or raw.")
    end select

    if ( (trim(InputStyle) == 'image') .or. (trim(OutputStyle) == 'sky') ) then
      AttitudeStyle = stringParameter('attstyle')
      select case(AttitudeStyle)
      case('binnedset')
        BinnedAttSetTabName = stringParameter('binnedattset')
        AvgCntToProcess = realParameter('avcnttoprocess')

      case('template')
      case default
        call error('badAttitudeStyle',"")
      end select
    end if

    WithRateImage   = booleanParameter('withrateimage')
    if (WithRateImage) then
      RateSetRoot     = stringParameter('ratesetroot')
    end if

    WithApproxPatFrac = booleanParameter('withapproxpatfrac')
    if (.not. WithApproxPatFrac) then
      call warning('withapproxpatfracNotSupported' &
         , 'withapproxpatfrac is not yet supported and is forced to be TRUE.')
      WithApproxPatFrac = .true.
    end if
    applyAttcalcFudge = booleanParameter('withboresightfudge')	! Global variable defined and used in ssclib/coordinate

  end subroutine readParameters


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module epileupmask_mod

