! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+
! Module: epileupmask_shared
!
! Authors:  Masaaki Sakano - SSC/LUX
!           Ed Chapin - ESAC
!
! Description: Shared routines for epiliupmask and eootepileupmask.  These basically read the data from the input file, calculate the mask etc and return them.
! 
! Comments: 
! 
! ToDo lists: 
!   * handle the attitude fractuation in calculation
!   * write generic coordinate transform tool
!   * write dumpArray in ssclib/dump
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module epileupmask_shared

  use dal
  use param  
  use errorHandling
  use cal
  use CalTypes		! for CAL_getPatternFractionServer, type(PatternFractionServerT)

  use binned_att	! in attcalc2
  use cube_aux		! in eimchip2sky
  use eimchip2sky_aux	! in eimchip2sky

  ! Followings in ssclib
  use array_utils
  use ssc_misc
  use cal_aux
  use dal_aux
  use epic_aux		! for getCcdNodeNum()
  !use intervals_aux	! for gtiVecType
  use wcs_aux		! in ssclib
  use coordinate	! in ssclib
  use dss_aux		! in ssclib

  use ebkgmap_skyframe 
  use ebkgmap_attcommon

  implicit none 

  private 

  ! public getCcdNodeNum
  public evlistColumns
  public initialSetUpForEpileup
  public getDefaultInfoCcdChipsT
  public instSetUpForEpileup
  public countNumberOfFrames
  public getPatSingleFractionArrayTmpl
  public getPatSingleFractionArrayAt
  public setupToReadDataForEpileup
  public calcStartFinisTime
  public readDataForEpileup
  public readEventForEpileup
  public readEvent2SkyForEpileup
  public readEvent2RawForEpileup
  public readEvent2RateCubeForEpileup
  public readImage2SkyForEpileup
  public readImage2RawForEpileup
  public prepareForReadImageForEpileup
  public getAttSkyCoordForEpileup
  public convertMask2Real
  public allocateInfoCcdChipsT
  public deallocateInfoCcdChipsT
  public allocateInfoFrameAttT
  public deallocateInfoFrameAttT

  type, public :: infoCcdChipsT
    ! NOTE: 
    !  To allocate and deallocate the pointer elements, use
    !    allocateInfoCcdChipsT() and deallocateInfoCcdChipsT()

    integer(int32) :: instrumentId = XMM	! EPN etc.
    integer(int16) :: chipXSize, chipYSize &	! EPN_MAX_X etc.
       , tlmin, tlmax
    integer(int8)  :: numCcds &			! EPN_CCD_LAST etc.
       , numNodes = 1	! Default.
    integer(int32) :: numCubePlanes
    type(keyInfoType), pointer :: rawKeyInfo(:)	! type defined in eimchip2sky/cube_aux
    integer,           pointer :: cubePlaneIndices(:)
    real(double),      pointer :: pattFracSingle(:)	! grade pattern
    logical :: isAllocatedRawKeyInfo       = .false. &
             , isAllocatedCubePlaneIndices = .false. &
             , isAllocatedPattFracSingle   = .false.
  end type infoCcdChipsT

  type, public :: infoFrameAttT
    ! NOTE: 
    !  To allocate and deallocate the pointer elements, use
    !    allocateInfoFrameAttT() and deallocateInfoFrameAttT()

    real(double) :: expStartTimeStamp
    type(SpacecraftAttitudeType) :: medianScAttitude
    type(attBinType), pointer :: binnedScAtt(:)		! type def is in attcalc2/binned_att.
    real(single),     pointer :: numTotFrames(:) &	        ! (1:numCubePlanes)
                               , binnedAttsNumFrames(:,:)	! (1:numCubePlanes,1:numAttBins)
    logical :: isAllocatedBinnedScAtt         = .false. &
             , isAllocatedNumTotFrames        = .false. &
             , isAllocatedBinnedAttsNumFrames = .false.
  end type infoFrameAttT

  real(single), dimension(10), parameter :: &
    StandardAlpha1 = (/ 0.0,0.0,0.0, 0.7754, 0.7754, 0.6778, 0.0,0.0,0.0,0.0 /)
    !StandardAlpha1 = (/ 0.0,0.0,0.0, 0.778, 0.778, 0.778, 0.0,0.0,0.0,0.0 /)
    ! Indices mean instumentId, defined in caloalutils/src/caloaldefs.f90 .
    ! Alpha1 is defined in J. Ballet (1999, A&AS); the "standard" fraction of
    ! single pixel event at the standard energy (1.49 keV (Al Ka line),
    ! in this case).  Here the ratio values at the (other) energy band
    ! will be normalised referring to this value.

  type evlistColumns
    integer(int8),  pointer :: ccdNum(:), pattern(:)
    integer(int16), pointer :: rawX(:), rawY(:), pi(:)
    real,           pointer :: skyX(:), skyY(:)
    real(double),   pointer :: times(:)
  end type evlistColumns

  interface convertMask2Real
    module procedure convertMask1d2Real, convertMask2d2Real, convertMask3d2Real
  end interface

  interface getAttSkyCoordForEpileup
    module procedure getAttSkyCoordForEpileupInfoC, getAttSkyCoordForEpileupManyA
  end interface

  interface countNumberOfFrames
    module procedure countNumberOfFramesInfoCcd, countNumberOfFramesManyArgs
  end interface
  
  interface getPatSingleFractionArrayTmpl
    module procedure getPatSingleFractionArrayTmplMA, getPatSingleFractionArrayTmplIC
  end interface

  interface setupToReadDataForEpileup
    module procedure setupToReadDataForEpileupInfoC, setupToReadDataForEpileupManyA
  end interface

  interface readEvent2SkyForEpileup
    module procedure readEvent2SkyForEpileupInfoC, readEvent2SkyForEpileupManyA
  end interface

  interface readEvent2RawForEpileup
    module procedure readEvent2RawForEpileupInfoC, readEvent2RawForEpileupManyA
  end interface

  interface readEvent2RateCubeForEpileup
    module procedure readEvent2RateCubeForEpileupMA, readEvent2RateCubeForEpileupIC 
  end interface

  interface readImage2SkyForEpileup
    module procedure readImage2SkyForEpileupInfoC, readImage2SkyForEpileupManyA
  end interface

  interface readImage2RawForEpileup
    module procedure readImage2RawForEpileupInfoC, readImage2RawForEpileupManyA
  end interface

contains 

  !---------------------------------------------------------------------
  subroutine initialSetUpForEpileup(style, eventName, imageName&
     , outputRefSetName, inputSet, messageInputFile, rateImage)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! Set up for
    !  inputSet, messageInputFile, outputRefSetName, 
    !  rateImage if input is 'image' (in that case, rateImage must be present!)
    !  Either of eventName or imageName can be null character set.

    character(*), intent(in) ::&
      style,&
      eventName,&
      imageName
    character(*), intent(out) ::&
      outputRefSetName,&
      messageInputFile
    type(DataSetT), intent(out) :: inputSet
    real(single), pointer, optional :: rateImage(:,:)

    !-------------------------------------------------------------------
    ! Start:

    select case(style)
    case('evlist')
      messageInputFile = 'event list'
      call message(appMsg, Verbose, 'Reading info from '//trim(messageInputFile)//' header.')

      inputSet = dataSet(trim(eventName), READ)
      outputRefSetName = trim(eventName)

    case('image')
      messageInputFile = 'image FITS'

      if (present(rateImage)) then
        call message(appMsg, Verbose, 'Reading info from '//trim(messageInputFile)//' header.')
        call readArrayData(trim(imageName), rateImage)	! in ssclib/dal_aux.
        inputSet = dataSet(trim(imageName), READ)
        outputRefSetName = trim(imageName)
      end if

    case default
      call error('badInputStyleB',"Should not happen.  Contact the code developer.")
    end select

  end subroutine initialSetUpForEpileup


  !---------------------------------------------------------------------
  function getDefaultInfoCcdChipsT(instrumentId, allowNodes) result(infoCcdChips)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! Returns the default infoCcdChipsT variable, based on instrumentId,
    ! which may be given explicitly or taken from the current CAL seting.
    !
    ! NOTE: the number of the nodes is in default 1 even for MOSs,
    ! unless the allowNodes option is given and .true.

    type(infoCcdChipsT) :: infoCcdChips	! Defined in this code.
    integer(int32), intent(in), optional :: instrumentId
    logical,        intent(in), optional :: allowNodes	! .false. in default

    !-------------------------------------------------------------------
    ! Start

    if (present(instrumentId)) then
      infoCcdChips%instrumentId = instrumentId
    else
      call CAL_getState(instrument=infoCcdChips%instrumentId)
    end if

    select case(infoCcdChips%instrumentId)
    case(EPN)
      infoCcdChips%chipXSize = EPN_MAX_X	! All these are defined in caloaldefs.
      infoCcdChips%chipYSize = EPN_MAX_Y
      infoCcdChips%numCcds   = EPN_CCD_LAST
      infoCcdChips%numNodes  = 1
      infoCcdChips%tlmin = 0
      infoCcdChips%tlmax = 32767	! Default, probably used if the input is Image.

    case(EMOS1, EMOS2)
      infoCcdChips%chipXSize = EMOS_MAX_X
      infoCcdChips%chipYSize = EMOS_MAX_Y
      infoCcdChips%numCcds   = EMOS_CCD_LAST
      infoCcdChips%numNodes  = 1
      if (present(allowNodes)) then
        if (allowNodes) infoCcdChips%numNodes = 2
        ! In principle MOSs can have 2 nodes.  However, in practice,
        ! two nodes have never been used in the actual observations
        ! except for a few calibration observations.
      end if
      infoCcdChips%tlmin = 0
      infoCcdChips%tlmax = 15000	! Default, probably used if the input is Image.

    case default
call error('badInstrument',"in ebkgmap/epileupmask_shared")
    end select	! select case(infoCcdChips%instrumentId)

  end function getDefaultInfoCcdChipsT


  !---------------------------------------------------------------------
  subroutine instSetUpForEpileup(inputSet, infoCcdChips, withTlminmax)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! Set up for instrument-related stuffs.

    type(DataSetT),      intent(in)  :: inputSet
    type(infoCcdChipsT), intent(out) :: infoCcdChips	! Defined in this code.
    logical, intent(in), optional :: withTlminmax

    integer(int32) :: instrumentId
    character(20)  :: instrumentName
    logical :: withTlminmaxLocal

    !-------------------------------------------------------------------
    ! Start

    withTlminmaxLocal = .false.
    if (present(withTlminmax)) withTlminmaxLocal=withTlminmax

    instrumentName = stringAttribute(inputSet, "INSTRUME")
    instrumentId = instId(instrumentName) ! in caloalutils/src/caloaldefs.f90
    call CAL_setState(instrument=instrumentId)

    !! Get it.
    infoCcdChips = getDefaultInfoCcdChipsT(instrumentId)

    !! Optional setting
    if (withTlminmaxLocal) then
      if (hasAttribute(inputSet, 'TLMIN9')) then
        infoCcdChips%tlmin = int16Attribute(inputSet, 'TLMIN9')
      else	! Probably the input is Image.  It is USED but in the extention 1!!!????
        call message(appMsg, Verbose &
           , "TLMIN9 attribute is not found in the input dataset.  Default value is used.")
      end if

      if (hasAttribute(inputSet, 'TLMAX9')) then
        infoCcdChips%tlmax = int16Attribute(inputSet, 'TLMAX9')
      else	! Probably the input is Image.
        call message(appMsg, Verbose &
           , "TLMAX9 attribute is not found in the input dataset.  Default value is used.")
      end if
    end if	! if (withTlminmaxLocal) then

  end subroutine instSetUpForEpileup


  !---------------------------------------------------------------------
  subroutine instSetUpForEpileup_org(inputSet, instrumentId, chipXSize, chipYSize&
     , numCcds, numNodes, tlmin, tlmax)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! Set up for instrument-related stuffs.

    type(DataSetT), intent(in) :: inputSet
    integer(int32), intent(out) :: instrumentId
    integer(int16), intent(out) :: chipXSize, chipYSize
    integer(int8), intent(out) :: numCcds, numNodes
    integer(int16), intent(out), optional :: tlmin, tlmax
    character(20) :: instrumentName

    integer(int16) :: tlminLocal, tlmaxLocal

    !-------------------------------------------------------------------
    ! Start

    instrumentName = stringAttribute(inputSet, "INSTRUME")
    instrumentId = instId(instrumentName) ! in caloalutils/src/caloaldefs.f90
    call CAL_setState(instrument=instrumentId)

    if (instrumentId == EPN) then
      chipXSize = EPN_MAX_X	! All these are defined in caloaldefs.
      chipYSize = EPN_MAX_Y
      numCcds = EPN_CCD_LAST
      numNodes = 1
      tlminLocal = 0
      tlmaxLocal = 32767	! Default, probably used if the input is Image.

    else if (instrumentId == EMOS1 .or. instrumentId == EMOS2) then
      chipXSize = EMOS_MAX_X
      chipYSize = EMOS_MAX_Y
      numCcds = EMOS_CCD_LAST
      numNodes = 1
      ! numNodes = 2
      ! In principle MOSs can have 2 nodes.  However, in practice,
      ! two nodes have never been used in the actual observations
      ! except for a few calibration observations.
      tlminLocal = 0
      tlmaxLocal = 15000	! Default, probably used if the input is Image.

    else
call error('badInstrument',"")
    end if

    if (present(tlmin)) then
      tlmin = tlminLocal
      if (hasAttribute(inputSet, 'TLMIN9')) then
        tlmin = int16Attribute(inputSet, 'TLMIN9')
      else	! Probably the input is Image.  It is USED but in the extention 1!!!????
        call message(appMsg, Noisy, "Input dataset does not have TLMIN9 attribute.  Default value is used.")
      end if
    end if

    if (present(tlmax)) then
      tlmax = tlmaxLocal
      if (hasAttribute(inputSet, 'TLMAX9')) then
        tlmax = int16Attribute(inputSet, 'TLMAX9')
      else	! Probably the input is Image.
        call message(appMsg, Noisy, "Input dataset does not have TLMAX9 attribute.  Default value is used.")
      end if
    end if

  end subroutine instSetUpForEpileup_org


  !---------------------------------------------------------------------
  subroutine countNumberOfFramesManyArgs(inputSet, styleIn, styleOut&
     , instrumentId, numCcds, numNodes, binnedScAtt & ! , numAttBins
     , messageInputFile&
     , rawKeyInfo, numTotFrames, cubePlaneIndices&
     , numCubePlanes & !, maxCcdNodeNum
     , binnedAttsNumFrames)
  !---------------------------------------------------------------------
  ! Make sure to deallocate rawKeyInfo(:), numTotFrames(:), cubePlaneIndices(:),
  ! after use.
  ! binnedScAtt(:) and binnedAttsNumFrames(:,:) are needed only
  ! if (styleIn=='evlist' && styleOut=='sky').  Otherwise they are
  ! left undefined.
  ! Apart from binnedScAtt(:), those should NOT be allocated before the call.

    implicit none 

    !-- Description --
    ! Count number of frames of each chip (=numTotFrames and
    ! optionally binnedAttsNumFrames, the latter of which gives
    ! the number of frames per each chip per attitude bin).
    ! 
    ! -- Parameters -- 
    ! numTotFrames(1:numCubePlanes) : Array for cubePlaneNum.  It is allocated in this subroutine.
    ! numCubePlanes : The effective last index for cubePlaneIndices(:), which is maxCcdNodeNum or smaller if any CCD chip(s) is missing.

    character(*),   intent(in) :: styleIn, styleOut, messageInputFile	! styleIn/Out: (In|Out)putStyle
    type(DataSetT), intent(in) :: inputSet
    integer(int32), intent(in) :: instrumentId
    integer(int8),  intent(in) :: numCcds, numNodes
    type(attBinType), pointer :: binnedScAtt(:)	! intent(in)	! type def is in attcalc2/binned_att.   !!! Only needed if (styleIn=='evlist' && styleOut=='sky')
    ! integer,        intent(in) :: numAttBins
    type(keyInfoType), pointer :: rawKeyInfo(:)	! intent(out)	! defined in eimchip2sky/cube_aux
    real(single), pointer :: numTotFrames(:)	! intent(out)
    integer, pointer :: cubePlaneIndices(:)	! intent(out)
    integer(int32), intent(out) :: numCubePlanes
    ! integer,        intent(out) :: maxCcdNodeNum
    real(single), pointer, optional :: binnedAttsNumFrames(:,:)	! if (styleIn=='evlist' && styleOut=='sky'), this is mandatory!

    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT) :: infoFrameAtt	! Defined in epileupmask_shared
logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    infoCcdChips%instrumentId = instrumentId
    infoCcdChips%numCcds  = numCcds
    infoCcdChips%numNodes = numNodes
    if (styleIn=='evlist' .and. styleOut=='sky') then
      call allocateInfoFrameAttT(infoFrameAtt,binnedScAttPtr=binnedScAtt)
    end if

    call countNumberOfFramesInfoCcd(inputSet, styleIn, styleOut&
     , infoCcdChips, infoFrameAtt &
     , messageInputFile )

    rawKeyInfo       => infoCcdChips%rawKeyInfo
    numTotFrames     => infoFrameAtt%numTotFrames
    cubePlaneIndices => infoCcdChips%cubePlaneIndices
    numCubePlanes    =  infoCcdChips%numCubePlanes
    if (infoFrameAtt%isAllocatedBinnedAttsNumFrames) then
      binnedAttsNumFrames => infoFrameAtt%binnedAttsNumFrames
    end if

  end subroutine countNumberOfFramesManyArgs


  !---------------------------------------------------------------------
  subroutine countNumberOfFramesInfoCcd(inputSet, styleIn, styleOut&
     , infoCcdChips, infoFrameAtt &
     ! , instrumentId, numCcds, numNodes, binnedScAtt & ! , numAttBins
     ! , binnedScAtt & ! , numAttBins
     , messageInputFile )
     ! , rawKeyInfo, numTotFrames, cubePlaneIndices&
     ! , numTotFrames &
     ! , numCubePlanes & !, maxCcdNodeNum
     !, binnedAttsNumFrames)
  !---------------------------------------------------------------------
  ! Make sure to do
  !   call deallocateInfoCcdChipsT(infoCcdChips)
  !   call deallocateInfoFrameAttT(infoFrameAtt)
  ! after use.
  ! infoFrameAtt%binnedScAtt(:) are needed only if (styleIn=='evlist' && styleOut=='sky').
  ! and if so, infoFrameAtt%binnedAttsNumFrames(1:numCubePlanes,1:numAttBins)
  ! will be defined.  Otherwise the latter is left undefined.
  ! Apart from infoFrameAtt%binnedScAtt(:) (<= intent(in)),
  ! those should NOT be allocated before the call.

    implicit none 

    !-- Description --
    ! Count number of frames of each chip (=infoFrameAtt%numTotFrames and
    ! optionally infoFrameAtt%binnedAttsNumFrames, the latter of which gives
    ! the number of frames per each chip per attitude bin).
    ! 
    ! -- Parameters -- 
    ! numTotFrames(1:numCubePlanes) : Array for cubePlaneNum.  It is allocated in this subroutine.
    ! numCubePlanes : The effective last index for cubePlaneIndices(:), which is maxCcdNodeNum or smaller if any CCD chip(s) is missing.
    !
    !! *Basic Concept*
    ! Cube has planes, each of which represents each chip(CCD) and node.
    ! Thus, the cube has a structure or (Serial-No, x, y), where
    ! the Serial-No here is called "iCubePlaneNum".
    ! Note the "numCubePlanes" returned from this subroutine is
    ! the practical maximum value of these "iCubePlaneNum".
    ! Confusingly "maxPossibleNumCubePlanes" used in this subroutine is
    ! the maximum possible value of these "iCubePlaneNum" in any circumstance
    ! in the given instrument.
    !
    ! In the most simple case, Serial-No can be just the CCD chip number.
    ! However, MOSs can have 2 nodes, which make the things complicated.
    ! In addition, when there are uninterested CCD chips, to keep
    ! the large sum of empty planes in the cube is a waste of memory
    ! (well, though it would make the things a lot simpler...).
    !
    ! For that reason, the idea is the meaning of each "iCubePlaneNum"
    ! can be anything as a user wishes.
    ! Obviously the code needs know what each iCubePlaneNum physically means,
    ! hence the variables (e.g., rawKeyInfo(iCubePlaneNum)) that define
    ! the relation between iCubePlaneNum and chip/node numbers must be defined.
    !
    ! The code that calls this subroutine creates the chip-cube by itself.
    ! This subroutine returns the consistent parameters, such as,
    ! rawKeyInfo(:) for the chip-cube.
    !
    ! [*Definitions*]
    !   iCubePlaneNum[=Scalar]: = 1,2,...
    !     In default, it is as follows IF it is MOS with 2 nodes,
    !   (1:ccdNum=1,nodeNum=1, 2:ccdNum=1,nodeNum=2, 3:ccdNum=2,nodeNum=1, ...)
    !     or as follows IF it is MOS with 1 node,
    !   (1:ccdNum=1,nodeNum=1, 2:ccdNum=2,nodeNum=1, 3:ccdNum=3,nodeNum=1, ...)
    !     May not be continuous, i.e., some chips may be missing/undefined.
    !     Therefore the following is, for example, possible in principle:
    !   (1:ccdNum=2,nodeNum=1, 2:ccdNum=4,nodeNum=1, 3:ccdNum=4,nodeNum=2, ...)
    !     In fact, even the random order would be allowed (though not recommended):
    !   (1:ccdNum=6,nodeNum=1, 2:ccdNum=1,nodeNum=2, 3:ccdNum=3,nodeNum=1, ...)
    !
    !    Note this is necessary in terms of the interface with
    !     other routines, such as rebinChipImagesToSky in eimchip2sky.
    !     (The justification is, to save the memory!)
    !    Note the "numCubePlanes" returned from this subroutine is
    !     the practical maximum value of these "iCubePlaneNum".
    !
    ! The information related to each "iCubePlaneNum" is stored in:
    !
    ! [*Definitions*]
    !   rawKeyInfo(iCubePlaneNum)%ccdNum =  ccdNum,
    !                           %nodeNum = nodeNum
    !   ccdNums(iCubePlaneNum)[=Array] = ccdNum	! Not used any more...
    !  Plus, in this subroutine, we also calculate
    !   numTotFrames(iCubePlaneNum)[=Array]: Number of Frames for iCubePlaneNum
    !
    ! In addition, it is convenient to define another single Serial-No that
    ! is unique for each combination of ccdNum and nodeNum, namely
    ! an "inverse" variable of the above rawKeyInfo:
    !
    ! [*Definitions*]
    !   ccdNodeNum[=Scalar]: = 10 * (nodeNum - 1) + ccdNum (<=getCcdNodeNum(ccdNum [, nodeNum]))
    !     pn ( 1(=ccdNum),2,...)
    !     MOS( 1(ccdNum=1,node=1), 2(ccdNum=2,node=1),..., 7(ccdNum=7,node=1))
    !       possibly, further (11(ccdNum=1,node=2),..., 17(ccdNum=7,node=2))
    !   cubePlaneIndices(ccdNodeNum)[=Array] = iCubePlaneNum
    !
    !  Namely, ccdNodeNum is a unique Scalar for a given ccdNum (and nodeNum).
    !  Then cubePlaneIndices(ccdNodeNum) gives the index number (=iCubePlaneNum)
    !  to be used for the array.

    character(*),   intent(in) :: styleIn, styleOut, messageInputFile	! styleIn/Out: (In|Out)putStyle
    type(DataSetT), intent(in) :: inputSet
    type(infoCcdChipsT), intent(inout) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT), intent(inout) :: infoFrameAtt	! Defined in epileupmask_shared
    ! integer(int32) :: instrumentId
    integer(int8)  :: numCcds, numNodes
    ! type(attBinType), pointer :: binnedScAtt(:)	! intent(in)	! type def is in attcalc2/binned_att.   !!! Only needed if (styleIn=='evlist' && styleOut=='sky')
    ! integer,        intent(in) :: numAttBins
    ! type(keyInfoType), pointer :: rawKeyInfo(:)	! intent(out)	! defined in eimchip2sky/cube_aux
    ! real(single), pointer :: numTotFrames(:)	! intent(out)
    integer, pointer :: cubePlaneIndices(:)	! intent(out)
    integer(int32) :: numCubePlanes
    ! integer,        intent(out) :: maxCcdNodeNum
    ! real(single), pointer, optional :: binnedAttsNumFrames(:,:)	! if (styleIn=='evlist' && styleOut=='sky'), this is mandatory!

    integer(int32) :: iCubePlaneNum
    integer :: numAttBins
    ! integer(int8), pointer :: ccdNums(:)
    integer(int8) :: ccdNum, nodeNum
    integer(int32) :: maxPossibleNumCubePlanes, ccdNodeNum
    type(BlockT) :: eventBlock, exposuBlock, primaryBlock
    type(ModeParameterDataType) :: modeParams	! type def is in cal/caltypes.
    type(IntervalT), pointer :: gtiPtr(:)	! type def is in cal/caltypes.
    type(keyInfoType), allocatable :: rawKeyInfoLoc(:)	! defined in eimchip2sky/cube_aux
    character(8) :: exposuTabName, exposuAttName, ontimeAttName, livetimeName
    real(double) :: totalGoodTime
    real(single), allocatable :: numTotFramesLoc(:), binnedAttsNumFramesLoc(:,:)
    type(attBinType), allocatable, target :: binnedAttTmp(:)
    type(attBinType), pointer :: binnedAttsNow(:)	! type def is in attcalc2/binned_att.
    logical, parameter :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start:

    ! instrumentId = infoCcdChips%instrumentId
    ! chipXSize    = infoCcdChips%chipXSize
    ! chipYSize    = infoCcdChips%chipYSize
    numCcds      = infoCcdChips%numCcds
    numNodes     = infoCcdChips%numNodes

    ! if (infoFrameAtt%isAllocatedBinnedScAtt) then
    !   binnedScAtt => infoFrameAtt%binnedScAtt
    ! end if

    !.................................................................
    ! Determine the number of frames for each chip.

    maxPossibleNumCubePlanes = numCcds * numNodes
    allocate(rawKeyInfoLoc(maxPossibleNumCubePlanes))	! , ccdNums(maxPossibleNumCubePlanes))
!    call initializeKeyInfoType(rawKeyInfoLoc)	! in eimchip2sky/cube_aux.
    ! rawKeyInfoLoc%gtiType = 0
    rawKeyInfoLoc%gtiTableName = ''
    ! ccdNums = 0
    allocate(numTotFramesLoc(maxPossibleNumCubePlanes))
    numTotFramesLoc(:) = 0.0

    call CAL_setState(instrument=infoCcdChips%instrumentId)

    call message(appMsg, verbose, 'Obtaining frame info from the EXPOSU tables of the '&
         //trim(messageInputFile)//'.')

    ! maxCcdNodeNum = getCcdNodeNum(numCcds, numNodes)
    ! allocate(cubePlaneIndices(maxCcdNodeNum))
    allocate(cubePlaneIndices(getCcdNodeNum(numCcds, numNodes)))
    cubePlaneIndices = 0 ! default

    select case(styleIn)
    case('evlist')
      eventBlock = block(inputSet, 'EVENTS')	! for OutputStyle=='sky'
      select case(styleOut)	! this preferrably includes double-loop below
      case('sky')
!print *,'before binnedAttsNum, maxPossibleNumCubePlanes=',maxPossibleNumCubePlanes,", numAttBins=", numAttBins
        if (.not. infoFrameAtt%isAllocatedBinnedScAtt) call error('isAllocatedBinnedScAttFalse','Contact the code developer.')
        numAttBins = size(infoFrameAtt%binnedScAtt)
        allocate(binnedAttsNumFramesLoc(maxPossibleNumCubePlanes, numAttBins))
        allocate(binnedAttTmp(numAttBins))
        ! binnedAttTmp = infoFrameAtt%binnedScAtt	! set later
        ! binnedAttsNow => binnedAttTmp			! set later
      end select

      iCubePlaneNum = 0
      do ccdNum = 1, numCcds
        do nodeNum = 1, numNodes

          ccdNodeNum = getCcdNodeNum(ccdNum, nodeNum)
          write(exposuTabName, '(a6, i2.2)') 'EXPOSU', ccdNodeNum
          write(livetimeName, '(a6, i2.2)') 'LIVETI', ccdNodeNum

          if (.not. hasBlock(inputSet, trim(exposuTabName))) then
            ! numTotFramesLoc(iCubePlaneNum) = -1.0
             write(errStr, '(a,i2,a,i1,a)') "Block ("//trim(exposuTabName) &
                //") for CCD=", ccdNum, ", NODE=", nodeNum &
                , " is not found.  Therefore the chip/node is assumed to be not used."
             if (nodeNum > 1) then
               call message(appMsg, Noisy)	! Print the message for Node==2, if any, only when Noisy.  In practice, the calling routine may give numNodes=1 first of all, then this would never be the case.
             else
               call message(appMsg, Verbose)
             end if
            cycle	! cubePlaneIndices(ccdNodeNum) == 0 (default)
          end if

          iCubePlaneNum = iCubePlaneNum + 1
          ! ccdNums(   iCubePlaneNum) = ccdNum	! Node is not taken into account
          rawKeyInfoLoc(iCubePlaneNum)%ccdNum  = ccdNum
          rawKeyInfoLoc(iCubePlaneNum)%nodeNum = nodeNum
          cubePlaneIndices(ccdNodeNum) = iCubePlaneNum

          exposuBlock = block(inputSet, exposuTabName)
          call CAL_setState(exposuBlock)
          call CAL_getModeParameters(modeParams)

          select case(styleOut)	! this is preferrably outside of double-loop
          case('raw')
            ! EC: Previously totalGoodTime was obtained using:
            !   totalGoodTime = real64Attribute(exposuBlock, 'LIVETIME')
            ! However, the EXPOSU extensions seem to have incorrect values
            ! if evselect was used to select based on TIME, or using GTIs.
            ! The per-ccd LIVETI keywords seem to be correct in the
            ! main EVENTS extension.
            totalGoodTime = real64Attribute(eventBlock, livetimeName)

            ! = sum of GTIs times the 'dead time' (which, at least in MOS
            ! parlance, means the average time lost to cosmic rays).

            numTotFramesLoc(iCubePlaneNum) = totalGoodTime * 1000.0d0&
               / modeParams%frameTime ! %frameTime is in milliseconds.

          case('sky')	! from eootepiliupmask_mod
            call readGtisFromDss(inputSet, eventBlock, ccdNodeNum, gtiPtr)
            ! in ssclib/dss_aux.

            binnedAttTmp = infoFrameAtt%binnedScAtt
            binnedAttsNow => binnedAttTmp
if (testFlag) then
print *, 'DEBUG:(ccdNodeNum=',ccdNodeNum,',size=',size(gtiPtr) &
,')binnedAttsNow%gtiDuration=',binnedAttsNow%gtiDuration
if (size(gtiPtr) .ne. 0) then
print *, 'DEBUG:gti%lower=',gtiPtr(:)%lower,' gti%upper=',gtiPtr(:)%upper
print *, 'DEBUG:gti Diff=',gtiPtr%upper-gtiPtr%lower
end if
end if
            call gtiWeightBinnedAttitude(gtiPtr, binnedAttsNow)
            ! in attcalc2/binned_att.

            ! The call to gtiWeightBinnedAttitude here has a different function to the same call in, for example, eimchip2sky. In the latter, the end result is a real-valued image in sky coordinates which is a superposition of contributions, each of which consists of a projection of the detector CCDs at a separate sample of attitude. In this case one needs to know the weight of each attitude sample, which is given by the length of good time spent at this attitude. The routine gtiWeightBinnedAttitude returns these weight values. In the present case however the output sky image is (conceptually at least) a logical-valued mask, formed by performing a logical OR of the contributing masks. There is no need to know weights to perform this OR (except in the special case that the weight was zero). Thus there is no reason to call gtiWeightBinnedAttitude from this point of view.
            !
            ! On the other hand, the constituent masks at each attitude sample are generated by calculating the amount of pileup during the period at which that attitude sample is applicable. The pileup calculation requires the total GTI during that period - hence we find that gtiWeightBinnedAttitude must be called after all. Note however that in the call to rebinChipImagesToSky, the dummyWeights argument is used to simply set the attitude weights to 1.

            deallocate(gtiPtr)

if (testFlag) then
print *, 'DEBUG(ccd=',ccdNum,'):modeParams%frameTime=',modeParams%frameTime
print *, 'DEBUG:binnedAttsNow%gtiDuration=',binnedAttsNow%gtiDuration
end if

            binnedAttsNumFramesLoc(iCubePlaneNum,:) = binnedAttsNow%gtiDuration&
             * 1000.0d0 / modeParams%frameTime	! %frameTime is in milliseconds.
            numTotFramesLoc(iCubePlaneNum) = numTotFramesLoc(iCubePlaneNum) + sum(binnedAttsNumFramesLoc(iCubePlaneNum,:))

          case default
            call error('NotSupported', '')
          end select	! select case(styleOut)

          call release(exposuBlock)
        end do	! do ccdNum = 1, numCcds
      end do	! do nodeNum = 1, numNodes
      call release(eventBlock)

      ! If this subroutine ever returns before this point, make sure to be aware of the following.
      select case(styleOut)
      case('sky')
        numCubePlanes = iCubePlaneNum
        call allocateInfoFrameAttT(infoFrameAtt, binnedAttsNumFramesSizes=(/ numCubePlanes, numAttBins /))
        infoFrameAtt%binnedAttsNumFrames = binnedAttsNumFramesLoc(1:numCubePlanes,:)
        ! if (present(binnedAttsNumFrames)) then
        !   allocate(binnedAttsNumFrames(iCubePlaneNum,numAttBins))
        !   binnedAttsNumFrames = binnedAttsNumFramesLoc(1:iCubePlaneNum,:)
        ! end if
        deallocate(binnedAttsNumFramesLoc)	! allocatable
      end select

    case('image')
      call CAL_setState(inputSet)
      call CAL_getModeParameters(modeParams)

      primaryBlock = block(inputSet, 0)
      iCubePlaneNum = 0
      do ccdNum = 1, numCcds
        exposuAttName = 'EXPOSURE'
        write(ontimeAttName, '(a6, i2.2)') 'ONTIME', ccdNum
        if (     hasAttribute(inputSet, ontimeAttName)) then
          totalGoodTime = real64Attribute(inputSet, ontimeAttName)
          write(errStr, '(a,es15.8)') trim(ontimeAttName)//' is found in the primary header:', totalGoodTime
        else if (hasAttribute(inputSet, exposuAttName)) then
          totalGoodTime = real64Attribute(inputSet, exposuAttName)
          write(errStr, '(a,es15.8)') trim(exposuAttName)//' is found in the primary header:', totalGoodTime
        else if (hasAttribute(primaryBlock, ontimeAttName)) then
          totalGoodTime = real64Attribute(primaryBlock, ontimeAttName)
          write(errStr, '(a,es15.8)') trim(ontimeAttName)//' is found:', totalGoodTime
        else if (hasAttribute(primaryBlock, exposuAttName)) then
          totalGoodTime = real64Attribute(primaryBlock, exposuAttName)
          write(errStr, '(a,es15.8)') trim(exposuAttName)//' is found:', totalGoodTime
        else
          call error('noExposureOntimeInfo'&
             , "Neither ONTIME nor EXPOSURE info is available in the input image.")
        end if
        call message(appMsg, Noisy)

        do nodeNum = 1, numNodes
          ccdNodeNum = getCcdNodeNum(ccdNum, nodeNum)

          iCubePlaneNum = iCubePlaneNum + 1

          rawKeyInfoLoc(iCubePlaneNum)%ccdNum  = ccdNum
          rawKeyInfoLoc(iCubePlaneNum)%nodeNum = nodeNum
          cubePlaneIndices(ccdNodeNum) = iCubePlaneNum

          numTotFramesLoc(iCubePlaneNum) = totalGoodTime * 1000.0d0&
             / modeParams%frameTime ! %frameTime is in milliseconds.
        end do
      end do
      call release(primaryBlock)

    case default
      call error('badInputStyleB3',"Should not happen.  Contact the code developer.")
    end select

    ! If this subroutine ever returns before this point, make sure to be aware of the following.

    numCubePlanes = iCubePlaneNum
    infoCcdChips%numCubePlanes = numCubePlanes

    call allocateInfoCcdChipsT(infoCcdChips &
       , rawKeyInfoSize=numCubePlanes &
       , cubePlaneIndicesPtr=cubePlaneIndices)

    ! allocate(rawKeyInfo(numCubePlanes), numTotFrames(numCubePlanes))
    infoCcdChips%rawKeyInfo = rawKeyInfoLoc(1:numCubePlanes)	! The contents after the index numCubePlanes in rawKeyInfoLoc() are not used, hence are cut off.
    ! rawKeyInfo   = rawKeyInfoLoc(1:numCubePlanes)

    call allocateInfoFrameAttT(infoFrameAtt, numTotFramesSize=numCubePlanes)
    infoFrameAtt%numTotFrames = numTotFramesLoc(1:numCubePlanes)	! The contents after the index numCubePlanes in numTotFramesLoc() are not used, hence are cut off.
    ! numTotFrames = numTotFramesLoc(1:numCubePlanes)

if (testFlag) &
print *, 'DEBUG:shared: numTotFrames = ', numTotFramesLoc

    deallocate(rawKeyInfoLoc, numTotFramesLoc)	! allocatable

  end subroutine countNumberOfFramesInfoCcd


  !---------------------------------------------------------------------
  subroutine getPatSingleFractionArrayTmplMA(chipXSize, chipYSize, pattFracSingle, tlmax)
  !---------------------------------------------------------------------
  ! CAL must be set for instrumentID.

    implicit none 

    !-- Description --
    ! get the pattern fraction (grade branching ratio).

    integer(int16), intent(in) :: chipXSize, chipYSize
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    integer(int16), intent(in), optional :: tlmax

    integer(int16), dimension(:), pointer :: poChipX, poChipY
    integer(int16) :: xi, yi
    integer(int16) :: arXi(1), arYi(1)
    logical(bool) :: isChipXYDetermined = .false.

    !-------------------------------------------------------------------
    ! Start:

    allocate(poChipX(1), poChipY(1))

    !! Determine poChipX & poChipY that are OK.
    do xi=8, chipXSize
      arXi(1)=xi
      do yi=8, max(8, min(chipYSize-8, 30))	! 8 <= y <= (30 or chipYSize-8)
        arYi(1)=yi
        if (CAL_isPixelBad(xi, yi)) then
          cycle
        else
          call CAL_pixCoord1ToChipCoord(arXi, arYi, poChipX, poChipY)	! xi is rawX
          ! This routine is called only once.
          ! Therefore, the array (arXi etc) with the size of 1 is efficient.
          isChipXYDetermined = .true.
          exit
        end if
      end do
    end do
    if (.not. isChipXYDetermined) then
      ! if no good combination is found in ([8:],[8:]).
      call warning('noGoodChipXYFound',&
         "No good ChipX/Y is found.  The pattern fraction is unreliable.")
      poChipX(1) = 5
      poChipY(1) = 5
    end if

    write(errStr, '("Chip coords of (",I3,",",I3,") is used for pattern fraction estimate.")')&
       poChipX(1), poChipY(1)
    call message(appMsg, Noisy)

    call getPatSingleFractionArrayAt(poChipX(1), poChipY(1), pattFracSingle, tlmax)

    deallocate(poChipX, poChipY)

  end subroutine getPatSingleFractionArrayTmplMA


  !---------------------------------------------------------------------
  subroutine getPatSingleFractionArrayTmplIC(infoCcdChips)
  !---------------------------------------------------------------------
  ! CAL must be set for instrumentID.

    implicit none 

    !-- Description --
    ! get the pattern fraction (grade branching ratio).

    type(infoCcdChipsT), intent(inout) :: infoCcdChips	! Defined in epileupmask_shared
    real(double), pointer :: pattFracSingle(:)	! grade pattern

    !-------------------------------------------------------------------
    ! Start:

    if (infoCcdChips%tlmax > 1) then
      call getPatSingleFractionArrayTmplMA(infoCcdChips%chipXSize &
                                         , infoCcdChips%chipYSize &
                                         , pattFracSingle &
                                         , infoCcdChips%tlmax)
    else	! tlmax looks invalid.
      call getPatSingleFractionArrayTmplMA(infoCcdChips%chipXSize &
                                         , infoCcdChips%chipYSize &
                                         , pattFracSingle)
    end if

    call allocateInfoCcdChipsT(infoCcdChips, pattFracSinglePtr=pattFracSingle)

  end subroutine getPatSingleFractionArrayTmplIC


  !---------------------------------------------------------------------
  subroutine getPatSingleFractionArrayAt(chipX, chipY, pattFracSingle, tlmax)
  ! subroutine getPatSingleFractionArray(chipX, chipY, iLo, iHi, pattFracSingle)	! iLo==tlmin, iHi=tlmax
  !---------------------------------------------------------------------
  ! CAL must be set for instrumentID.

    implicit none 

    !-- Description --
    ! get the pattern fraction (grade branching ratio) at a position.

    integer(int16), parameter :: ePatMin = 1

    integer(int16), intent(in) :: chipX, chipY
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    integer(int16), intent(in), optional :: tlmax

    type(PatternFractionServerT) :: patternFractionServer	! in cal
    ! type(PatternFractionServerT) :: patternFractionServer, patternFractionAt	! in cal
    type(curveT) :: curvePatFrac	! in CalTypes
    integer(int32) :: pi, instrumentId
    integer(int16) :: ePatMax

    !-------------------------------------------------------------------
    ! Start:

    call CAL_getState(instrument=instrumentId)
    select case(instrumentId)
    case(EPN)
      ePatMax = 32767	! Default, probably used if the input is Image.
    case(EMOS1, EMOS2)
      ePatMax = 15000	! Default, probably used if the input is Image.
    case default
call error('badInstrument',"in ebkgmap/epileupmask_shared")
    end select	! select case(instrumentId)

    if (present(tlmax)) ePatMax = tlmax

    patternFractionServer = CAL_getPatternFractionServer()

    allocate(pattFracSingle(ePatMin:ePatMax))
    !! Get the array of alpha1(PI)
    !allocate(pattFracSingle(tlmin:tlmax))

    call CAL_patternFractionAt(patternFractionServer, chipX, chipY)

    curvePatFrac = CAL_getPatternFractionInEnergy(patternFractionServer, GRADE_SINGLE)
    do pi=ePatMin,ePatMax
      pattFracSingle(pi)=CAL_curveValueAt(curvePatFrac, real(pi, kind=double))
    end do

    ! do pi=ePatMin,ePatMax,100
    !   print *,'Frac(',pi,')=',pattFracSingle(pi)
    ! end do
    write(errStr, '("pattFracSingle(1478eV)  =",f7.4)') pattFracSingle(1487)
    call message(appMsg, Noisy)
    write(errStr, '("pattFracSingle(12000eV) =",f7.4)') pattFracSingle(12000)
    call message(appMsg, Noisy)

! print *,'before pattFracSingle'
!     pattFracSingle = CAL_getPatternFractionInChannel(patternFractionServer, GRADE_SINGLE)
! print *,'after pattFracSingle'
! print *,"pattFracSingle= "
! print *,pattFracSingle
!       ! GRADE_SINGLE = 0, GRADE_DOUBLE = 1, GRADE_TRIPLE = 2, GRADE_QUADRUPLE = 3, etc.
!       ! defined in /cal/doc/calsources/caltypes.f90


    !!!!! DEBUG!!!!! ?????
!    allocate(pattFracSingle(0:15000))
!    if (chipX > 0 .and. chipY > 0)&	! Dummy statement to suppress warning.
!       pattFracSingle=0.778

  end subroutine getPatSingleFractionArrayAt


  !---------------------------------------------------------------------
  subroutine setupToReadDataForEpileupManyA(styleIn, styleOut&
     , inputSet, chipXSize, chipYSize & !, maxCcdNodeNum&
     , cubePlaneIndices& 
     , evCol, eventTabRet)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! set up for reading data from the input event/image, eg., setup pointers.

    character(*), intent(in) :: styleIn, styleOut	! InputStyle, OutputStyle
    type(DataSetT), intent(in) :: inputSet
    ! integer, intent(in) :: maxCcdNodeNum
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer, pointer :: cubePlaneIndices(:)
    type(evlistColumns), intent(out) :: evCol
    type(TableT), intent(out), optional :: eventTabRet

    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared

    !-------------------------------------------------------------------
    ! Start:

    infoCcdChips%chipXSize = chipXSize
    infoCcdChips%chipYSize = chipYSize
    call allocateInfoCcdChipsT(infoCcdChips &
       , cubePlaneIndicesPtr=cubePlaneIndices)

    call setupToReadDataForEpileupInfoC(styleIn, styleOut&
       , inputSet, infoCcdChips &
       , evCol, eventTabRet)

    ! infoCcdChips%isAllocatedCubePlaneIndices = .false.	! so that it will not be deallocated.
    call deallocateInfoCcdChipsT(infoCcdChips, doCubePlaneIndices=.false.)

  end subroutine setupToReadDataForEpileupManyA


  !---------------------------------------------------------------------
  subroutine setupToReadDataForEpileupInfoC(styleIn, styleOut&
     , inputSet, infoCcdChips &
     ! , chipXSize, chipYSize & !, maxCcdNodeNum&
     ! , cubePlaneIndices& 
     , evCol, eventTabRet)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! set up for reading data from the input event/image, eg., setup pointers.

    character(*), intent(in) :: styleIn, styleOut	! InputStyle, OutputStyle
    type(DataSetT), intent(in) :: inputSet
    ! integer, intent(in) :: maxCcdNodeNum
    type(infoCcdChipsT), intent(in) :: infoCcdChips	! Defined in epileupmask_shared
    ! integer(int16), intent(in) :: chipXSize, chipYSize
    ! integer, pointer :: cubePlaneIndices(:)
    type(evlistColumns), intent(out) :: evCol
    type(TableT), intent(out), optional :: eventTabRet

    type(TableT) :: eventTab
    ! integer :: numEvents
    integer(int32) :: ccdNodeNum
    integer(int8), pointer :: ccdNumValuesInEvlist(:) 
    integer(int32) :: i

    !-------------------------------------------------------------------
    ! Start:

    select case(styleIn)
    case('evlist')

      eventTab = table(inputSet, 'EVENTS', READ)
      ! numEvents = numberOfRows(eventTab)
      evCol%ccdNum  => int8Data( column(eventTab, 'CCDNR'))
      evCol%pattern => int8Data( column(eventTab, 'PATTERN'))
      evCol%pi      => int16Data(column(eventTab, 'PI'))
      evCol%rawX    => int16Data(column(eventTab, 'RAWX'))
      evCol%rawY    => int16Data(column(eventTab, 'RAWY'))
      evCol%times   => real64Data(column(eventTab,'TIME'))

      ! Do some checks:

      ccdNumValuesInEvlist => findUniqueValues(evcol%ccdNum) ! in ssclib/array_utils.
      if (minval(ccdNumValuesInEvlist) < 1&
      .or.maxval(ccdNumValuesInEvlist) > size(infoCcdChips%cubePlaneIndices))&
      ! .or.maxval(ccdNumValuesInEvlist) > maxCcdNodeNum)&
call error('',"")
      do i = 1, size(ccdNumValuesInEvlist)
        ccdNodeNum = ccdNumValuesInEvlist(i)
        if (infoCcdChips%cubePlaneIndices(ccdNodeNum) <= 0)&
call error('',"")
      end do
      deallocate(ccdNumValuesInEvlist)


      select case (styleOut)
      case('sky')
        ! Do nothing
      case('raw')
        ! Do some checks:

        if (minval(evCol%rawX) < 1&
        .or.maxval(evCol%rawX) > infoCcdChips%chipXSize)&
call error('',"")

        if (minval(evCol%rawY) < 1&
        .or.maxval(evCol%rawY) > infoCcdChips%chipYSize)&
call error('',"")

      case default
         call error('badOutputStyleB1',"Should not happen.  Contact the code developer.")
      end select	! select case (styleOut) in case(styleIn)=('evlist')

    case('image')	! select case(styleIn)
      ! Do nothing
    case default
      call error('badInputStyleB2',"Should not happen.  Contact the code developer.")
    end select	! select case(styleIn)

    if (present(eventTabRet)) eventTabRet = eventTab

  end subroutine setupToReadDataForEpileupInfoC


  !---------------------------------------------------------------------
  subroutine calcStartFinisTime(infoFrameAtt, attBinNum &
     , evCol, minFrameToProcess &
     , startTime, finisTime, isCycle)
  !---------------------------------------------------------------------
  ! Returns startTime, finisTime, isCycle.
  ! 
  !-- Description --
  ! Calculate conservative start and end times for the applicability
  ! of each attitude sample. Note that it isn't necessary to consider
  ! the affect of GTIs on this period because we assume that the event
  ! list contains only events within these GTIs. These
  ! crudely-calculated start and stop times will therefore net the
  ! same number of events as a more sophisticated calculation would.
  ! Note `crudely' means GTI is not taken into account below.

    implicit none 

    type(infoFrameAttT), intent(in) :: infoFrameAtt	! Defined in epileupmask_shared
    integer,             intent(in) :: attBinNum
    type(evlistColumns), intent(in) :: evCol
    integer(int32),      intent(in) :: minFrameToProcess
    real(double),        intent(out) :: startTime, finisTime	! Return
    logical,             intent(out) :: isCycle	! Whether skip this frame (=attBinNum) or not.

    integer :: numAttBins
    real(single) :: maxNumFrames
logical, parameter :: testFlag = .false.
!logical, parameter :: testFlag = .true.

    !-------------------------------------------------------------------
    ! Start:

    isCycle = .false.

    numAttBins = size(infoFrameAtt%binnedAttsNumFrames, 2)
    ! numFrames => binnedAttsNumFrames(:,attBinNum)
if (testFlag) print *, 'DEBUG:numFrames=',infoFrameAtt%binnedAttsNumFrames(:,attBinNum)
    maxNumFrames = maxval(infoFrameAtt%binnedAttsNumFrames(:,attBinNum))	! The first rank is equivalent to (1:numCubePlanes).


    if (attBinNum == 1) then
       startTime = minval(evCol%times)
    else
       startTime = infoFrameAtt%binnedScAtt(attBinNum)%binStartTime
    end if

    if (attBinNum == numAttBins) then
       finisTime = maxval(evCol%times)
if (testFlag) print *, 'DEBUG:finisTime=',finisTime
    else
       finisTime = infoFrameAtt%binnedScAtt(attBinNum+1)%binStartTime
    end if

    if (finisTime < startTime) then
      write(errStr, '(a, i4, a, es21.15, " > ", es21.15, a)') &
         'Skip the attitude bin', attBinNum&
         , ' because the start time is later than the end time: (' &
         , startTime , finisTime &
         , '), which likely implies that the event file'&
         //' does not contain the data for the period in this attitude bin.'
      call warning('skipAttitudeBin')
      ! call message(AppMsg, Verbose)
      isCycle = .true.	! Will skip this Frame.
      return
    end if

    if (maxNumFrames >= minFrameToProcess) then
      write(errStr, '(a, i4, a, i4, a, f10.1, ".")') 'Processing the attitude bin'&
         , attBinNum, ' out of', numAttBins&
         , ', which has the maximum number of frames of', maxNumFrames
      call message(AppMsg, Verbose)
    else
      write(errStr, '(a, i4, a, f9.1)') 'Skip the attitude bin', attBinNum&
         , ' due to too small number of frames: (at most)', maxNumFrames
      call warning('skipAttitudeBin')
      ! call message(AppMsg, Verbose)

      isCycle = .true.	! Will skip this Frame.
    end if

    write(errStr, '(a,es21.15," - ",es21.15, a, f8.1, ".")') &
       ' which starts/ends roughly at ' &
       , startTime, finisTime &
       , ' with the time interval of ', finisTime - startTime
    call message(AppMsg, Noisy)

  end subroutine calcStartFinisTime


  !---------------------------------------------------------------------
  subroutine readDataForEpileup(styleIn, styleOut, rateThre&
     , flagIsSingleEventOnly, flagWithApproxPatFrac&
     , xiSize, yiSize, chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     , numFrames, rateImage, rateCube, maskImage, maskImageInt, maskCube&
     , evCol, expStartTimeStamp&
     , rawKeyInfo, wcs, att, withRateImage&
     , eVForCorrection, startTime, finisTime)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! read data from the input event/image.

    character(*), intent(in) :: styleIn, styleOut	! InputStyle, OutputStyle
    real(single), intent(in) :: rateThre
    logical, intent(in) :: flagIsSingleEventOnly, flagWithApproxPatFrac
    integer, intent(in) :: xiSize, yiSize
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    integer, pointer :: cubePlaneIndices(:)
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    real(single), pointer :: rateImage(:,:), rateCube(:,:,:), maskImage(:,:), numFrames(:)
    integer(int8), pointer :: maskImageInt(:,:)
    logical, pointer :: maskCube(:,:,:)
    ! logical, allocatable :: puMaskCube(:,:,:)
    type(evlistColumns), intent(in), optional :: evCol	! mandatory when InputStyle=="evlist"
    ! type(TableT), intent(in), optional :: eventTab
    real(double), intent(in), optional :: expStartTimeStamp
    type(keyInfoType), pointer, optional :: rawKeyInfo(:)	! intent(in)	! defined in eimchip2sky/cube_aux
    type(WcsT), intent(in), optional :: wcs	! type def in ssclib/wcs_aux
    type(SpacecraftAttitudeType), intent(in), optional :: att	! == binnedScAtt(attBinNum)%att
    logical, intent(in), optional :: withRateImage
    real, intent(in), optional :: eVForCorrection	! Energy(eV) to be used in correction, mandatory when InputStyle=="image"
    real(double), intent(in), optional :: startTime, finisTime	! unused if styleIn=='image'
     ! if this is <= 0, then the values defined in the event list are used.

    !-------------------------------------------------------------------
    ! Start:

    select case(styleIn)
    case('evlist')

      call readEventForEpileup(styleOut, rateThre&
         ! , flagWithApproxPatFrac&
         , chipXSize, chipYSize&
         , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
         , numFrames, rateImage, rateCube, maskImage, maskCube&
         , evCol, expStartTimeStamp&
         , rawKeyInfo, wcs, att, withRateImage&
         , startTime, finisTime)

    case('image')	! select case(styleIn)
      select case (styleOut)
      case('sky')
        call readImage2SkyForEpileup(rateThre&
           , flagIsSingleEventOnly, flagWithApproxPatFrac&
           , xiSize, yiSize&
           , instrumentId, cubePlaneIndices, pattFracSingle&
           , numFrames, rateImage, maskImageInt&
           , expStartTimeStamp, wcs, att, eVForCorrection)

      case('raw')
        allocate(rateCube(numCubePlanes, chipXSize, chipYSize))
        allocate(maskCube(numCubePlanes, chipXSize, chipYSize))
        call readImage2RawForEpileup(rateThre&
           , flagIsSingleEventOnly, flagWithApproxPatFrac&
           , xiSize, yiSize&
           , instrumentId, cubePlaneIndices, pattFracSingle&
           , numFrames, rateImage, rateCube, maskCube&
           , expStartTimeStamp, wcs, att, eVForCorrection)

      case default
         call error('badOutputStyleB4',"Should not happen.  Contact the code developer.")
      end select	! select case (styleOut) in case(styleIn)=('image')

    case default
      call error('badInputStyleB2',"Should not happen.  Contact the code developer.")
    end select	! select case(styleIn)

  end subroutine readDataForEpileup


  !---------------------------------------------------------------------
  subroutine readEventForEpileup(styleOut, rateThre&
     ! , flagWithApproxPatFrac&
     , chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     , numFrames, rateImage, rateCube, maskImage, maskCube&
     , evCol, expStartTimeStamp&
     , rawKeyInfo, wcs, att, withRateImage&
     , startTime, finisTime)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! read data from the input event.

    character(*), intent(in) :: styleOut	! OutputStyle
    real(single), intent(in) :: rateThre
    ! logical, intent(in) :: flagWithApproxPatFrac
    ! integer, intent(in) :: xiSize, yiSize
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    integer, pointer :: cubePlaneIndices(:)
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    real(single), pointer :: rateImage(:,:), rateCube(:,:,:), maskImage(:,:), numFrames(:)
    logical, pointer :: maskCube(:,:,:)
    ! logical, allocatable :: puMaskCube(:,:,:)
    type(evlistColumns), intent(in), optional :: evCol
    ! type(TableT), intent(in), optional :: eventTab
    real(double), intent(in), optional :: expStartTimeStamp
    type(keyInfoType), pointer, optional :: rawKeyInfo(:)	! intent(in)	! defined in eimchip2sky/cube_aux
    type(WcsT), intent(in), optional :: wcs	! type def in ssclib/wcs_aux
    type(SpacecraftAttitudeType), intent(in), optional :: att	! == binnedScAtt(attBinNum)%att
    logical, intent(in), optional :: withRateImage
    real(double), intent(in), optional :: startTime, finisTime	! unused if styleIn=='image'
     ! if this is <= 0, then the values defined in the event list are used.

    !-------------------------------------------------------------------
    ! Start:

    select case (styleOut)
    case('sky')
      call readEvent2SkyForEpileup(rateThre&
         ! , flagWithApproxPatFrac&
         , chipXSize, chipYSize&
         , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
         , numFrames, rateImage, maskImage&
         , evCol, expStartTimeStamp&
         , rawKeyInfo, wcs, att, withRateImage&
         , startTime, finisTime, rateCube)

    case('raw')
      call readEvent2RawForEpileup(rateThre&
         ! , flagWithApproxPatFrac&
         , chipXSize, chipYSize&
         , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
         , numFrames, rateCube, maskCube&
         , evCol, startTime, finisTime)

    case default
       call error('badOutputStyleB1',"Should not happen.  Contact the code developer.")
    end select	! select case (styleOut) in case(styleIn)=('evlist')

  end subroutine readEventForEpileup


  !---------------------------------------------------------------------
  subroutine readEvent2SkyForEpileupManyA(rateThre&
     ! , flagWithApproxPatFrac&
     , chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     , numFrames, rateImage, maskImage&
     , evCol, expStartTimeStamp&
     , rawKeyInfo, wcs, att, withRateImage&
     , startTime, finisTime, rateCubeRet)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! read data from the input event for the sky-coord output.
    !-- Note --
    ! numFrames specified here is probably binnedAttsNumFrames(:,attBinNum)

    real(single), intent(in) :: rateThre
    ! logical, intent(in) :: flagWithApproxPatFrac
    ! integer, intent(in) :: xiSize, yiSize
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    integer, pointer :: cubePlaneIndices(:)
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    real(single), pointer :: rateImage(:,:), maskImage(:,:), numFrames(:)
    ! real(single), pointer :: maskCubeReal(:,:,:)
    ! logical, allocatable :: puMaskCube(:,:,:)
    type(evlistColumns), intent(in) :: evCol
    ! type(TableT), intent(in), optional :: eventTab
    real(double), intent(in) :: expStartTimeStamp
    type(keyInfoType), pointer :: rawKeyInfo(:)	! intent(in)	! defined in eimchip2sky/cube_aux
    type(WcsT), intent(in) :: wcs	! type def in ssclib/wcs_aux
    type(SpacecraftAttitudeType), intent(in) :: att	! == binnedScAtt(attBinNum)%att
    logical, intent(in), optional :: withRateImage
    real(double), intent(in), optional :: startTime, finisTime
     ! if this is <= 0, then the values defined in the event list are used.
    real(single), pointer, optional :: rateCubeRet(:,:,:)	! intent(out)

    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT) :: infoFrameAtt	! Defined in epileupmask_shared

    !-------------------------------------------------------------------
    ! Start:

    infoCcdChips%instrumentId  = instrumentId  
    infoCcdChips%chipXSize     = chipXSize     
    infoCcdChips%chipYSize     = chipYSize     
    infoCcdChips%numCubePlanes = numCubePlanes 

    call allocateInfoCcdChipsT(infoCcdChips &
         , rawKeyInfoPtr=rawKeyInfo &
         , cubePlaneIndicesPtr=cubePlaneIndices &
         , pattFracSinglePtr=pattFracSingle)

    call allocateInfoFrameAttT(infoFrameAtt &
         , binnedScAttSize=1 &
         , numTotFramesPtr=numFrames)
    infoFrameAtt%expStartTimeStamp = expStartTimeStamp

    call initializeBinnedAtt(infoFrameAtt%binnedScAtt)	! in attcalc2/binned_att.
    infoFrameAtt%binnedScAtt(1)%isGood = .true.
    infoFrameAtt%binnedScAtt(1)%gtiDuration = 1.0d0 ! dummy value.
    infoFrameAtt%binnedScAtt(1)%att = att

    call readEvent2SkyForEpileupInfoC(rateThre&
         , infoCcdChips, infoFrameAtt &
         , rateImage, maskImage&
         , evCol &
         , 1 &	! ==attBinNum
         , wcs &
         , withRateImage&
         , startTime, finisTime, rateCubeRet)

    ! infoFrameAtt%isAllocatedNumTotFrames = .false.	! %numTotFrames must not be deallocated here, as it is linked to the external pointer.
    call deallocateInfoFrameAttT(infoFrameAtt, doNumTotFrames=.false.)

  end subroutine readEvent2SkyForEpileupManyA


  !---------------------------------------------------------------------
  subroutine readEvent2SkyForEpileupInfoC(rateThre&
       , infoCcdChips, infoFrameAtt &
                                ! , chipXSize, chipYSize&
                                ! , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
                                ! , numFrames &
       , rateImage, maskImage&
       , evCol & !, expStartTimeStamp&
       , attBinNum &
                                ! , rawKeyInfo &
       , wcs &
                                ! , att &
       , withRateImage&
       , startTime, finisTime, rateCubeRet)
    !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! read data from the input event for the sky-coord output.
    !-- Note --
    ! numFrames specified here is probably binnedAttsNumFrames(:,attBinNum)

    real(single), intent(in) :: rateThre
    type(infoCcdChipsT), intent(in) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT), intent(in) :: infoFrameAtt	! Defined in epileupmask_shared
    ! logical, intent(in) :: flagWithApproxPatFrac
    ! integer, intent(in) :: xiSize, yiSize
    ! integer(int16) :: chipXSize, chipYSize
    ! integer(int32) :: instrumentId, numCubePlanes
    ! integer, pointer :: cubePlaneIndices(:)
    ! real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    real(single), pointer :: rateImage(:,:), maskImage(:,:) ! , numFrames(:)
    real(single), pointer :: maskCubeReal(:,:,:)
    ! logical, allocatable :: puMaskCube(:,:,:)
    type(evlistColumns), intent(in) :: evCol
    integer, intent(in) :: attBinNum	! Index for binnedScAtt
    ! type(TableT), intent(in), optional :: eventTab
    ! real(double) :: expStartTimeStamp
    ! type(keyInfoType), pointer :: rawKeyInfo(:)	! intent(in)	! defined in eimchip2sky/cube_aux
    type(WcsT), intent(in) :: wcs	! type def in ssclib/wcs_aux
    type(SpacecraftAttitudeType) :: att	! == binnedScAtt(attBinNum)%att
    logical, intent(in), optional :: withRateImage
    real(double), intent(in), optional :: startTime, finisTime
    ! if this is <= 0, then the values defined in the event list are used.
    real(single), pointer, optional :: rateCubeRet(:,:,:)	! intent(out), initialised in this routine.

    real(single), pointer :: rateCube(:,:,:)
    ! integer :: numEvents, row
    ! real,           pointer :: evlistSkyX(:), evlistSkyY(:)
    ! integer(int32), allocatable :: eventCountCube(:,:,:)
    real(single), allocatable :: dummyAttWeights(:,:)
    logical :: withRateImageLocal

    !-------------------------------------------------------------------
    ! Start:

    ! instrumentId  = infoCcdChips%instrumentId
    ! chipXSize     = infoCcdChips%chipXSize
    ! chipYSize     = infoCcdChips%chipYSize
    ! numCubePlanes = infoCcdChips%numCubePlanes
    ! cubePlaneIndices  => infoCcdChips%cubePlaneIndices
    ! pattFracSingle => infoCcdChips%pattFracSingle
    ! rawKeyInfo     => infoCcdChips%rawKeyInfo

    ! numFrames         => infoFrameAtt%numTotFrames
    ! expStartTimeStamp =  infoFrameAtt%expStartTimeStamp
    att = infoFrameAtt%binnedScAtt(attBinNum)%att

    withRateImageLocal = .false.
    if (present(withRateImage)) withRateImageLocal = withRateImage

    call readEvent2RateCubeForEpileup(infoCcdChips, infoFrameAtt &
       , rateCube &
       , evCol, startTime, finisTime)	! rateCube(:,:,:) is allocated.
    ! call readEvent2RateCubeForEpileup( & ! flagWithApproxPatFrac&
    !      chipXSize, chipYSize&
    !    , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
    !    , numFrames, rateCube&
    !    , evCol, startTime, finisTime)
    ! call readEvent2RawForEpileup(rateThre& ...)	!! would make unnecessary maskCube.


    allocate(maskCubeReal(size(rateCube,1), size(rateCube,2), size(rateCube,3)))
    ! allocate(maskCubeReal(infoCcdChips%numCubePlanes, infoCcdChips%chipXSize, infoCcdChips%chipYSize))
    maskCubeReal = 0.0

    where (rateCube > rateThre)
      ! maskCube = .true.
      maskCubeReal = 1.0
    end where

    allocate(dummyAttWeights(size(rateCube,1), 1))	! Will be used to save both rateCube and maskCubeReal
    ! allocate(dummyAttWeights(infoCcdChips%numCubePlanes, 1))
    dummyAttWeights = 1.0

    call message(appMsg, Verbose, 'Converting mask images into Sky coordinates.')
    call rebinChipImagesToSky(maskCubeReal&
       , infoCcdChips%rawKeyInfo%ccdNum, wcs, (/att/)&	! ==(/binnedScAtt(attBinNum)%att/)&
       , dummyAttWeights, infoFrameAtt%expStartTimeStamp, 'rate'&
       , .false., maskImage)
     ! in eimchip2sky/eimchip2sky_aux.
     ! Note: Here only one bin in attitude is used, so (/att/) does the job.
     !  rebinChipImagesToSky() is meant to "rebin" the input over the given
     !  attitude.  In this case only we need is just the simple sky-coordinate
     !  conversion, but not further rebinning.
     ! Note: In rebinChipImagesToSky() in eimchip2sky_aux, the variable
     !  "binnedScAtt" is different from the same name ones used elsewhere!
     !  There, it actually refers to binnedScAtt(:)%att in the ordinary notation.

    deallocate(maskCubeReal)

    if (withRateImageLocal) then
      call message(appMsg, Verbose, 'Converting rate images into Sky coordinates.')
      call rebinChipImagesToSky(rateCube&
         , infoCcdChips%rawKeyInfo%ccdNum, wcs, (/att/)&	! ==(/binnedScAtt(attBinNum)%att/)&
         , dummyAttWeights, infoFrameAtt%expStartTimeStamp, 'rate'&
         , .false., rateImage)
    end if

    deallocate(dummyAttWeights)

    if (present(rateCubeRet)) then
      rateCubeRet => rateCube
    else
      deallocate(rateCube)
    end if

      ! How does this call work to make a mask? Ok, we have sent to
      ! rebinChipImagesToSky a cube in which a value of 0 represents a
      ! CCD column for which the pileup criterion has not been exceeded
      ! at any pixel during the present attitude interval, and a value
      ! of 1 represents the contrary. This cube is reprojected to sky
      ! coordinates and added to outImage. At the end of the process
      ! outImage will therefore have 0-valued and non-0-valued pixels.
      ! The 0 values represent those places on the sky where the camera
      ! has never had a piled-up column; >0 values represent the contrary.
      ! These >0 areas of outImage are translated into a logical mask
      ! of the piled-up areas during the final call to saveImage.

  end subroutine readEvent2SkyForEpileupInfoC


  !---------------------------------------------------------------------
  subroutine readEvent2RawForEpileupManyA(rateThre&
     ! , flagWithApproxPatFrac&
     , chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     , numFrames, rateCube, maskCube&
     , evCol, startTime, finisTime)
  !---------------------------------------------------------------------
  ! Make sure to deallocate(maskCube) .

    implicit none 

    !-- Description --
    ! read data from the input event for chipcube output.
    ! Note that maskCube is allocated in this routine.

    real(single), intent(in) :: rateThre
    ! logical, intent(in) :: flagWithApproxPatFrac
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    integer, pointer :: cubePlaneIndices(:)	! intent(in)
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern	! intent(in)
    real(single), pointer :: numFrames(:)	! intent(in)
    real(single), pointer :: rateCube(:,:,:)	! intent(out)
    logical, pointer :: maskCube(:,:,:)	! intent(out)
    ! logical, allocatable :: puMaskCube(:,:,:)
    type(evlistColumns), intent(in) :: evCol
    real(double), intent(in), optional :: startTime, finisTime	! unused if styleIn=='image'
     ! if this is <= 0, then the values defined in the event list are used.

    !-------------------------------------------------------------------
    ! Start:

    allocate(maskCube(numCubePlanes, chipXSize, chipYSize))
    maskCube = .false.

    call readEvent2RateCubeForEpileup( & ! flagWithApproxPatFrac&
         chipXSize, chipYSize&
       , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
       , numFrames, rateCube&
       , evCol, startTime, finisTime)

    where ( rateCube > rateThre )
      maskCube = .true.
    end where

  end subroutine readEvent2RawForEpileupManyA


  !---------------------------------------------------------------------
  subroutine readEvent2RawForEpileupInfoC(rateThre&
     , infoCcdChips, infoFrameAtt &
     ! , flagWithApproxPatFrac&
     ! , chipXSize, chipYSize&
     ! , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     ! , numFrames &
     , rateCube, maskCube&
     , evCol, startTime, finisTime)
  !---------------------------------------------------------------------
  ! Make sure to deallocate(maskCube) .

    implicit none 

    !-- Description --
    ! read data from the input event for chipcube output.
    ! Note that maskCube is allocated in this routine.

    real(single), intent(in) :: rateThre
    type(infoCcdChipsT), intent(in) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT), intent(in) :: infoFrameAtt	! Defined in epileupmask_shared
    ! logical, intent(in) :: flagWithApproxPatFrac
    ! integer(int16), intent(in) :: chipXSize, chipYSize
    ! integer(int32), intent(in) :: instrumentId, numCubePlanes
    ! integer, pointer :: cubePlaneIndices(:)	! intent(in)
    ! real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern	! intent(in)
    ! real(single), pointer :: numFrames(:)	! intent(in)
    real(single), pointer :: rateCube(:,:,:)	! intent(out)
    logical, pointer :: maskCube(:,:,:)	! intent(out)
    ! logical, allocatable :: puMaskCube(:,:,:)
    type(evlistColumns), intent(in) :: evCol
    real(double), intent(in), optional :: startTime, finisTime	! unused if styleIn=='image'
     ! if this is <= 0, then the values defined in the event list are used.

    !-------------------------------------------------------------------
    ! Start:

    allocate(maskCube(infoCcdChips%numCubePlanes &
                    , infoCcdChips%chipXSize &
                    , infoCcdChips%chipYSize))
    maskCube = .false.

    call readEvent2RateCubeForEpileup(infoCcdChips, infoFrameAtt &
     ! , flagWithApproxPatFrac&
     , rateCube&
     , evCol, startTime, finisTime)
    ! call readEvent2RateCubeForEpileup(! flagWithApproxPatFrac&
    !      chipXSize, chipYSize&
    !    , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
    !    , numFrames, rateCube&
    !    , evCol, startTime, finisTime)

    if (size(rateCube, 1) > 0) then	! Otherwise rateCube is not defined, probably because "No events in this attitude bin."
      where ( rateCube > rateThre )
        maskCube = .true.
      end where
    end if

!print *, 'maxval(eventCountCube)=', maxval(eventCountCube)

    ! ! Now count the probability:
    ! !  Algorithm: Final_puProbCube = 1.0 - exp(-Image/Frame)
    ! !              which is calculated later.

    ! allocate(puProbCube(numCubePlanes, chipXSize, chipYSize))
    ! puProbCube = 0.0
    ! do cubePlaneNum = 1, numCubePlanes
!print *, 'numFrames(',cubePlaneNum,')=', numFrames(cubePlaneNum)
    !   if (numFrames(cubePlaneNum) <= 0.0) then
    !     if (any(eventCountCube(cubePlaneNum,:,:) .ne. 0))&
    !       call warning('eventsOutOfGoodFrameTime',"")
    !   else
    !     puProbCube(cubePlaneNum,:,:) = eventCountCube(cubePlaneNum,:,:)&
    !       / numFrames(cubePlaneNum)
    !   end if
    ! end do

!!! from eootepileupmask
!           puProbPlane = eventCountCube(cubePlaneNum,:,:)&
!             / binnedAttsNumFrames(cubePlaneNum, attBinNum)
! 
!           ! At this point the array puProbPlane holds not the PU
!           ! probability but just the average number of events per
!           ! frame for this pixel, CCD and attitude interval. The
!           ! probability is calculated at the next step:
! 
!           puProbPlane = 1.0 - exp(-puProbPlane)
!           maxPileupProb = max(maxPileupProb, maxval(puProbPlane))

  end subroutine readEvent2RawForEpileupInfoC


  !---------------------------------------------------------------------
  subroutine readEvent2RateCubeForEpileupMA( &! flagWithApproxPatFrac&
       chipXSize, chipYSize&
     , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     , numFrames, rateCube&
     , evCol, startTime, finisTime)
  !---------------------------------------------------------------------
  ! Make sure to deallocate(rateCube) after use.

    implicit none 

    !-- Description --
    ! read data from the input event, and output a rate chipcube.

    ! logical, intent(in) :: flagWithApproxPatFrac
    integer(int16), intent(in) :: chipXSize, chipYSize
    integer(int32), intent(in) :: instrumentId, numCubePlanes
    integer,      pointer :: cubePlaneIndices(:)
    real(double), pointer :: pattFracSingle(:)	! grade pattern
    real(single), pointer :: rateCube(:,:,:), numFrames(:)
    type(evlistColumns), intent(in) :: evCol
    real(double),   intent(in), optional :: startTime, finisTime	! unused if styleIn=='image'
     ! if this is <= 0, then the values defined in the event list are used.

    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT) :: infoFrameAtt	! Defined in epileupmask_shared

    !-------------------------------------------------------------------
    ! Start:

    infoCcdChips%chipXSize     = chipXSize
    infoCcdChips%chipYSize     = chipYSize
    infoCcdChips%instrumentId  = instrumentId
    infoCcdChips%numCubePlanes = numCubePlanes
    call allocateInfoCcdChipsT(infoCcdChips &
       , cubePlaneIndicesPtr=cubePlaneIndices, pattFracSinglePtr=pattFracSingle)
    call allocateInfoFrameAttT(infoFrameAtt, numTotFramesPtr=numFrames)

    call readEvent2RateCubeForEpileupIC(infoCcdChips, infoFrameAtt &
       , rateCube &
       , evCol, startTime, finisTime)

    call deallocateInfoCcdChipsT(infoCcdChips &
       , doCubePlaneIndices=.false., doPattFracSingle=.false.)	! Practically do nothing.
    call deallocateInfoFrameAttT(infoFrameAtt, doNumTotFrames=.false.)	! Practically do nothing.

  end subroutine readEvent2RateCubeForEpileupMA

  !---------------------------------------------------------------------
  subroutine readEvent2RateCubeForEpileupIC(infoCcdChips, infoFrameAtt &
     ! , flagWithApproxPatFrac&
     ! , chipXSize, chipYSize&
     ! , instrumentId, numCubePlanes, cubePlaneIndices, pattFracSingle&
     ! , numFrames &
     , rateCube &
     , evCol, startTime, finisTime)
  !---------------------------------------------------------------------
  ! Make sure to deallocate(rateCube) after use.

    implicit none 

    !-- Description --
    ! read data from the input event, and output a rate chipcube.

    type(infoCcdChipsT), intent(in) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT), intent(in) :: infoFrameAtt	! Defined in epileupmask_shared
    ! logical, intent(in) :: flagWithApproxPatFrac
    integer(int16) :: chipXSize, chipYSize
    integer(int32) :: instrumentId, numCubePlanes
    integer, pointer :: cubePlaneIndices(:)
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    real(single), pointer :: rateCube(:,:,:), numFrames(:)
    type(evlistColumns), intent(in) :: evCol
    real(double), intent(in), optional :: startTime, finisTime	! unused if styleIn=='image'
     ! if this is <= 0, then the values defined in the event list are used.

    integer :: row
    integer(int32) :: cubePlaneNum
    ! integer(int32), allocatable :: eventCountCube(:,:,:)
    integer(int16) :: xi, yi
    real(double) :: corrFactor, alpha1
    ! real(double) :: startTimeRev, finisTimeRev	! ==startTime if present && >0.
    integer :: startRow, finisRow
    integer :: maxPattFracSingle
 
    !-------------------------------------------------------------------
    ! Start:

    instrumentId  = infoCcdChips%instrumentId
    chipXSize     = infoCcdChips%chipXSize
    chipYSize     = infoCcdChips%chipYSize
    numCubePlanes = infoCcdChips%numCubePlanes
    cubePlaneIndices  => infoCcdChips%cubePlaneIndices
    pattFracSingle    => infoCcdChips%pattFracSingle

    numFrames         => infoFrameAtt%numTotFrames

    maxPattFracSingle = ubound(pattFracSingle,1)

    call readEventSharedForEpileup(evCol, startRow, finisRow&
       , startTime=startTime, finisTime=finisTime)
       ! , startTimeRev, finisTimeRev, startTime, finisTime)

    if (startRow > finisRow) then
      if (present(startTime) .and. present(finisTime)) then
        write(errStr,'(a,es16.10,"/",es16.10,a,i8,"/",i8,").")') &
           "No events in this attitude bin: Start/EndTime=(" &
           , startTime, finisTime, "), Start/EndRow=(" &
           , startRow,  finisRow
      else
        write(errStr,'(a,i8,"/",i8,").")') &
           "No events in this attitude bin: Start/EndRow=(" &
           , startRow,  finisRow
      end if
      call message(appMsg, Sparse)
! call error('abc')
      ! call message(AppMsg, verbose, "No events in this attitude bin.")
      allocate(rateCube(0,0,0))
      ! if (.not. associated(rateCube))  rateCube = 0
      return
    end if

    ! Now count the events:

    ! allocate(eventCountCube(numCubePlanes,chipXSize,chipYSize))
    ! eventCountCube = 0
    allocate(rateCube(numCubePlanes, chipXSize, chipYSize))
    rateCube = 0.0

    do row = startRow, finisRow
      errStr = sprintEvery(row, finisRow, fraction=0.2) ! in ssclib/ssc_misc.

      if (errStr .ne. '') call message(appMsg, Verbose)

      if (evCol%pattern(row) .ne. 0) cycle	! Pat=0 (single-pixel event)

      cubePlaneNum = cubePlaneIndices(evCol%ccdNum(row))	! numNode==1 is assumed	; Rewrite with getCcdNodeNum() in epic_aux!!!
      xi = evCol%rawX(row)
      yi = evCol%rawY(row)

      ! eventCountCube(cubePlaneNum, xi, yi) = eventCountCube(cubePlaneNum, xi, yi) + 1

    ! if (WithApproxPatFrac) then
      if (evCol%pi(row) > maxPattFracSingle) then
        write(errStr,'("PI(=",i5,") is beyond the range for row=",i8,".")') evCol%pi(row), row+1
        call warning('piTooLarge')
        cycle
      end if

      alpha1 = pattFracSingle(evCol%pi(row))
    ! else
    !   call error('notSupported')
    ! end if
      if (alpha1 <= 0) then
        call warning('zeroPatternFraction','Pattern fraction for a given energy is zero.  Skip.')
        cycle
      end if

      corrFactor = StandardAlpha1(instrumentId) / alpha1
         ! (0.778 / alpha1) for MOS
      rateCube(cubePlaneNum, xi, yi) = rateCube(cubePlaneNum, xi, yi) &
         + corrFactor / numFrames(cubePlaneIndices(evCol%ccdNum(row)))
!print *,'Corr/numFrame=',corrFactor/numFrames(cubePlaneIndices(evCol%ccdNum(row)))

    end do		! do row = startRow, finisRow

    !! Check about Bad frames.
    do cubePlaneNum = 1, numCubePlanes
!print *, 'numFrames(',cubePlaneNum,')=', numFrames(cubePlaneNum)
      if (numFrames(cubePlaneNum) <= 0.0) then
        ! if (any(eventCountCube(cubePlaneNum,:,:) .ne. 0))&
        !    call warning('eventsOutOfGoodFrameTime',"")
        if (any(rateCube(cubePlaneNum,:,:) .ne. 0)) then
          call warning('eventsOutOfGoodFrameTime',"")
          rateCube(cubePlaneNum,:,:) = 0.0	! rateCube for that chip is set to be 0.
        end if
      end if
    end do

  end subroutine readEvent2RateCubeForEpileupIC


  !---------------------------------------------------------------------
  subroutine readEventSharedForEpileup(evCol, startRow, finisRow&
     , startTimeRev, finisTimeRev, startTime, finisTime)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! Shared routine for readEvent2*ForEpileup.
    ! Returns start/finisRow and/or start/finisTimeRev (the realistic
    ! start/finisTime constrained with the event file),
    ! taking account of evCol and optional start/finisTime
    ! (the latter can be arbitrary, namely possibly well beyond the OBS times).
    ! evCol must be present if you want any significant
    ! result from this subroutine.

    type(evlistColumns), intent(in), optional :: evCol
    ! type(TableT), intent(in), optional :: eventTab
    integer,      intent(out), optional :: startRow, finisRow
    real(double), intent(out), optional :: startTimeRev, finisTimeRev	! ==startTime if present && >0.
    real(double), intent(in),  optional :: startTime, finisTime
     ! if this is <= 0, then the values defined in the event list are used.

    integer :: startRowLocal, finisRowLocal
    real(double) :: startTimeRevLocal, finisTimeRevLocal

    !-------------------------------------------------------------------
    ! Start:

    if (.not. present(evCol)) return

    startRowLocal = 0
    startTimeRevLocal = evCol%times(0)
    if (present(startTime)) then
      if (startTime > 0) then	! If startTime is negative, default is given.
        startRowLocal = smallestTrueIndex(evCol%times >= startTime) - 1	! smallestTrueIndex() defined in ssclib/array_utils.
        startTimeRevLocal = startTime
      end if
    else
    end if
    if (present(startRow)) startRow = startRowLocal
    if (present(startTimeRev)) startTimeRev = startTimeRevLocal

    ! numEvents = numberOfRows(eventTab)
    ! finisRow = numEvents-1
    finisRowLocal = size(evCol%times) - 1
    finisTimeRevLocal = evCol%times(finisRowLocal)
    if (present(finisTime)) then
      if (finisTime > 0) then	! If finisTime is negative, default is given.
        finisRowLocal = largestTrueIndex(evCol%times <= finisTime) - 1	! largestTrueIndex() defined in ssclib/array_utils.
        finisTimeRevLocal = finisTime
      end if
    end if
    if (present(finisRow)) finisRow = finisRowLocal
    if (present(finisTimeRev)) finisTimeRev = finisTimeRevLocal

  end subroutine readEventSharedForEpileup


  !---------------------------------------------------------------------
  subroutine readImage2SkyForEpileupManyA(rateThre&
     , flagIsSingleEventOnly, flagWithApproxPatFrac&
     , xiSize, yiSize&
     , instrumentId, cubePlaneIndices, pattFracSingle&
     , numFrames, rateImage, maskImageInt&
     , expStartTimeStamp, wcsin, medianScAttitude&
     , eVForCorrection)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! read data from the input image for the sky-coord output.

    real(single), intent(in) :: rateThre
    logical, intent(in) :: flagIsSingleEventOnly, flagWithApproxPatFrac
    integer, intent(in) :: xiSize, yiSize
    integer(int32), intent(in) :: instrumentId
    integer, pointer :: cubePlaneIndices(:)
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    real(single), pointer :: rateImage(:,:), numFrames(:)
    integer(int8), pointer :: maskImageInt(:,:)
    ! logical, allocatable :: puMaskCube(:,:,:)
    real(double), intent(in) :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(in) :: medianScAttitude
    type(WcsT), intent(in) :: wcsin	! type def in ssclib/wcs_aux
    real, intent(in) :: eVForCorrection	! Energy(eV) to be used in correction.

    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT) :: infoFrameAtt	! Defined in epileupmask_shared

    !-------------------------------------------------------------------
    ! Start:

    infoCcdChips%instrumentId = instrumentId
    call allocateInfoCcdChipsT(infoCcdChips &
       , cubePlaneIndicesPtr=cubePlaneIndices, pattFracSinglePtr=pattFracSingle)

    infoFrameAtt%expStartTimeStamp = expStartTimeStamp
    infoFrameAtt%medianScAttitude  = medianScAttitude
    call allocateInfoFrameAttT(infoFrameAtt, numTotFramesPtr=numFrames)

    call readImage2SkyForEpileupInfoC(rateThre&
       , flagIsSingleEventOnly, flagWithApproxPatFrac&
       , xiSize, yiSize&
       , infoCcdChips, infoFrameAtt &
       , rateImage, maskImageInt&
       , wcsin &
       , eVForCorrection)

    call deallocateInfoCcdChipsT(infoCcdChips &
       , doCubePlaneIndices=.false., doPattFracSingle=.false.)	! Practically do nothing.
    call deallocateInfoFrameAttT(infoFrameAtt, doNumTotFrames=.false.)	! Practically do nothing.

  end subroutine readImage2SkyForEpileupManyA


  !---------------------------------------------------------------------
  subroutine readImage2SkyForEpileupInfoC(rateThre&
     , flagIsSingleEventOnly, flagWithApproxPatFrac&
     , xiSize, yiSize&
     , infoCcdChips, infoFrameAtt &
     ! , instrumentId, cubePlaneIndices, pattFracSingle&
     ! , numFrames &
     , rateImage, maskImageInt&
     ! , expStartTimeStamp &
     , wcsin & ! , medianScAttitude&
     , eVForCorrection)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! read data from the input image for the sky-coord output.

    real(single), intent(in) :: rateThre
    logical, intent(in) :: flagIsSingleEventOnly, flagWithApproxPatFrac
    integer, intent(in) :: xiSize, yiSize
    type(infoCcdChipsT), intent(in) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT), intent(in) :: infoFrameAtt	! Defined in epileupmask_shared
    integer(int32) :: instrumentId
    integer, pointer :: cubePlaneIndices(:)
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    real(single), pointer :: rateImage(:,:), numFrames(:)
    integer(int8), pointer :: maskImageInt(:,:)
    ! logical, allocatable :: puMaskCube(:,:,:)
    real(double) :: expStartTimeStamp
    type(SpacecraftAttitudeType) :: medianScAttitude
    type(WcsT), intent(in) :: wcsin	! type def in ssclib/wcs_aux
    real, intent(in) :: eVForCorrection	! Energy(eV) to be used in correction.

    integer(int16) :: xi, yi
    real(double) :: corrFactor
    real(double) :: tanX, tanY, detX, detY, rawX, rawY
    logical(bool) :: isOffChip
    integer(int8) :: ccdNum
 
    !-------------------------------------------------------------------
    ! Start:

    instrumentId  = infoCcdChips%instrumentId
    ! chipXSize     = infoCcdChips%chipXSize
    ! chipYSize     = infoCcdChips%chipYSize
    cubePlaneIndices  => infoCcdChips%cubePlaneIndices
    pattFracSingle    => infoCcdChips%pattFracSingle

    numFrames         => infoFrameAtt%numTotFrames
    expStartTimeStamp =  infoFrameAtt%expStartTimeStamp
    medianScAttitude  =  infoFrameAtt%medianScAttitude

    maskImageInt = 0

    call prepareForReadImageForEpileup(flagIsSingleEventOnly, flagWithApproxPatFrac&
     , instrumentId, pattFracSingle&
     , eVForCorrection&
     , corrFactor)

    ! call readArrayData(inputSet, rateImage)
    ! ! in ssclib/dal_aux.

    !! Calculate the corrected single pixel event rate, and the mask.
    write(errStr,'(a,i5," x ",i5,").")') &
       'Calculating the corrected single pixel event rate for (', xiSize, yiSize
    call message(appMsg, Verbose)
    do xi=1, xiSize
      errStr = sprintEvery(int(xi), xiSize, fraction=0.2) ! in ssclib/ssc_misc.
      if (errStr .ne. '') call message(appMsg, Verbose)

      do yi=1, yiSize
        !! Get the chip number.
        call pixelsToTan(real(xi),real(yi),wcsin,tanX,tanY)	! ssclib/wcs_aux
        call tanToDet(tanX, tanY, wcsin, medianScAttitude&
             , expStartTimeStamp, detX, detY)
        call detToRaw(detX, detY, rawX, rawY, isOffChip, ccdNum)	! ccdNum is the only parameter required.

        !! calculate rateImage
        rateImage(xi,yi) = rateImage(xi,yi) / numFrames(cubePlaneIndices(ccdNum))
        ! Assume (numFrames(numNode=1) == numFrames(numNode=2))

      end do
    end do
    call message(appMsg, Verbose, 'Finished loop.')

    rateImage = rateImage * corrFactor	! 2d-Array times a scalar operation

    !! calculate maskImage
    where (rateImage > rateThre) maskImageInt = 1

  end subroutine readImage2SkyForEpileupInfoC


  !---------------------------------------------------------------------
  subroutine readImage2RawForEpileupManyA(rateThre&
     , flagIsSingleEventOnly, flagWithApproxPatFrac&
     , xiSize, yiSize&
     , instrumentId, cubePlaneIndices, pattFracSingle&
     , numFrames, rateImage, rateCube, maskCube&
     , expStartTimeStamp, wcsin, medianScAttitude&
     , eVForCorrection)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! read data from the input image, and return rate & mask cube (Raw).
    !-- NOTE --
    ! The input rateCube and maskCube must be pre-allocated.
    ! maskCube is initialised in this subroutine.

    real(single),   intent(in) :: rateThre
    logical,        intent(in) :: flagIsSingleEventOnly, flagWithApproxPatFrac
    integer,        intent(in) :: xiSize, yiSize
    integer(int32), intent(in) :: instrumentId
    integer,      pointer :: cubePlaneIndices(:)	! intent(in)
    real(double), pointer :: pattFracSingle(:)	! grade pattern	! intent(in)
    real(single), pointer :: numFrames(:) &	! intent(in)
                           , rateImage(:,:) &	! intent(in)
                           , rateCube(:,:,:)	! intent(inout), should be allocated.
    logical,      pointer :: maskCube(:,:,:)	! intent(inout), should be allocated, but will be initialised in this routine.
    real(double), intent(in) :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(in) :: medianScAttitude
    type(WcsT),   intent(in) :: wcsin	! type def in ssclib/wcs_aux
    real, intent(in) :: eVForCorrection	! Energy(eV) to be used in correction.

    type(infoCcdChipsT) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT) :: infoFrameAtt	! Defined in epileupmask_shared

    !-------------------------------------------------------------------
    ! Start:

    infoCcdChips%instrumentId = instrumentId
    call allocateInfoCcdChipsT(infoCcdChips &
       , cubePlaneIndicesPtr=cubePlaneIndices, pattFracSinglePtr=pattFracSingle)

    infoFrameAtt%expStartTimeStamp = expStartTimeStamp
    infoFrameAtt%medianScAttitude  = medianScAttitude
    call allocateInfoFrameAttT(infoFrameAtt, numTotFramesPtr=numFrames)

    call readImage2RawForEpileupInfoC(rateThre&
       , flagIsSingleEventOnly, flagWithApproxPatFrac&
       , xiSize, yiSize&
       , infoCcdChips, infoFrameAtt &
       , rateImage, rateCube, maskCube&
       , wcsin &
       , eVForCorrection)

    call deallocateInfoCcdChipsT(infoCcdChips &
       , doCubePlaneIndices=.false., doPattFracSingle=.false.)	! Practically do nothing.
    call deallocateInfoFrameAttT(infoFrameAtt, doNumTotFrames=.false.)	! Practically do nothing.

  end subroutine readImage2RawForEpileupManyA


  !---------------------------------------------------------------------
  subroutine readImage2RawForEpileupInfoC(rateThre&
     , flagIsSingleEventOnly, flagWithApproxPatFrac&
     , xiSize, yiSize&
     , infoCcdChips, infoFrameAtt &
     ! , instrumentId, cubePlaneIndices, pattFracSingle&
     ! , numFrames &
     , rateImage, rateCube, maskCube&
     ! , expStartTimeStamp &
     , wcsin & ! , medianScAttitude&
     , eVForCorrection)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! read data from the input image, and return rate & mask cube (Raw).
    !-- NOTE --
    ! The input rateCube and maskCube must be pre-allocated.
    ! maskCube is initialised in this subroutine.

    real(single),   intent(in) :: rateThre
    logical,        intent(in) :: flagIsSingleEventOnly, flagWithApproxPatFrac
    integer,        intent(in) :: xiSize, yiSize
    type(infoCcdChipsT), intent(in) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT), intent(in) :: infoFrameAtt	! Defined in epileupmask_shared
    integer(int32) :: instrumentId
    integer,      pointer :: cubePlaneIndices(:)	! intent(in)
    real(double), pointer :: pattFracSingle(:)	! grade pattern	! intent(in)
    real(single), pointer :: numFrames(:) &	! intent(in)
                           , rateImage(:,:) &	! intent(in)
                           , rateCube(:,:,:)	! intent(inout), should be allocated.
    logical,      pointer :: maskCube(:,:,:)	! intent(inout), should be allocated, but will be initialised in this routine.
    real(double) :: expStartTimeStamp
    type(SpacecraftAttitudeType) :: medianScAttitude
    type(WcsT),   intent(in) :: wcsin	! type def in ssclib/wcs_aux
    real, intent(in) :: eVForCorrection	! Energy(eV) to be used in correction.

    integer :: weightCube(size(rateCube,1),size(rateCube,2),size(rateCube,3))
    integer(int16) :: xi, yi, x, y, cn
    real(double) :: corrFactor
    real(double) :: tanX, tanY, detX, detY, rawX, rawY
    logical(bool) :: isOffChip
    integer(int8) :: ccdNum, maxCcdNum
 
    !-------------------------------------------------------------------
    ! Start:

    instrumentId  = infoCcdChips%instrumentId
    ! chipXSize     = infoCcdChips%chipXSize
    ! chipYSize     = infoCcdChips%chipYSize
    cubePlaneIndices  => infoCcdChips%cubePlaneIndices
    pattFracSingle    => infoCcdChips%pattFracSingle

    numFrames         => infoFrameAtt%numTotFrames
    expStartTimeStamp =  infoFrameAtt%expStartTimeStamp
    medianScAttitude  =  infoFrameAtt%medianScAttitude

    rateCube(:,:,:) = 0.0
    weightCube(:,:,:) = 0
    ! maskCube(:,:,:) = .false.	! Initialised later
    maxCcdNum = ubound(rateCube,1)

    call prepareForReadImageForEpileup(flagIsSingleEventOnly, flagWithApproxPatFrac&
     , instrumentId, pattFracSingle&
     , eVForCorrection&
     , corrFactor)

    ! call readArrayData(inputSet, rateImage)
    ! ! in ssclib/dal_aux.

    !! Calculate the corrected single pixel event rate, and the mask.
    write(errStr,'(a,i5," x ",i5,").")') &
       'Calculating the corrected single pixel event rate for (', xiSize, yiSize
    call message(appMsg, Verbose)
    do xi=1, xiSize
      errStr = sprintEvery(int(xi), xiSize, fraction=0.2) ! in ssclib/ssc_misc.
      if (errStr .ne. '') call message(appMsg, Verbose)

      do yi=1, yiSize
        if (rateImage(xi,yi) == 0) cycle	! If the image value is zero, then the following coordinate conversion is a waste of processing time.

        !! Get the chip number.
        call pixelsToTan(real(xi),real(yi),wcsin,tanX,tanY)	! ssclib/wcs_aux
        call tanToDet(tanX, tanY, wcsin, medianScAttitude&
           , expStartTimeStamp, detX, detY)
        call detToRaw(detX, detY, rawX, rawY, isOffChip, ccdNum)
 
        if (isOffChip) then
          write(errStr,'("Off chip event at (",i4,",",i4,") in the image.")') xi, yi
          call warning('offChipEvent')
          cycle
        else if (isOffChip .or. (cubePlaneIndices(ccdNum) > maxCcdNum)) then
          write(errStr,'("CCD number is unexpectedly large at (",i4,",",i4,").")') xi, yi
          call error('unexpectedCcdNumber')
        else
          !! calculate rateCube
          cn = cubePlaneIndices(ccdNum)
          x  = nint(rawX)
          y  = nint(rawY)
          rateCube(  cn, x, y) = rateCube(  cn, x, y) + rateImage(xi,yi)
          weightCube(cn, x, y) = weightCube(cn, x, y) + 1
          ! NOTE: Chances are rateImage(xi,yi) for a different (xi,yi)
          !  corresponds to rateCube(cn,rawX,rawY) for the same (rawX,rawY),
          !  that is, (xi_1,yi_1) => (rawX_1, rawY_1)
          !           (xi_2,yi_2) => (rawX_1, rawY_1),
          !  particularly when rateImage has a small bin size.
          !  Therefore it is necessary to divide it by weight later
          !  to obtain the appropriate rateCube(cn,RAWX,RAWY).
          !
          ! NOTE: The case where the image has a larger bin-size is
          !  not considered.  In that case, ideally, all the potential
          !  pixels in (RAWX,RAWY) that correspond to a pixel in the image
          !  should be flagged (masked).  However at the moment,
          ! only one representative RAWX/Y pixel per Sky-image pixel is 
          ! flagged.

          ! rateCube(cubePlaneIndices(ccdNum), nint(rawX), nint(rawY))&
          !    = rateImage(xi,yi) * corrFactor / numFrames(cubePlaneIndices(ccdNum))
        end if
 
        ! !! calculate maskCube
        ! if ( rateCube(cubePlaneIndices(ccdNum), nint(rawX), nint(rawY)) > rateThre ) then
        !   maskCube(cubePlaneIndices(ccdNum), nint(rawX), nint(rawY)) = .true.
        ! end if
      end do
    end do
    call message(appMsg, Verbose, 'Finished loop.')

    !! adjuste rateCube
    where (weightCube == 0)
      weightCube = 1
    end where
    do cn = lbound(cubePlaneIndices,1), ubound(cubePlaneIndices,1)
      rateCube(cn,:,:) = rateCube(cn,:,:) * corrFactor / numFrames(cn) / weightCube(cn,:,:) 
    end do

    !! calculate maskCube
    maskCube = .false.
    where ( rateCube > rateThre )
      maskCube = .true.
    end where

  end subroutine readImage2RawForEpileupInfoC


  !---------------------------------------------------------------------
  subroutine prepareForReadImageForEpileup(flagIsSingleEventOnly, flagWithApproxPatFrac&
     , instrumentId, pattFracSingle&
     , eVForCorrection&
     , corrFactor)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! prepare some essentials for subroutine readImageForEpileup
    ! Basically set corrFactor, wcs, expStartTimeStamp, medianScAttitude.

    logical, intent(in) :: flagIsSingleEventOnly, flagWithApproxPatFrac
    integer(int32), intent(in) :: instrumentId
    real(kind=double), dimension(:), pointer :: pattFracSingle	! grade pattern
    ! logical, allocatable :: puMaskCube(:,:,:)
    real, intent(in) :: eVForCorrection	! Energy(eV) to be used in correction.
    real(double), intent(out) :: corrFactor

    integer(int32) :: i
    real(double) :: alpha1
    real(single), dimension(1) :: arReal
    real(single), dimension(:), pointer :: poPiReal
    integer :: maxPattFracSingle
 
    !-------------------------------------------------------------------
    ! Start:

    maxPattFracSingle = ubound(pattFracSingle,1)
    !! Estimate the correction factor for (ideal = no-pile-up) single-pixel event rate.
    if (flagIsSingleEventOnly) then
      if (flagWithApproxPatFrac) then
        allocate(poPiReal(1))
        arReal(1)=eVForCorrection
        call CAL_eVToPi(arReal, poPiReal)
        i = ceiling(poPiReal(1))
        deallocate(poPiReal)

        if (i > maxPattFracSingle) then
          write(errStr,'("The energy specified is beyond the PI range.  Use PI=",i5,", instead.")')&
             maxPattFracSingle
          call warning('energyTooLarge')
          alpha1 = pattFracSingle(maxPattFracSingle)
        else
          alpha1 = pattFracSingle(i)
        end if
      else
        call error('notSupported', 'flagWithApproxPatFrac must be TRUE at the moment.')
      end if

      if (alpha1 <= 0) then
        call error('zeroPatternFraction','Pattern fraction for the given energy is zero.')
      end if
      corrFactor = StandardAlpha1(instrumentId) / alpha1
      	! (0.778 / alpha1) for MOS
    else
      corrFactor = StandardAlpha1(instrumentId)
    end if

  end subroutine prepareForReadImageForEpileup


  !---------------------------------------------------------------------
  subroutine getAttSkyCoordForEpileupManyA(skyTemplateImageSetName, attStyleKeyWord, binnedAttFileName&
       , wcs, xiSize, yiSize&
       , binnedScAtt & !, numAttBins
       , expStartTimeStamp, medianScAttitude)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! get Attitude data, and sky coordinate data for output image.

    character(*), intent(in) :: skyTemplateImageSetName, attStyleKeyWord, binnedAttFileName
    type(WcsT), intent(out) :: wcs	! type def in ssclib/wcs_aux
    integer, intent(out) :: xiSize, yiSize
    type(attBinType), pointer :: binnedScAtt(:)	! type def in attcalc2/binned_att.
    ! integer, intent(out) :: numAttBins
    real(double), intent(out) :: expStartTimeStamp
    type(SpacecraftAttitudeType), intent(out) :: medianScAttitude

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, "Reading data from template image ("//trim(skyTemplateImageSetName)//").")
    call getSkyCoordFrame(skyTemplateImageSetName&
       , wcs, xiSize=xiSize, yiSize=yiSize&
       , expStartTimeStamp=expStartTimeStamp, medianScAttitude=medianScAttitude)	! in ./ebkgmap_skyframe.f90

    call message(AppMsg, Verbose, "Constructing attitude samples.")
    call getBinnedScAtt(attStyleKeyWord&
       , binnedScAtt & !, numAttBins&
       , medianScAttitude, binnedAttFileName)	! in ./ebkgmap_attcommon.f90
    ! numAttBins = size(binnedScAtt)

  end subroutine getAttSkyCoordForEpileupManyA


  !---------------------------------------------------------------------
  subroutine getAttSkyCoordForEpileupInfoC(skyTemplateImageSetName, attStyleKeyWord, binnedAttFileName&
       , wcs, xiSize, yiSize&
       , infoFrameAtt )
       ! , binnedScAtt & !, numAttBins
       ! , expStartTimeStamp, medianScAttitude)
  !---------------------------------------------------------------------

    implicit none 

    !-- Description --
    ! get Attitude data, and sky coordinate data for output image.

    character(*), intent(in) :: skyTemplateImageSetName, attStyleKeyWord, binnedAttFileName
    type(WcsT), intent(out) :: wcs	! type def in ssclib/wcs_aux
    integer, intent(out) :: xiSize, yiSize
    ! type(infoCcdChipsT), intent(inout) :: infoCcdChips	! Defined in epileupmask_shared
    type(infoFrameAttT), intent(inout) :: infoFrameAtt	! Defined in epileupmask_shared
    type(attBinType), pointer :: binnedScAtt(:)	! type def in attcalc2/binned_att.
    ! integer, intent(out) :: numAttBins
    ! real(double), intent(out) :: expStartTimeStamp
    ! type(SpacecraftAttitudeType), intent(out) :: medianScAttitude

    !-------------------------------------------------------------------
    ! Start:

    ! Reading wcs, xiSize, yiSize, expStartTimeStamp, medianScAttitude
    call message(AppMsg, Verbose, "Reading data from template image ("//trim(skyTemplateImageSetName)//").")
    call getSkyCoordFrame(skyTemplateImageSetName&
       , wcs, xiSize=xiSize, yiSize=yiSize&
       , expStartTimeStamp=infoFrameAtt%expStartTimeStamp &
       , medianScAttitude=infoFrameAtt%medianScAttitude)	! in ./ebkgmap_skyframe.f90

    ! Reading binnedScAtt
    call message(AppMsg, Verbose, "Constructing attitude samples.")
    call getBinnedScAtt(attStyleKeyWord&
       , binnedScAtt & !, numAttBins&
       , infoFrameAtt%medianScAttitude, binnedAttFileName)	! in ./ebkgmap_attcommon.f90
    ! numAttBins = size(binnedScAtt)
    call allocateInfoFrameAttT(infoFrameAtt, binnedScAttPtr=binnedScAtt)

  end subroutine getAttSkyCoordForEpileupInfoC


  !---------------------------------------------------------------------
  subroutine convertMask1d2Real(maskLogical, maskReal)
  !---------------------------------------------------------------------
  ! Returns maskReal (single), which is newly allocated here.
  ! Make sure to deallocate it after use.

    implicit none 

    !-- Description --
    ! 

    logical,      pointer :: maskLogical(:)	! intent(in)
    real(single), pointer :: maskReal(:)	! intent(out)

    !-------------------------------------------------------------------
    ! Start:

    allocate(maskReal(size(maskLogical,1))) 

    where (maskLogical)
      maskReal = 1.0
    elsewhere
      maskReal = 0.0
    end where

  end subroutine convertMask1d2Real


  !---------------------------------------------------------------------
  subroutine convertMask2d2Real(maskLogical, maskReal)
  !---------------------------------------------------------------------
  ! Returns maskReal (single), which is newly allocated here.
  ! Make sure to deallocate it after use.

    implicit none 

    logical,      pointer :: maskLogical(:,:)	! intent(in)
    real(single), pointer :: maskReal(:,:)	! intent(out)

    !-------------------------------------------------------------------
    ! Start:

    allocate(maskReal(size(maskLogical,1), size(maskLogical,2))) 

    where (maskLogical)
      maskReal = 1.0
    elsewhere
      maskReal = 0.0
    end where

  end subroutine convertMask2d2Real


  !---------------------------------------------------------------------
  subroutine convertMask3d2Real(maskLogical, maskReal)
  !---------------------------------------------------------------------
  ! Returns maskReal (single), which is newly allocated here.
  ! Make sure to deallocate it after use.

    implicit none 

    logical,      pointer :: maskLogical(:,:,:)	! intent(in)
    real(single), pointer :: maskReal(:,:,:)	! intent(out)

    !-------------------------------------------------------------------
    ! Start:

    allocate(maskReal(size(maskLogical,1), size(maskLogical,2), size(maskLogical,3))) 

    where (maskLogical)
      maskReal = 1.0
    elsewhere
      maskReal = 0.0
    end where

  end subroutine convertMask3d2Real


  !---------------------------------------------------------------------
  subroutine allocateInfoCcdChipsT(infoCcdChips &
     , rawKeyInfoPtr,       rawKeyInfoSize &
     , cubePlaneIndicesPtr, cubePlaneIndicesSize &
     , pattFracSinglePtr,   pattFracSingleSize)
  !---------------------------------------------------------------------

    implicit none 

    type(infoCcdChipsT), intent(inout) :: infoCcdChips
    type(keyInfoType), pointer, optional :: rawKeyInfoPtr(:)	! type defined in eimchip2sky/cube_aux
    integer,           pointer, optional :: cubePlaneIndicesPtr(:)
    real(double),      pointer, optional :: pattFracSinglePtr(:)	! grade pattern
    integer,        intent(in), optional :: rawKeyInfoSize, cubePlaneIndicesSize, pattFracSingleSize

    !-------------------------------------------------------------------
    ! Start:

    if (present(rawKeyInfoPtr) .or. present(rawKeyInfoSize)) then 
      if (infoCcdChips%isAllocatedRawKeyInfo) then
        call error('infoCcdChipsAlreadyAllocated', 'infoCcdChips%rawKeyInfo is already allocated.')
      else
        infoCcdChips%isAllocatedRawKeyInfo = .true.

        if (present(rawKeyInfoPtr) .and. present(rawKeyInfoSize)) &
           call error('bugDoubleArgs', 'Contact the code developer (%rawKeyInfo)')
        if (present(rawKeyInfoPtr))           infoCcdChips%rawKeyInfo => rawKeyInfoPtr
        if (present(rawKeyInfoSize)) allocate(infoCcdChips%rawKeyInfo(rawKeyInfoSize))

      end if
    end if

    if (present(cubePlaneIndicesPtr) .or. present(cubePlaneIndicesSize)) then 
      if (infoCcdChips%isAllocatedCubePlaneIndices) then
        call error('infoCcdChipsAlreadyAllocated', 'infoCcdChips%cubePlaneIndices is already allocated.')
      else
        infoCcdChips%isAllocatedCubePlaneIndices = .true.

        if (present(cubePlaneIndicesPtr) .and. present(cubePlaneIndicesSize)) &
           call error('bugDoubleArgs', 'Contact the code developer (%cubePlaneIndices)')
        if (present(cubePlaneIndicesPtr))           infoCcdChips%cubePlaneIndices => cubePlaneIndicesPtr
        if (present(cubePlaneIndicesSize)) allocate(infoCcdChips%cubePlaneIndices(cubePlaneIndicesSize))

      end if
    end if

    if (present(pattFracSinglePtr) .or. present(pattFracSingleSize)) then 
      if (infoCcdChips%isAllocatedPattFracSingle) then
        call error('infoCcdChipsAlreadyAllocated', 'infoCcdChips%pattFracSingle is already allocated.')
      else
        infoCcdChips%isAllocatedPattFracSingle = .true.

        if (present(pattFracSinglePtr) .and. present(pattFracSingleSize)) &
           call error('bugDoubleArgs', 'Contact the code developer (%pattFracSingle)')
        if (present(pattFracSinglePtr))           infoCcdChips%pattFracSingle => pattFracSinglePtr
        if (present(pattFracSingleSize)) allocate(infoCcdChips%pattFracSingle(pattFracSingleSize))

      end if
    end if

  end subroutine allocateInfoCcdChipsT


  !---------------------------------------------------------------------
  subroutine deallocateInfoCcdChipsT(infoCcdChips &
     , doRawKeyInfo, doCubePlaneIndices, doPattFracSingle)
  !---------------------------------------------------------------------

    implicit none 

    type(infoCcdChipsT), intent(inout) :: infoCcdChips
    logical, intent(in), optional :: doRawKeyInfo, doCubePlaneIndices, doPattFracSingle	! Specify .false. if you do not deallocate these.

    logical :: localDoRawKeyInfo, localDoCubePlaneIndices, localDoPattFracSingle

    !-------------------------------------------------------------------
    ! Start:

    localDoRawKeyInfo       = .true.
    localDoCubePlaneIndices = .true. 
    localDoPattFracSingle   = .true. 
    if (present(doRawKeyInfo))       localDoRawKeyInfo       = doRawKeyInfo
    if (present(doCubePlaneIndices)) localDoCubePlaneIndices = doCubePlaneIndices
    if (present(doPattFracSingle))   localDoPattFracSingle   = doPattFracSingle

    if (localDoRawKeyInfo .and. infoCcdChips%isAllocatedRawKeyInfo) then
      deallocate(infoCcdChips%rawKeyInfo)
      infoCcdChips%isAllocatedRawKeyInfo = .false.
    end if

    if (localDoCubePlaneIndices .and. infoCcdChips%isAllocatedCubePlaneIndices) then
      deallocate(infoCcdChips%cubePlaneIndices)
      infoCcdChips%isAllocatedCubePlaneIndices = .false.
    end if

    if (localDoPattFracSingle .and. infoCcdChips%isAllocatedPattFracSingle) then
      deallocate(infoCcdChips%pattFracSingle)
      infoCcdChips%isAllocatedPattFracSingle = .false.
    end if

  end subroutine deallocateInfoCcdChipsT


  !---------------------------------------------------------------------
  subroutine allocateInfoFrameAttT(infoFrameAtt &
     , binnedScAttPtr,         binnedScAttSize & 
     , numTotFramesPtr,        numTotFramesSize & 
     , binnedAttsNumFramesPtr, binnedAttsNumFramesSizes)
  !---------------------------------------------------------------------

    implicit none 

    type(infoFrameAttT), intent(inout) :: infoFrameAtt
    type(attBinType), pointer, optional :: binnedScAttPtr(:)		! type def is in attcalc2/binned_att.
    real(single),     pointer, optional :: numTotFramesPtr(:) &		! (1:numCubePlanes)
                                         , binnedAttsNumFramesPtr(:,:)	! (1:numCubePlanes,1:numAttBins)
    integer,       intent(in), optional :: binnedScAttSize, numTotFramesSize &
                                         , binnedAttsNumFramesSizes(2)

    !-------------------------------------------------------------------
    ! Start:

    if (present(binnedScAttPtr) .or. present(binnedScAttSize)) then 
      if (infoFrameAtt%isAllocatedBinnedScAtt) then
        call error('infoFrameAttAlreadyAllocated', 'infoFrameAtt%binnedScAtt is already allocated.')
      else
        infoFrameAtt%isAllocatedBinnedScAtt = .true.

        if (present(binnedScAttPtr) .and. present(binnedScAttSize)) &
           call error('bugDoubleArgs', 'Contact the code developer (%binnedScAtt)')
        if (present(binnedScAttPtr))           infoFrameAtt%binnedScAtt => binnedScAttPtr
        if (present(binnedScAttSize)) allocate(infoFrameAtt%binnedScAtt(binnedScAttSize))

        ! Sanity check
        if (infoFrameAtt%isAllocatedBinnedAttsNumFrames) then
          if (size(infoFrameAtt%binnedScAtt) .ne. size(infoFrameAtt%binnedAttsNumFrames, 2)) then
            write(errStr,'(a,i5,a,i5,").")') &
               "Size of %binnedScAtt (=", size(infoFrameAtt%binnedScAtt) &
             , ") differs from of %binnedAttsNumFrames (=", size(infoFrameAtt%binnedAttsNumFrames, 2)
            call warning('inconsistentPtrArraySize')
          end if
        end if
      end if
    end if

    if (present(numTotFramesPtr) .or. present(numTotFramesSize)) then 
      if (infoFrameAtt%isAllocatedNumTotFrames) then
        call error('infoFrameAttAlreadyAllocated', 'infoFrameAtt%numTotFrames is already allocated.')
      else
        infoFrameAtt%isAllocatedNumTotFrames = .true.

        if (present(numTotFramesPtr) .and. present(numTotFramesSize)) &
           call error('bugDoubleArgs', 'Contact the code developer (%numTotFrames)')
        if (present(numTotFramesPtr))           infoFrameAtt%numTotFrames => numTotFramesPtr
        if (present(numTotFramesSize)) allocate(infoFrameAtt%numTotFrames(numTotFramesSize))

        ! Sanity check
        if (infoFrameAtt%isAllocatedBinnedAttsNumFrames) then
          if (size(infoFrameAtt%numTotFrames) .ne. size(infoFrameAtt%binnedAttsNumFrames, 1)) then
            write(errStr,'(a,i5,a,i5,").")') &
               "Size of %numTotFrames (=", size(infoFrameAtt%numTotFrames) &
             , ") differs from of %binnedAttsNumFrames (=", size(infoFrameAtt%binnedAttsNumFrames, 1)
            call warning('inconsistentPtrArraySize')
          end if
        end if
      end if
    end if

    if (present(binnedAttsNumFramesPtr) .or. present(binnedAttsNumFramesSizes)) then 
      if (infoFrameAtt%isAllocatedBinnedAttsNumFrames) then
        call error('infoFrameAttAlreadyAllocated', 'infoFrameAtt%binnedAttsNumFrames is already allocated.')
      else
        infoFrameAtt%isAllocatedBinnedAttsNumFrames = .true.

        if (present(binnedAttsNumFramesPtr) .and. present(binnedAttsNumFramesSizes)) &
           call error('bugDoubleArgs', 'Contact the code developer (%binnedAttsNumFrames)')
        if (present(binnedAttsNumFramesPtr))           infoFrameAtt%binnedAttsNumFrames => binnedAttsNumFramesPtr
        if (present(binnedAttsNumFramesSizes)) &
           allocate(infoFrameAtt%binnedAttsNumFrames(binnedAttsNumFramesSizes(1), binnedAttsNumFramesSizes(2)))

        ! Sanity check
        if (infoFrameAtt%isAllocatedNumTotFrames) then
          if (size(infoFrameAtt%numTotFrames) .ne. size(infoFrameAtt%binnedAttsNumFrames, 1)) then
            write(errStr,'(a,i5,a,i5,").")') &
               "Size of %numTotFrames (=", size(infoFrameAtt%numTotFrames) &
             , ") differs from of %binnedAttsNumFrames (=", size(infoFrameAtt%binnedAttsNumFrames, 1)
            call warning('inconsistentPtrArraySize')
          end if
        end if

        ! Sanity check
        if (infoFrameAtt%isAllocatedBinnedScAtt) then
          if (size(infoFrameAtt%binnedScAtt) .ne. size(infoFrameAtt%binnedAttsNumFrames, 2)) then
            write(errStr,'(a,i5,a,i5,").")') &
               "Size of %binnedScAtt (=", size(infoFrameAtt%binnedScAtt) &
             , ") differs from of %binnedAttsNumFrames (=", size(infoFrameAtt%binnedAttsNumFrames, 2)
            call warning('inconsistentPtrArraySize')
          end if
        end if
      end if
    end if

  end subroutine allocateInfoFrameAttT

  !---------------------------------------------------------------------
  subroutine deallocateInfoFrameAttT(infoFrameAtt &
     , doBinnedScAtt, doNumTotFrames, doBinnedAttsNumFrames)
  !---------------------------------------------------------------------

    implicit none 

    type(infoFrameAttT), intent(inout) :: infoFrameAtt
    logical, intent(in), optional :: doBinnedScAtt, doNumTotFrames, doBinnedAttsNumFrames	! Specify .false. if you do not deallocate these.

    logical :: localDoBinnedScAtt, localDoNumTotFrames, localDoBinnedAttsNumFrames

    !-------------------------------------------------------------------
    ! Start:

    localDoBinnedScAtt         = .true.
    localDoNumTotFrames        = .true.
    localDoBinnedAttsNumFrames = .true.
    if (present(doBinnedScAtt))         localDoBinnedScAtt         = doBinnedScAtt
    if (present(doNumTotFrames))        localDoNumTotFrames        = doNumTotFrames
    if (present(doBinnedAttsNumFrames)) localDoBinnedAttsNumFrames = doBinnedAttsNumFrames

    if (localDoBinnedScAtt         .and. infoFrameAtt%isAllocatedBinnedScAtt) then
      deallocate(infoFrameAtt%binnedScAtt)
      infoFrameAtt%isAllocatedBinnedScAtt = .false.
    end if
    
    if (localDoNumTotFrames        .and. infoFrameAtt%isAllocatedNumTotFrames) then
      deallocate(infoFrameAtt%numTotFrames)
      infoFrameAtt%isAllocatedNumTotFrames = .false.
    end if

    if (localDoBinnedAttsNumFrames .and. infoFrameAtt%isAllocatedBinnedAttsNumFrames) then
      deallocate(infoFrameAtt%binnedAttsNumFrames)
      infoFrameAtt%isAllocatedBinnedAttsNumFrames = .false.
    end if

  end subroutine deallocateInfoFrameAttT


  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none 

    !-------------------------------------------------------------------
    ! Start:

  end subroutine template

end module epileupmask_shared

