! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: emrgamap_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task makes an image of the scatter in MOS images caused by the RGA.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module emrgamap_mod

  use dal
  use cal
  use param  
  use errorHandling
  use dataFilter	! in selectlib
!  use dss		! for RangeT

  use cal_aux
  use constants
  use coordinate
  use dal_aux
  use dss_ranges_aux
  use geometric_types
  use regridders	! for regrid()
  use save_image
  use ssc_misc
  use wcs_aux

  implicit none 

  private 

  ! Command-line parameters:
  character(255) ::&
    srcListSetTabName='',&
    templateImageSetName='',&
    outputImageSetName='',&
    selectionExpression='',&
    tempImageSetName='',&
    tempSetName=''
  character(20) ::&
    raColName='',&
    decColName='',&
    rateColName=''
  real(double) ::&
    psfEeV=0.0d0,&
    energyFraction=0.0d0

  ! Other globals:

  public emrgamap

contains 

  !---------------------------------------------------------------------
  subroutine emrgamap
  !---------------------------------------------------------------------

    implicit none

    integer, parameter :: numDetPixelsX=95, numDetPixelsY=2375
    real(single), parameter :: sizeDetPixelsX=400.0, sizeDetPixelsY=16.0
    ! There's not too rapid change in the detx direction so can afford to make the pixels fairly large in that direction.

    character(255) :: srcListSetName, srcListTabName
    type(DataSetT) :: srcListSet, templateSet
    type(TableT) :: tab
    integer(int32) :: numSrcs, srcNum
    real(double), pointer :: srcRaDeg(:), srcDecDeg(:)
    real(single), pointer :: srcRate(:)
    type(ArrayT) :: templateArray
    type(WcsT) :: wcs
    integer, pointer :: imageDims(:)
    integer(int16) :: xSize, ySize, xi, yi
!, xiStart, xiFinis, yiStart, yiFinis
    real(single), allocatable :: skyImage(:,:)
    integer(int32) :: instrumentId
    real(double) :: expStartTimeStamp
    type(SpacecraftAttitudeType) :: scAttitude, instrumAttitude
    real(double), allocatable :: srcThetaArcsec(:), srcPhi(:)
    real(single) ::&
      detXValues(0:numDetPixelsX), detYValues(0:numDetPixelsY),&
      detXStart, detXWidth, detYHeight, detYStart,&
      binnedTemplateFraction(numDetPixelsX),&
      detXYImage(numDetPixelsX,numDetPixelsY), psf1D(1,numDetPixelsY)
    real(double), allocatable :: xTanArray(:,:), yTanArray(:,:)
    type(Point2DT) :: detImgPixCnrPixels(0:numDetPixelsX,0:numDetPixelsY)
    real(single), pointer :: templateDetX(:), templateFraction(:), psfImage(:,:)
    type(PsfBinSizeT) :: psfPixelSizeMm
    type(Point2dDbleT), allocatable ::&
      psfImgPixCrnrTel(:,:), psfImgPixCrnrDet(:,:)
    type(Point2DT), allocatable :: psfImgPixCrnr1d(:,:)
    integer :: foreignStatus

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    ! Load the source list:

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)

    srcListSet = clone(srcListSetName, tempSetName, TEMP)
    tab = table(srcListSet, srcListTabName, MODIFY)
    call initFilter(selectionExpression)
    numSrcs = doFilter(tab)

    if (numSrcs <= 0)&
call error('noSourcesSelected'&
, 'No sources passed the specified filtering expression.')

    call readColDataToPtr(tab, raColName,  srcRaDeg)
    call readColDataToPtr(tab, decColName, srcDecDeg)
    call readColDataToPtr(tab, rateColName, srcRate)

    ! Get details of template image:

    templateSet = dataSet(templateImageSetName, READ)
    templateArray = array(templateSet, 0)
    call readWcs(templateSet, templateArray, wcs)
!    call easifyWcsType(wcs%x%ctype, wcs%y%ctype, wcsType)
!    if (wcsType .ne. 'DETXY' .and. wcsType .ne. 'XY')&
!call error('badWcsCtype','')

    imageDims => dimensions(templateArray)
    xSize = imageDims(0)
    ySize = imageDims(1)
    allocate(skyImage(xSize, ySize))

    call getCalInfo(templateSet, instrumentId, expStartTimeStamp, scAttitude&
      , instrumAttitude) ! in ssclib/cal_aux.
    call release(templateSet)
    call CAL_setState(accuracyLevel=ACCURACY_MEDIUM)

    ! Get src thetas and phis:

    allocate(srcThetaArcsec(numSrcs), srcPhi(numSrcs))
    call raDecToTel(srcRaDeg*PiOn180, srcDecDeg*PiOn180, scAttitude&
      , expStartTimeStamp, srcThetaArcsec, srcPhi)

!    xiStart = -numDetPixelsX / 2
!    xiFinis = xiStart + numDetPixelsX
!    yiStart = -numDetPixelsY / 2
!    yiFinis = yiStart + numDetPixelsY
!    detXValues = (/(sizeDetPixelsX * xi, xi=xiStart, xiFinis)/)
!    detYValues = (/(sizeDetPixelsY * yi, yi=yiStart, yiFinis)/)
detXStart = -(numDetPixelsX / 2.0) * sizeDetPixelsX
detYStart = -(numDetPixelsY / 2.0) * sizeDetPixelsY
detXWidth = numDetPixelsX * sizeDetPixelsX
detYHeight = numDetPixelsY * sizeDetPixelsY
detXValues = detXStart + (/(sizeDetPixelsX * xi, xi=0, numDetPixelsX)/)
detYValues = detYStart + (/(sizeDetPixelsY * yi, yi=0, numDetPixelsY)/)

    allocate(xTanArray(numDetPixelsX+1,numDetPixelsY+1))
    allocate(yTanArray(numDetPixelsX+1,numDetPixelsY+1))

    call detToTan(real(spread(detXValues, 2, numDetPixelsY+1), double)&
      , real(spread(detYValues, 1, numDetPixelsX+1), double)&
      , wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180, scAttitude&
      , expStartTimeStamp, xTanArray, yTanArray) ! in ssclib/coordinate.

    call tanToPixels(wcs, xTanArray, yTanArray, detImgPixCnrPixels%x&
      , detImgPixCnrPixels%y) ! in ssclib/wcs_aux.

    deallocate(xTanArray, yTanArray)

    call loadScatterTemplate(instrumentId, templateDetX, templateFraction)
    do xi = 1, numDetPixelsX
      binnedTemplateFraction(xi) = integrateRanges(templateDetX&
        , templateFraction, detXValues(xi-1), detXValues(xi))
      ! integrateRanges is in ssclib/dss_ranges_aux.
    end do

    ! Now make the DETXY image:

!vvvv
    skyImage = 0.0
!*** do separately for separate att bins.
    detXYImage = 0.0
    do srcNum = 1, numSrcs
print *, srcNum,'/',numSrcs

      call getPSFImage(energyFraction, psfEeV, srcThetaArcsec(srcNum)&
        , srcPhi(srcNum), psfImage, psfPixelSizeMm)

      allocate(psfImgPixCrnrTel(size(psfImage,1)+1, size(psfImage,2)+1))
      allocate(psfImgPixCrnrDet(size(psfImage,1)+1, size(psfImage,2)+1))
      allocate(psfImgPixCrnr1d( size(psfImage,1)+1, size(psfImage,2)+1))

      call getPsfImagePixelCorners(srcThetaArcsec(srcNum), srcPhi(srcNum)&
        , psfPixelSizeMm, psfImage, psfImgPixCrnrTel)
      ! in ssclib/coordinate.
      !
      ! The pairs of numbers (type(Point2dT)) in this array represent the
      ! coordinates of the corners of the psfImage pixels, expressed in
      ! the exposure-map image (ie, sky) coordinate system.

      call psfToDet(psfImgPixCrnrTel%x, psfImgPixCrnrTel%y&
        , psfImgPixCrnrDet%x, psfImgPixCrnrDet%y)
      ! in ssclib/coordinate.

      ! Now we are going to do a bit of a trick. We want to rebin the PSF to detxy coordinates; however what we want to end up with is the detx-integrated form of this. So let us collapse these two steps into one. We'll do this by defining a pseudo-2-dimensional array psf1D which has an x dimension of only 1 pixel but a y dimension of numDetPixelsY. The single x bin extends from detXValues(0) to detXValues(numDetPixelsX). Rebinning to such an array is equivalent to the two steps described above.
      !
      ! First step is to transform psfImgPixCrnrDet to the pixel coordinate system of this new array:

      psfImgPixCrnr1d%x = (psfImgPixCrnrDet%x - detXValues(0)) / detXWidth
      psfImgPixCrnr1d%y = fracVal(0.5_single, real(numDetPixelsY+0.5, single)&
        , real((psfImgPixCrnrDet%y - detYValues(0)) / detYHeight, single))

      call regridCartesian(psfImgPixCrnr1d, psfImage&
        , psf1D, status=foreignStatus) ! in ssclib/regridders.

if (foreignStatus < 0) call error('','')
      if (foreignStatus > 0) then
        print *, 'regrid exit status:', foreignStatus
      end if

      call CAL_releaseMemory(psfImage)
      deallocate(psfImgPixCrnrTel, psfImgPixCrnrDet, psfImgPixCrnr1d)

      detXYImage = detXYImage + spread(psf1D(1,:), 1, numDetPixelsX)&
        * spread(binnedTemplateFraction, 2, numDetPixelsY) * srcRate(srcNum-1)
    end do

    ! Now rebin image to sky coords and add to output rate image:

    call regridCartesian(detImgPixCnrPixels, detXYImage&
        , skyImage, status=foreignStatus) ! in ssclib/regridders.

if (foreignStatus < 0) call error('','')
    if (foreignStatus > 0) then
      print *, 'regrid exit status:', foreignStatus
    end if
!^^^^

    call release(srcListSet)
    deallocate(srcRaDeg, srcDecDeg, srcRate)
    deallocate(templateDetX, templateFraction)

    call saveImage(skyImage, templateImageSetName, outputImageSetName&
      , tempImageSetName)

  end subroutine emrgamap

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    srcListSetTabName = stringParameter('srclisttab')
    selectionExpression = stringParameter('expression')
    raColName = stringParameter('racol')
    decColName = stringParameter('deccol')
    rateColName = stringParameter('ratecol')
    templateImageSetName = stringParameter('templateset')
    outputImageSetName = stringParameter('outset')
    psfEeV = realParameter('psfenergy')
    energyFraction = realParameter('energyfraction')
    tempImageSetName = stringParameter('tempimageset')
tempSetName = 'tempset.ds'

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine getPSFImage(psfEnergyFraction, psfEeV, srcThetaArcsec&
    , srcPhi, psfImage, psfPixelSizeMm)
  !---------------------------------------------------------------------

    implicit none

    real(double),      intent(in)  :: psfEnergyFraction,&
                                      psfEeV, srcThetaArcsec,&
                                      srcPhi
    real(single),      pointer     :: psfImage(:,:)
    type(PsfBinSizeT), intent(out) :: psfPixelSizeMm

    real(double) :: effectiveSrcThetaArcsec
    type(PsfValidityRangesT) :: validPsfRanges
    type(PsfT) :: psfHandle
    type(PsfWindowT) :: psfWindow

    !-------------------------------------------------------------------
    ! Start

    validPsfRanges = CAL_psfValidityRanges()

    if  (psfEeV < validPsfRanges%eMin&
    .or. psfEeV > validPsfRanges%eMax)&
call error('energyOutOfRange',"")

    if (srcThetaArcsec < validPsfRanges%thetaMin)&
call error('thetaOutOfRange',"")

    if (srcPhi < validPsfRanges%phiMin&
    .or. srcPhi > validPsfRanges%phiMax)&
call error('phiOutOfRange',"")

    if (srcThetaArcsec > validPsfRanges%thetaMax) then
      call warning('thetaOutOfRange','')
      effectiveSrcThetaArcsec = validPsfRanges%thetaMax&
        * (1.0d0 - epsilon(srcThetaArcsec))
    else
      effectiveSrcThetaArcsec = srcThetaArcsec
    end if

    psfHandle = CAL_getPSF(psfEeV, effectiveSrcThetaArcsec, srcPhi)
    psfPixelSizeMm = CAL_psfBinSize(psfHandle)
    psfWindow = CAL_psfEnboxedEnergy(psfHandle, fraction=psfEnergyFraction)
    call CAL_psfImage(psfHandle, psfWindow, psfImage)

    ! Normalize the psf image:

    psfImage = psfImage * psfPixelSizeMm%binX * psfPixelSizeMm%binY

  end subroutine getPSFImage

  !---------------------------------------------------------------------
  subroutine loadScatterTemplate(instrumentId, templateDetX, templateFraction)
  !---------------------------------------------------------------------
  ! NOTE: the calling routine should deallocate the pointers templateDetX and templateFraction after use.

    implicit none

    integer(int32), intent(in) :: instrumentId
    real(single), pointer :: templateDetX(:), templateFraction(:)

    real(single), allocatable :: brightestDetYCount(:), baselineCount(:)
    integer :: i
    real(single) ::&
      histo1PeakCts, histo1BaseCts, histo3PeakCts, histo3BaseCts,&
      histo6PeakCts, histo6BaseCts, histo1TotalCts, histo3TotalCts,&
      histo6TotalCts, integral3, integral6, n
!    type(RangeT) :: x3, x6
real(single) :: x3Lo, x3Hi, x6Lo, x6Hi

    !-------------------------------------------------------------------
    ! Start

if (instrumentId .ne. EMOS1) call error('',"")
!*** I'll measure for mos2 later.

    ! For obsid 0084020601 I used evselect to make an image (with only #XMMEA_EM filtering) in DETX/DETY with bin size (800,400). At each DETX value on CCDs 3 and 6, the brightest pixel as a function of DETY was at DETY=-105.5. I measured the value at this DETXY for several values of DETX and also estimated the baseline value. These numbers are as follows:

    allocate(templateFraction(27))
    allocate(templateDetX(27), brightestDetYCount(27), baselineCount(27))

    templateDetX(1) = -17000.0
    templateDetX(2:13)  = -16168.5 + (/(800.0*i, i=0,11)/)
    templateDetX(14:26) =   7031.5 + (/(800.0*i, i=0,12)/)
    templateDetX(27) = 17200.0

    brightestDetYCount = (/&
      121,&
      121,&
      118,&
      199,&
      224,&
      259,&
      277,&
      230,&
      240,&
      277,&
      343,&
      325,&
      405,&
      623,&
      507,&
      438,&
      395,&
      381,&
      318,&
      293,&
      263,&
      246,&
      266,&
      223,&
      175,&
      152,&
      152&
    /)
    baselineCount = (/&
       50,&
       50,&
       50,&
       60,&
       75,&
       80,&
       90,&
      100,&
      105,&
      130,&
      165,&
      200,&
      260,&
      300,&
      270,&
      200,&
      170,&
      130,&
      110,&
      100,&
       80,&
       70,&
       60,&
       50,&
       50,&
       45,&
       45&
    /)

    ! Note that extra points at DETX = -17000 and +17200 have been included so as to have coverage right out to the edge of the FOV. Counts values from the next innermost-lying points have arbitrarily been copied to these extrema. Beyond these limits the amount of scatter is zero.
    !
    ! Clearly the first step in processing these raw numbers is to subtract the baseline from the peak values:

    templateFraction = brightestDetYCount - baselineCount

    ! The remainder is proportional to the final values of templateFraction which we want this subroutine to return. To determine the normalisation constant we need to compare the scatter in CCDs 3 and 6 with the counts in the central PSF. For observation 0084020601 the central CCD was (necessarily, considering the expected extreme count rate for this bright source) in timing mode. Another observation was examined, 0134720901, in which the central CCD was in imaging mode. A histogram vs DETY (in 200-unit-wide bins) was accumulated for CCDs 1 (=central), 3 (=high DETX) and 6 (=low DETX). The peak value vs estimated baseline counts from this are as follows:

!    obs0134720901ccd1PeakCts = 42400
!    obs0134720901ccd1BaseCts =   100
!    obs0134720901ccd3PeakCts =   118
!    obs0134720901ccd3BaseCts =    60
!    obs0134720901ccd6PeakCts =    95
!    obs0134720901ccd6BaseCts =    65
histo1PeakCts = 42400
histo1BaseCts =   100
histo3PeakCts =   118
histo3BaseCts =    60
histo6PeakCts =    95
histo6BaseCts =    65

!    ! To enable comparison with obs 0084020601 I did the same for CCDs 3 and 6 of this obs:
!
!    obs0084020601ccd3PeakCts =  2330
!    obs0084020601ccd3BaseCts =   760
!    obs0084020601ccd6PeakCts =  1741
!    obs0084020601ccd6BaseCts =   770

    ! To calculate the normalisation constant I lay out the theory of the whole business as follows.
    !
    ! What we want to do is return a function F(DETX) such that the count rate per square DETXY unit due to the RGA scatter, s(DETX,DETY), is given by
    !
    !
    !	s(DETX,DETY) = F(DETX) * P(DETY)
    !
    !	                         inf
    !	                         /
    !	             = F(DETX) * | dDETX psf(DETX,DETY).
    !	                         /
    !	                       -inf
    !
    ! The units of F are d^-1, where 'd' here signifies the DETXY unit, currently set at 0.05 arcsec. The vector templateFraction at the present time holds a series of samples of F, multiplied by a normalization constant N. The counts H(ccd,i) in bin i of the histograms described above can be expressed in terms of this N*F as follows:
    !
    !	              200*(DETY_i+1)   x_hi(ccd)
    !	            1  /               /
    !	H(ccd,i) = --- | dDETY P(DETY) | dDETX N*F(DETX).
    !	            N  /               /
    !	              200*DETY_i      x_lo(ccd)
    !
    ! However for the central ccd we have to add in the contribution from the unscattered psf itself. This dominates the expression, so we may approximate H(1,i) as
    !
    !	        200*(DETY_i+1)
    !	         /
    !	H(1,i) = | dDETY P(DETY).
    !	         /
    !	        200*DETY_i
    !
    ! The following expression for N may thus be derived:
    !
    !	               ( x_hi(3)               x_hi(6)            )
    !                  (  /                     /                 )
    !	      H(1,i) * (  | dDETX N*F(DETX)  +  | dDETX N*F(DETX) )
    !                  (  /                     /                 )
    !	               ( x_lo(3)               x_lo(6)            )
    !	N = --------------------------------------------------------.
    !	                   H(3,i) + H(6,i)
    !
    ! I'll use N*F from obs 0084020601 and H values from obs 0134720901:

    histo1TotalCts = histo1PeakCts - histo1BaseCts
    histo3TotalCts = histo3PeakCts - histo3BaseCts
    histo6TotalCts = histo6PeakCts - histo6BaseCts

!    x3%lower%value =   6670
!    x3%upper%value =  17200
!    x6%lower%value = -17000
!    x6%upper%value =  -6530
!    x3%lower%type = INCLUSIVE
!    x3%upper%type = INCLUSIVE
!    x6%lower%type = INCLUSIVE
!    x6%upper%type = INCLUSIVE
!
!    integral3 = integrateRanges(templateDetX, templateFraction, x3)
!    integral6 = integrateRanges(templateDetX, templateFraction, x6)

x3Lo =   6670
x3Hi =  17200
x6Lo = -17000
x6Hi =  -6530
integral3 = integrateRanges(templateDetX, templateFraction, x3Lo, x3Hi)
integral6 = integrateRanges(templateDetX, templateFraction, x6Lo, x6Hi)
    ! integrateRanges is in ssclib/dss_ranges_aux.

    n = histo1TotalCts * (integral3 + integral6)&
      / (histo3TotalCts + histo6TotalCts)

    ! Finally, use this to normalize the F values:

    templateFraction = templateFraction / n

  end subroutine loadScatterTemplate

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module emrgamap_mod
