! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: esrcmap_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task makes an image from an input source list.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module esrcmap_mod

  use dal
  use cal
  use param  
  use errorHandling
  use dataFilter	! in selectlib

  use array_utils
  use cal_aux
  use constants
  use coordinate
  use dal_aux
!  use fftw_aux
  use geometric_types
  use regridders	! for regrid()
  use save_image
  use sort_mod
  use ssc_misc
  use wcs_aux
use dump
!use temp_ssclib

  implicit none 

  private 

  ! Command-line parameters:
  character(1000) ::&
    selectionExpression=''
  character(256) ::&
    srcListSetTabName='',&
    maskSetName='',&
    templateImageSetName='',&
    tempImageSetName='',&
    tempSetName=''
  character(256), allocatable ::&
    outImageSetNames(:)
  character(20) ::&
    idColName='',&
    raColName='',&
    decColName='',&
    ximaColName='',&
    yimaColName='',&
    rateColName='',&
    psfEnergyStyle='',&
    positionStyle=''
  real(double), allocatable ::&
    userPsfMeanEnergies(:)
  real(double) ::&
    energyFraction=0.0d0
  integer ::&
    numOutImages=0,&
    idInstr=0
  integer, allocatable ::&
    userIdBands(:)
  logical ::&
    withIdColName=.false.,&
    withMaskSet=.false.


  ! Other globals:
  integer(int32) ::&
    instrumentId
  real(double) ::&
    expStartTimeStamp
  type(SpacecraftAttitudeType) ::&
    scAttitude, instrumAttitude

  public esrcmap

contains 

  !---------------------------------------------------------------------
  subroutine esrcmap
  !---------------------------------------------------------------------

    implicit none

    character(1) :: bandChars(9)
    character(256) :: srcListSetName, srcListTabName
    type(DataSetT) :: srcListSet, templateSet
    type(TableT) :: sourceTab
    integer :: numFilteredRows, numSources, srcNum, outImageNum, xi, yi
!, xiSize, yiSize
    logical, pointer :: mask(:,:)
    type(ArrayT) :: templateArray
    real(double), pointer :: srcRaDeg(:), srcDecDeg(:)
    real(single), pointer :: srcXima(:), srcYima(:), srcRate(:,:)
    logical,      pointer :: srcIgnore(:,:)
    type(WcsT) :: wcs
    integer, pointer :: imageDims(:)
    integer(int16) :: xiSize, yiSize
    real(single), allocatable :: imageCube(:,:,:), srcXpixels(:), srcYpixels(:)
    real(single), pointer :: patchCube(:,:,:)
    real(double), allocatable ::&
      srcXTan(:), srcYTan(:), srcThetaArcsec(:), srcPhi(:), psfMeanEnergies(:)
    integer :: foreignStatus
    character(8) :: srcNumStr, maxSrcNumStr, prefix, eLoKwd, eHiKwd
    real(double) :: eLo, eHi
    integer(int32) :: templateInstrumentId
    real(double), allocatable :: srcTanX(:), srcTanY(:)
    real(single) :: x, y

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, 'aaa'
    bandChars = (/'1', '2', '3', '4', '5', '6', '7', '8', '9'/)

    call readParameters

    !...................................................................
    ! Read image template and mask details:

    templateSet = dataSet(templateImageSetName, READ)
    templateArray = array(templateSet, 0)
    call readWcs(templateSet, templateArray, wcs) ! in ssclib/wcs_aux.

    imageDims => dimensions(templateArray)
    xiSize = imageDims(0)
    yiSize = imageDims(1)

    if (withMaskSet) then
      call readArrayData(maskSetName, mask) ! in ssclib/dal_aux.
      if (size(mask, 1) .ne. xiSize .or. size(mask, 2) .ne. yiSize)&
call error('sizeMismatch',"")
    else
      allocate(mask(xiSize,yiSize))
      mask = .true.
    end if

if (testFlag) print *, 'bbb'

    call getCalInfo(templateSet, templateInstrumentId, expStartTimeStamp&
      , scAttitude, instrumAttitude)

    call release(templateArray)
    call release(templateSet)

    if (withIdColName) then
      select case(idInstr)
      case(1)
        instrumentId = EPN
      case(2)
        instrumentId = EMOS1
      case(3)
        instrumentId = EMOS2
      case default
call error('badIdInstr',"")
      end select

      if (templateInstrumentId .ne. instrumentId) call warning('inconsistentInstruments',"")

    else
      instrumentId = templateInstrumentId
!********** print it - just to allow the user to make sure they're doing the right thing.
    end if
if (testFlag) print *, 'ccc'

    select case(instrumentId)
    case(EPN)
      prefix = 'PN_'
    case(EMOS1)
      prefix = 'M1_'
    case(EMOS2)
      prefix = 'M2_'
    case default
call error('badInstrument',"")
    end select

    call CAL_setState(instrument=instrumentId, accuracyLevel=ACCURACY_MEDIUM)

    !...................................................................
    ! Open source set and load the columns:

    call splitSetTabName(srcListSetTabName, srcListSetName, srcListTabName)
    ! in ssclib/dal_aux.
    srcListSet = clone(srcListSetName, tempSetName, TEMP)
if (testFlag) print *, 'ddd'

    allocate(psfMeanEnergies(numOutImages))
    select case(psfEnergyStyle)
    case('user')
      psfMeanEnergies = userPsfMeanEnergies
    case('set')
      do outImageNum = 1, numOutImages
        eLoKwd = prefix//trim(bandChars(userIdBands(outImageNum)))//'_ELO'
        eHiKwd = prefix//trim(bandChars(userIdBands(outImageNum)))//'_EHI'
        eLo = real64Attribute(srcListSet, eLoKwd)
        eHi = real64Attribute(srcListSet, eHiKwd)
        psfMeanEnergies(outImageNum) = (eLo + eHi) / 2.0d0
      end do

    case default
call error('badPsfEnergyStyle',"")
    end select
if (testFlag) print *, 'eee'

    sourceTab = table(srcListSet, srcListTabName, MODIFY)
    call initFilter(selectionExpression)
    numFilteredRows = doFilter(sourceTab)

    if (numFilteredRows <= 0)&
      call warning('noSourcesSelected'&
      , 'No sources passed the specified filtering expression.')

    allocate(imageCube(numOutImages, xiSize, yiSize))

if (testFlag) print *, 'fff'
    select case(positionStyle)
    case('radec')
      if (withIdColName) then
if (testFlag) print *, 'fff_aa0'
        call readSrcList(sourceTab, raColName, decColName, rateColName&
          , srcRaDeg, srcDecDeg, srcRate, srcIgnore&
          , idColName, idInstr, userIdBands)
if (testFlag) print *, 'fff_aa1'
      else
if (testFlag) print *, 'fff_ab0'
        call readSrcList(sourceTab, raColName, decColName, rateColName&
          , srcRaDeg, srcDecDeg, srcRate, srcIgnore)
      end if

!********** calculate srcXima, srcYima
if (testFlag) print *, 'fff_a1'

      numSources = size(srcRaDeg)
      if (all(srcIgnore)) call warning('noValidSources',"")

if (testFlag) then
  print *, '  ra=', srcRaDeg(1)
  print *, ' dec=', srcDecDeg(1)
  print *, 'scts=', srcRate(1,1)
end if

      !.................................................................
      ! Convert source Ra/Dec to image coordinates:

      allocate(srcTanX(numSources), srcTanY(numSources))
      call raDecToTan(srcRaDeg*PiOn180, srcDecDeg*PiOn180, wcs&
        , srcTanX, srcTanY) ! in ssclib/coordinate

      allocate(srcXpixels(numSources), srcYpixels(numSources))
      call tanToPixels(wcs, srcTanX, srcTanY, srcXpixels, srcYpixels)
      ! in ssclib/wcs_aux
      deallocate(srcTanX, srcTanY)

      ! Get src thetas and phis:

      allocate(srcThetaArcsec(numSources), srcPhi(numSources))
      call raDecToTel(srcRaDeg*PiOn180, srcDecDeg*PiOn180, scAttitude&
        , expStartTimeStamp, srcThetaArcsec, srcPhi)

if (testFlag) then
  print *, ''
  print *, 'theta=', srcThetaArcsec(1)
  print *, '  phi=', srcPhi(1)
end if

      ! Now make the image:

      imageCube = 0.0
!******** have to allocate it first!!?
      maxSrcNumStr = intToStr(numSources)
      do srcNum = 1, numSources
      if (all(srcIgnore)) cycle

        x = srcXpixels(srcNum) + 0.5
        y = srcYpixels(srcNum) + 0.5
!        xi = nint(srcXpixels(srcNum))
!        yi = nint(srcYpixels(srcNum))
        xi = floor(x)
        yi = floor(y)

      if (xi < 1 .or. xi > xiSize .or. yi < 1 .or. yi > yiSize) cycle
      if (.not. mask(xi, yi)) cycle

        srcNumStr = intToStr(srcNum)
        errStr = 'Source '//trim(srcNumStr)//' out of '//trim(maxSrcNumStr)
        call message(AppMsg, verbose) 
if (testFlag) print *, 'fff_a2'

        call getCalPsfPatch(xiSize, yiSize, wcs, energyFraction&
          , psfMeanEnergies, srcThetaArcsec(srcNum), srcPhi(srcNum)&
          , patchCube, foreignStatus)
if (testFlag) print *, 'fff_a3'



!********* change it so that the same size patch is used for all sources...? Then can write patch size as kwd in image header.
!******* can also write feathering fraction as kwd.



        if (withIdColName) then
          call addPatchCubeVec(imageCube, patchCube, srcRate(srcNum,:))
        else
          call addPatchCubeScal(imageCube, patchCube, srcRate(srcNum,1))
        end if
if (testFlag) print *, 'fff_a4'

        deallocate(patchCube)
if (testFlag) print *, 'fff_a5'
      end do
      deallocate(srcRaDeg, srcDecDeg, srcRate, srcIgnore)

    case('pixel')
      if (withIdColName) then
if (testFlag) print *, 'fff_ba0'
        call readSrcListPixels(sourceTab, ximaColName, yimaColName, rateColName&
          , srcXima, srcYima, srcRate, srcIgnore&
          , idColName, idInstr, userIdBands)
      else
if (testFlag) print *, 'fff_bb0'
        call readSrcListPixels(sourceTab, ximaColName, yimaColName, rateColName&
          , srcXima, srcYima, srcRate, srcIgnore)
      end if

      numSources = size(srcXima)
      if (all(srcIgnore)) call warning('noValidSources',"")

      ! Get src thetas and phis:

      allocate(srcXTan(numSources), srcYTan(numSources))
      call pixelsToTan(srcXima, srcYima, wcs, srcXTan, srcYTan)
      ! in ssclib/wcs_aux.

      allocate(srcThetaArcsec(numSources), srcPhi(numSources))
      call tanToTel(srcXTan, srcYTan, wcs%x%refWorld * PiOn180&
        , wcs%y%refWorld * PiOn180, scAttitude, expStartTimeStamp&
        , srcThetaArcsec, srcPhi) ! in ssclib/coordinate.

      ! Now make the image:

      imageCube = 0.0
!******** have to allocate it first!!?
      maxSrcNumStr = intToStr(numSources)
      do srcNum = 1, numSources
      if (all(srcIgnore)) cycle

        xi = nint(srcXpixels(srcNum))
        yi = nint(srcYpixels(srcNum))

      if (xi < 1 .or. xi > xiSize .or. yi < 1 .or. yi > yiSize) cycle
      if (.not. mask(xi, yi)) cycle

        srcNumStr = intToStr(srcNum)
        errStr = 'Source '//trim(srcNumStr)//' out of '//trim(maxSrcNumStr)
        call message(AppMsg, verbose) 

        call getCalPsfPatchPixels(xiSize, yiSize, wcs, energyFraction&
          , psfMeanEnergies, srcThetaArcsec(srcNum), srcPhi(srcNum)&
          , srcXima(srcNum), srcYima(srcNum), patchCube, foreignStatus)

        if (withIdColName) then
          call addPatchCubeVec(imageCube, patchCube, srcRate(srcNum,:))
        else
          call addPatchCubeScal(imageCube, patchCube, srcRate(srcNum,1))
        end if

        deallocate(patchCube)
      end do
      deallocate(srcXima, srcYima, srcRate, srcIgnore)
    case default
call error('badPositionStyle',"")
    end select
if (testFlag) print *, 'fff'

    call release(sourceTab)
    call release(srcListSet)
    deallocate(mask)

    do outImageNum = 1, numOutImages
      call saveImage(imageCube(outImageNum,:,:), templateImageSetName&
        , outImageSetNames(outImageNum), tempImageSetName)
      ! In ssclib/save_image.
    end do
if (testFlag) print *, 'zzz'

  end subroutine esrcmap

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    integer :: outImageNum

    !-------------------------------------------------------------------
    ! Start

    srcListSetTabName = stringParameter('srclisttab')
    withMaskSet = booleanParameter('withmaskset')
    if (withMaskSet)&
      maskSetName = stringParameter('maskset')
    selectionExpression = stringParameter('expression')
    templateImageSetName = stringParameter('templateset')

    numOutImages = parameterCount('outsets')
    allocate(outImageSetNames(numOutImages))
    do outImageNum = 1, numOutImages
      outImageSetNames(outImageNum) = stringParameter('outsets'&
        , outImageNum-1)
    end do

    withIdColName = booleanParameter('withidcol')
    if (withIdColName) then
      idColName = stringParameter('idcol')
      idInstr = intParameter('idinst')

if (parameterCount('idbands') .ne. numOutImages) call error('badNumIdBands',"")
      allocate(userIdBands(numOutImages))
      do outImageNum = 1, numOutImages
        userIdBands(outImageNum) = intParameter('idbands'&
          , outImageNum-1)
      end do
    end if

    positionStyle = stringParameter('posstyle')
    select case(positionStyle)
    case('radec')
      raColName = stringParameter('racol')
      decColName = stringParameter('deccol')
    case('pixel')
      ximaColName = stringParameter('xcol')
      yimaColName = stringParameter('ycol')
    case default
call error('badPositionStyle',"")
    end select

    rateColName = stringParameter('ratecol')

    psfEnergyStyle = stringParameter('estyle')
    select case(psfEnergyStyle)
    case('user')
if (parameterCount('psfenergies') .ne. numOutImages) call error('badNumPsfEnergies',"")
      allocate(userPsfMeanEnergies(numOutImages))
      do outImageNum = 1, numOutImages
        userPsfMeanEnergies(outImageNum) = realParameter('psfenergies'&
          , outImageNum-1)
      end do

    case('set')
    case default
call error('badPsfEnergyStyle',"")
    end select

    energyFraction = realParameter('energyfraction')
    tempImageSetName = stringParameter('tempimageset')
    tempSetName = stringParameter('tempsrclistset')

    applyAttcalcFudge = booleanParameter('withboresightfudge')
    ! only needed while boresight psi has wrong sign. IMS 2005-11-17

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine readSrcList(tab, raColName, decColName, rateColName&
  , srcRaDeg, srcDecDeg, srcRate, srcIgnore, idColName, idInstr, idBands)
  !---------------------------------------------------------------------
  ! Note! The calling program should deallocate the pointers srcRaDeg,
  ! srcDecDeg, srcIgnore and srcRate after use.

    implicit none

    type(TableT), intent(in)           :: tab
    character(*), intent(in)           :: raColName, decColName, rateColName
    real(double), pointer              :: srcRaDeg(:), srcDecDeg(:)
    real(single), pointer              :: srcRate(:,:)
    logical,      pointer              :: srcIgnore(:,:)
    character(*), intent(in), optional :: idColName
    integer,      intent(in), optional :: idInstr, idBands(:)

    integer ::&
      numSources, numBands, numUniqueSrcs, srcNum, row, bandNum,&
      firstRowThisSource
    type(ColumnT) :: rateCol
    integer, pointer :: srcIdData(:), idBandData(:), idInstData(:)
    real(double), pointer :: tempSrcRaDeg(:), tempSrcDecDeg(:)
    real(single), pointer :: tempSrcRate(:)
    integer(int32), pointer :: uniqueIdValues(:)
    logical, allocatable :: mask(:), rateIsNull(:)
logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa'
    numSources = numberOfRows(tab)

    rateCol = column(tab, rateColName)
    call readColDataToPtr(tab, raColName,  tempSrcRaDeg)
    call readColDataToPtr(tab, decColName, tempSrcDecDeg)
    call readColDataToPtr(     rateCol,    tempSrcRate)
    ! readColDataToPtr is in ssclib/dal_aux.
if (testFlag) print *, '  bbb'

    allocate(rateIsNull(0:numSources-1))
    rateIsNull = .false. ! default
    if (nullDefined(rateCol)) then
      do row = 0, numSources-1
        if (isNull(rateCol, row)) rateIsNull(row) = .true.
      end do
    end if
if (testFlag) print *, '  ccc'

    if (present(idColName)) then
if (.not.present(idInstr)) call error('bug0',"")
if (.not.present(idBands)) call error('bug1',"")
if (testFlag) print *, '  ddd'

      numBands = size(idBands)
      call readColDataToPtr(tab, idColName,  srcIdData)
      call readColDataToPtr(tab, 'ID_BAND',  idBandData)
      call readColDataToPtr(tab, 'ID_INST',  idInstData)
if (testFlag) print *, '  eee'

      uniqueIdValues => findUniqueValues(srcIdData) ! in ssclib/array_utils.
      numUniqueSrcs = size(uniqueIdValues)
if (testFlag) print *, '  fff'

! Do checks:
allocate(mask(numSources))
do srcNum = 1, numUniqueSrcs
  mask = (srcIdData == uniqueIdValues(srcNum))
if (count(mask) <= 0) call error('bug2',"")
  mask = (mask .and. idInstData==idInstr)
  do bandNum = 1, numBands
    if (count(mask .and. idBandData==idBands(bandNum)) > 1)&
call error('tooManyBands',"")
  end do
end do
if (testFlag) print *, '  ggg'

      allocate(srcRaDeg( numUniqueSrcs), srcDecDeg(numUniqueSrcs))
      allocate(srcRate(  numUniqueSrcs, numBands))
      allocate(srcIgnore(numUniqueSrcs, numBands))
      srcRaDeg  = 0.0d0
      srcDecDeg = 0.0d0
      srcRate = 0.0
      srcIgnore = .true. ! default
if (testFlag) print *, '  hhh'

      do srcNum = 1, numUniqueSrcs
        firstRowThisSource&
          = smallestTrueIndex(srcIdData == uniqueIdValues(srcNum))
          ! in ssclib/array_utils.

        srcRaDeg( srcNum) = tempSrcRaDeg( firstRowThisSource-1)
        srcDecDeg(srcNum) = tempSrcDecDeg(firstRowThisSource-1)

        do row = firstRowThisSource, numSources - 1
        if (srcIdData( row) .ne. uniqueIdValues(srcNum)) cycle
        if (idInstData(row) .ne. idInstr)                cycle

          do bandNum = 1, numBands
          if (idBandData(row) .ne. idBands(bandNum)) cycle
          if (rateIsNull(row))                       cycle

            srcRate(  srcNum, bandNum) = tempSrcRate(row)
            srcIgnore(srcNum, bandNum) = .false.
          end do
        end do
      end do
if (testFlag) print *, '  iii'

      deallocate(srcIdData, idBandData, idInstData)
      deallocate(uniqueIdValues)
if (testFlag) print *, '  jjj'

    else ! .not. present(idColName). Treat all the rows as separate sources.
    ! This is designed for the generic, non-xmm case.

      numBands = 1

      allocate(srcRaDeg( numSources), srcDecDeg(numSources))
      allocate(srcRate(  numSources, 1))
      allocate(srcIgnore(numSources, 1))
      srcRaDeg  = tempSrcRaDeg
      srcDecDeg = tempSrcDecDeg
      srcRate(:,1) = tempSrcRate
      srcIgnore = .false.
      where(rateIsNull)
        srcIgnore(:,1) = .true.
        srcRate(  :,1) = 0.0
      end where
    end if

    call release(rateCol)
    deallocate(tempSrcRaDeg, tempSrcDecDeg, tempSrcRate)
if (testFlag) print *, '  zzz'

  end subroutine readSrcList

  !---------------------------------------------------------------------
  subroutine readSrcListPixels(tab, ximaColName, yimaColName, rateColName&
  , srcXima, srcYima, srcRate, srcIgnore, idColName, idInstr, idBands)
  !---------------------------------------------------------------------
  ! Note! The calling program should deallocate the pointers srcRaDeg,
  ! srcDecDeg, srcIgnore and srcRate after use.

    implicit none

    type(TableT), intent(in)           :: tab
    character(*), intent(in)           :: ximaColName, yimaColName, rateColName
    real(single), pointer              :: srcXima(:), srcYima(:), srcRate(:,:)
    logical,      pointer              :: srcIgnore(:,:)
    character(*), intent(in), optional :: idColName
    integer,      intent(in), optional :: idInstr, idBands(:)

    integer ::&
      numSources, numBands, numUniqueSrcs, srcNum, row, bandNum,&
      firstRowThisSource
    type(ColumnT) :: rateCol
    integer, pointer :: srcIdData(:), idBandData(:), idInstData(:)
    real(single), pointer :: tempSrcXima(:), tempSrcYima(:), tempSrcRate(:)
    integer(int32), pointer :: uniqueIdValues(:)
    logical, allocatable :: mask(:), rateIsNull(:)
!logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

    numSources = numberOfRows(tab)

    rateCol = column(tab, rateColName)
    call readColDataToPtr(tab, ximaColName, tempSrcXima)
    call readColDataToPtr(tab, yimaColName, tempSrcYima)
    call readColDataToPtr(     rateCol,     tempSrcRate)
    ! readColDataToPtr is in ssclib/dal_aux.

    allocate(rateIsNull(0:numSources-1))
    rateIsNull = .false. ! default
    if (nullDefined(rateCol)) then
      do row = 0, numSources-1
        if (isNull(rateCol, row)) rateIsNull(row) = .true.
      end do
    end if

    if (present(idColName)) then
if (.not.present(idInstr)) call error('bug0',"")
if (.not.present(idBands)) call error('bug1',"")

      numBands = size(idBands)
      call readColDataToPtr(tab, idColName,  srcIdData)
      call readColDataToPtr(tab, 'ID_BAND',  idBandData)
      call readColDataToPtr(tab, 'ID_INST',  idInstData)

      uniqueIdValues => findUniqueValues(srcIdData)
!***** this is present in ssclib/ssc_misc after ssclib_3.22.

      numUniqueSrcs = size(uniqueIdValues)

! Do checks:
allocate(mask(numSources))
do srcNum = 1, numUniqueSrcs
  mask = (srcIdData == uniqueIdValues(srcNum))
if (count(mask) <= 0) call error('bug2',"")
  mask = (mask .and. idInstData==idInstr)
  do bandNum = 1, numBands
    if (count(mask .and. idBandData==idBands(bandNum)) > 1)&
call error('tooManyBands',"")
  end do
end do

      allocate(srcXima( numUniqueSrcs), srcYima(numUniqueSrcs))
      allocate(srcRate(  numUniqueSrcs, numBands))
      allocate(srcIgnore(numUniqueSrcs, numBands))
      srcXima = 0.0
      srcYima = 0.0
      srcRate = 0.0
      srcIgnore = .true. ! default

      do srcNum = 1, numUniqueSrcs
        firstRowThisSource&
          = smallestTrueIndex(srcIdData == uniqueIdValues(srcNum))
          ! in ssclib/array_utils.

        srcXima(srcNum) = tempSrcXima(firstRowThisSource-1)
        srcYima(srcNum) = tempSrcYima(firstRowThisSource-1)

        do row = firstRowThisSource, numSources - 1
        if (srcIdData( row) .ne. uniqueIdValues(srcNum)) cycle
        if (idInstData(row) .ne. idInstr)                cycle

          do bandNum = 1, numBands
          if (idBandData(row) .ne. idBands(bandNum)) cycle
          if (rateIsNull(row))                       cycle

            srcRate(  srcNum, bandNum) = tempSrcRate(row)
            srcIgnore(srcNum, bandNum) = .false.
          end do
        end do
      end do

      deallocate(srcIdData, idBandData, idInstData)
      deallocate(uniqueIdValues)

    else ! .not. present(idColName). Treat all the rows as separate sources.
    ! This is designed for the generic, non-xmm case.

      numBands = 1

      allocate(srcXima(  numSources), srcYima(numSources))
      allocate(srcRate(  numSources, 1))
      allocate(srcIgnore(numSources, 1))
      srcXima = tempSrcXima
      srcYima = tempSrcYima
      srcRate(:,1) = tempSrcRate
      srcIgnore = .false.
      where(rateIsNull)
        srcIgnore(:,1) = .true.
        srcRate(  :,1) = 0.0
      end where
    end if

    call release(rateCol)
    deallocate(tempSrcXima, tempSrcYima, tempSrcRate)

  end subroutine readSrcListPixels

  !---------------------------------------------------------------------
  subroutine getCalPsfPatch(xiSize, yiSize, wcs, psfEnergyFraction&
  , psfMeanEnergies, srcThetaArcsec, srcPhi, patchCube, status)
  !---------------------------------------------------------------------
  ! Requires cal to have been set up for instrumentId and accuracyLevel.

    implicit none

    integer(int16), intent(in)  :: xiSize, yiSize
    type(WcsT),     intent(in)  :: wcs
    real(double),   intent(in)  :: psfEnergyFraction,&
                                   psfMeanEnergies(:),&
                                   srcThetaArcsec,&
                                   srcPhi
    real(single),   pointer     :: patchCube(:,:,:)
    integer,        intent(out) :: status

    real(double) :: effectiveSrcThetaArcsec, psfRadius
    type(PsfValidityRangesT) :: validPsfRanges
    type(PsfT) :: psfHandle
    type(PsfWindowT) :: psfWindow
    type(PsfBinSizeT) :: psfPixelSizeMm
    real(single), pointer :: psfImage(:,:)
    real(single), allocatable :: psfCube(:,:,:)
    integer(int16) :: startXi, finisXi, startYi, finisYi
    type(Point2dT), allocatable ::&
      psfImagePixelCorners(:,:)
    type(Point2dDbleT), allocatable ::&
      psfImgPixCrnrTel(:,:), psfImgPixCrnrTan(:,:), psfImgPixCrnrDet(:,:)
      ! types defined in coordinate.f90.
    integer :: foreignStatus, numBands, bandNum
!real(double) :: dummyTheta(3), dummyPhi(3)
!real(double), pointer :: cc2X(:), cc2Y(:), cc2Z(:)

logical :: testFlag = .false.

    !-------------------------------------------------------------------
    ! Start

if (testFlag) print *, '  aaa getCalPsfPatch'

    status = 0

    numBands = size(psfMeanEnergies)

    validPsfRanges = CAL_psfValidityRanges()
!print *, validPsfRanges%eMin, validPsfRanges%eMax
!print *, psfMeanEnergies
    if  (any(psfMeanEnergies < validPsfRanges%eMin)&
    .or. any(psfMeanEnergies > validPsfRanges%eMax))&
call error('energyOutOfRange',"")

    if (srcThetaArcsec < validPsfRanges%thetaMin)&
call error('thetaOutOfRange',"")

    if (srcPhi < validPsfRanges%phiMin&
    .or. srcPhi > validPsfRanges%phiMax)&
call error('phiOutOfRange',"")

    if (srcThetaArcsec > validPsfRanges%thetaMax) then
      call warning('thetaOutOfRange',"Theta is the off-axis angle. &
        &This out-of-range condition occurs from time to time, &
        &because the PSF in the CCF is defined for a range of &
        &values of theta which does not quite cover the whole field &
        &of view visible to the EPIC cameras. There is no cause for &
        &concern - the largest legal value of theta is used instead.")
      effectiveSrcThetaArcsec = validPsfRanges%thetaMax&
        * (1.0d0 - epsilon(srcThetaArcsec))
    else
      effectiveSrcThetaArcsec = srcThetaArcsec
    end if

if (testFlag) print *, '  bbb getCalPsfPatch'
    psfWindow%halfX = 0.0
    do bandNum = 1, numBands
!print *, '  band ID, mean EeV:', bandNum, psfMeanEnergies(bandNum)
      psfHandle = CAL_getPSF(psfMeanEnergies(bandNum), effectiveSrcThetaArcsec&
        , srcPhi)
      psfPixelSizeMm = CAL_psfBinSize(psfHandle)
      psfRadius = CAL_psfEncircledEnergy(psfHandle, fraction=psfEnergyFraction)
      psfWindow%halfX = max(psfWindow%halfX, psfRadius)
!call CAL_releaseMemory(psfHandle)
    end do
    psfWindow%halfY = psfWindow%halfX
if (testFlag) print *, '  ccc getCalPsfPatch'

    psfHandle = CAL_getPSF(psfMeanEnergies(1), effectiveSrcThetaArcsec&
      , srcPhi)
    call CAL_psfImage(psfHandle, psfWindow, psfImage)
!call dumpImageToFits(psfImage, 'srcmap_psf.ds')
    allocate(psfCube(numBands, size(psfImage,1), size(psfImage,2)))
    psfCube(1,:,:) = psfImage
    call CAL_releaseMemory(psfImage)
if (testFlag) print *, '  ddd getCalPsfPatch'
    do bandNum = 2, numBands
      psfHandle = CAL_getPSF(psfMeanEnergies(bandNum), effectiveSrcThetaArcsec&
        , srcPhi)
      call CAL_psfImage(psfHandle, psfWindow, psfImage)
      psfCube(bandNum,:,:) = psfImage
      call CAL_releaseMemory(psfImage)
    end do
if (testFlag) print *, '  eee getCalPsfPatch'

    ! Normalize, then vignet the psf image edges:

    psfCube = psfCube * psfPixelSizeMm%binX * psfPixelSizeMm%binY
    call vignetPatchCube(0.7, psfCube)

    ! Now rebin it to the image pixel coordinates:
if (testFlag) print *, '  fff getCalPsfPatch'

    allocate(psfImgPixCrnrTel(size(psfCube,2)+1, size(psfCube,3)+1))
    allocate(psfImgPixCrnrDet(size(psfCube,2)+1, size(psfCube,3)+1))
    allocate(psfImgPixCrnrTan(size(psfCube,2)+1, size(psfCube,3)+1))
    allocate(psfImagePixelCorners(size(psfCube,2)+1, size(psfCube,3)+1))
if (testFlag) print *, '  ggg getCalPsfPatch'

    call getPsfImagePixelCorners(srcThetaArcsec, srcPhi&
      , psfPixelSizeMm, psfCube(1,:,:), psfImgPixCrnrTel)
    ! in ssclib/coordinate.
    !
    ! The pairs of numbers (type(Point2dT)) in this array represent the
    ! coordinates of the corners of the psfImage pixels, expressed in
    ! the exposure-map image (ie, sky) coordinate system.
if (testFlag) print *, '  hhh getCalPsfPatch'

    call psfToDet(psfImgPixCrnrTel%x, psfImgPixCrnrTel%y&
      , psfImgPixCrnrDet%x, psfImgPixCrnrDet%y)
    ! in ssclib/coordinate.
if (testFlag) print *, '  iii getCalPsfPatch'

    call detToTan(psfImgPixCrnrDet%x, psfImgPixCrnrDet%y&
      , wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
      , scAttitude, expStartTimeStamp, psfImgPixCrnrTan%x&
      , psfImgPixCrnrTan%y)
    ! in ssclib/coordinate.
if (testFlag) print *, '  jjj getCalPsfPatch'

    call tanToPixels(wcs, psfImgPixCrnrTan%x, psfImgPixCrnrTan%y&
      , psfImagePixelCorners%x, psfImagePixelCorners%y)
    ! in ssclib/wcs_aux.
if (testFlag) print *, '  kkk getCalPsfPatch'

    startXi = max(1,     nint(minval(psfImagePixelCorners%x)))
    finisXi = min(xiSize, nint(maxval(psfImagePixelCorners%x), int16))
    startYi = max(1,     nint(minval(psfImagePixelCorners%y)))
    finisYi = min(yiSize, nint(maxval(psfImagePixelCorners%y), int16))
    allocate(patchCube(numBands, startXi:finisXi,startYi:finisYi))
if (testFlag) print *, '  lll getCalPsfPatch'

    psfImagePixelCorners%x = psfImagePixelCorners%x - real(startXi - 1)
    psfImagePixelCorners%y = psfImagePixelCorners%y - real(startYi - 1)
if (testFlag) print *, '  mmm getCalPsfPatch'

    call regridCartesian(psfImagePixelCorners, psfCube&
      , patchCube, status=foreignStatus) ! in ssclib/regridders.
if (testFlag) print *, '  nnn getCalPsfPatch'

if (foreignStatus < 0) call error('badRegridCartesianStatus','')
    if (foreignStatus > 0) then
      print *, 'regrid exit status:', foreignStatus
    end if

if (testFlag) print *, '  zzz getCalPsfPatch'

  end subroutine getCalPsfPatch

  !---------------------------------------------------------------------
  subroutine getCalPsfPatchPixels(xiSize, yiSize, wcs, psfEnergyFraction&
  , psfMeanEnergies, srcThetaArcsec, srcPhi, srcXima, srcYima, patchCube&
  , status)
  !---------------------------------------------------------------------
  ! Requires cal to have been set up for instrumentId and accuracyLevel.

    implicit none

    integer(int16), intent(in)  :: xiSize, yiSize
    type(WcsT),     intent(in)  :: wcs
    real(double),   intent(in)  :: psfEnergyFraction,&
                                   psfMeanEnergies(:),&
                                   srcThetaArcsec,&
                                   srcPhi
    real(single),   intent(in)  :: srcXima, srcYima
    real(single),   pointer     :: patchCube(:,:,:)
    integer,        intent(out) :: status

    real(double) :: effectiveSrcThetaArcsec, psfRadius
    type(PsfValidityRangesT) :: validPsfRanges
    type(PsfT) :: psfHandle
    type(PsfWindowT) :: psfWindow
    type(PsfBinSizeT) :: psfPixelSizeMm
    real(single), pointer :: psfImage(:,:)
    real(single), allocatable :: psfCube(:,:,:)
    integer(int16) :: startXi, finisXi, startYi, finisYi, xi, yi
    type(Point2dT), allocatable ::&
      psfImagePixelCorners(:,:)
    type(Point2dDbleT), allocatable ::&
      psfImgPixCrnrTel(:,:), psfImgPixCrnrTan(:,:), psfImgPixCrnrDet(:,:)
      ! types defined in coordinate.f90.
    integer :: foreignStatus, numBands, bandNum
!real(double) :: dummyTheta(3), dummyPhi(3)
!real(double), pointer :: cc2X(:), cc2Y(:), cc2Z(:)
    real(single) :: pixelGridCentreX, pixelGridCentreY
logical :: testflag = .false.

    !-------------------------------------------------------------------
    ! Start
if (testflag) print *, '  aaa getCalPsfPatchPixels'

    status = 0

    numBands = size(psfMeanEnergies)

    validPsfRanges = CAL_psfValidityRanges()

    if  (any(psfMeanEnergies < validPsfRanges%eMin)&
    .or. any(psfMeanEnergies > validPsfRanges%eMax))&
call error('energyOutOfRange',"")

    if (srcThetaArcsec < validPsfRanges%thetaMin)&
call error('thetaOutOfRange',"")

    if (srcPhi < validPsfRanges%phiMin&
    .or. srcPhi > validPsfRanges%phiMax)&
call error('phiOutOfRange',"")

    if (srcThetaArcsec > validPsfRanges%thetaMax) then
      call warning('thetaOutOfRange','')
      effectiveSrcThetaArcsec = validPsfRanges%thetaMax&
        * (1.0d0 - epsilon(srcThetaArcsec))
    else
      effectiveSrcThetaArcsec = srcThetaArcsec
    end if

    psfWindow%halfX = 0.0
    do bandNum = 1, numBands
      psfHandle = CAL_getPSF(psfMeanEnergies(bandNum), effectiveSrcThetaArcsec&
        , srcPhi)
      psfPixelSizeMm = CAL_psfBinSize(psfHandle)
      psfRadius = CAL_psfEncircledEnergy(psfHandle, fraction=psfEnergyFraction)
      psfWindow%halfX = max(psfWindow%halfX, psfRadius)
    end do
    psfWindow%halfY = psfWindow%halfX

    call CAL_psfImage(psfHandle, psfWindow, psfImage)
if (testflag) call dumpImageToFits(psfImage,'orig_psf_img_esm.ds')
    allocate(psfCube(numBands, size(psfImage,1), size(psfImage,2)))
    psfCube(1,:,:) = psfImage
    call CAL_releaseMemory(psfImage)
    do bandNum = 2, numBands
      call CAL_psfImage(psfHandle, psfWindow, psfImage)
      psfCube(bandNum,:,:) = psfImage
      call CAL_releaseMemory(psfImage)
    end do

    ! Normalize, then vignet the psf image edges:

    psfCube = psfCube * psfPixelSizeMm%binX * psfPixelSizeMm%binY
if (testflag) print *, '  new sum(psfImage)=', sum(psfCube(1,:,:))
    call vignetPatchCube(0.7, psfCube)

    ! Now rebin it to the image pixel coordinates:

    allocate(psfImgPixCrnrTel(size(psfCube,2)+1, size(psfCube,3)+1))
    allocate(psfImgPixCrnrDet(size(psfCube,2)+1, size(psfCube,3)+1))
    allocate(psfImgPixCrnrTan(size(psfCube,2)+1, size(psfCube,3)+1))
    allocate(psfImagePixelCorners(size(psfCube,2)+1, size(psfCube,3)+1))

    call getPsfImagePixelCorners(srcThetaArcsec, srcPhi&
      , psfPixelSizeMm, psfCube(1,:,:), psfImgPixCrnrTel)
    ! in ssclib/coordinate.

    call psfToDet(psfImgPixCrnrTel%x, psfImgPixCrnrTel%y&
      , psfImgPixCrnrDet%x, psfImgPixCrnrDet%y)
    ! in ssclib/coordinate.

    call detToTan(psfImgPixCrnrDet%x, psfImgPixCrnrDet%y&
      , wcs%x%refWorld*PiOn180, wcs%y%refWorld*PiOn180&
      , scAttitude, expStartTimeStamp, psfImgPixCrnrTan%x&
      , psfImgPixCrnrTan%y)
    ! in ssclib/coordinate.

    call tanToPixels(wcs, psfImgPixCrnrTan%x, psfImgPixCrnrTan%y&
      , psfImagePixelCorners%x, psfImagePixelCorners%y)
    ! in ssclib/wcs_aux.

! Now recenter it to (srcXima,srcYima):
if (isOdd(size(psfCube,2))) then
  xi = (size(psfCube,2) + 1) / 2.0
  if (isOdd(size(psfCube,3))) then
    yi = (size(psfCube,3) + 1) / 2.0
    pixelGridCentreX&
      = (psfImagePixelCorners(xi  ,yi  )%x&
      +  psfImagePixelCorners(xi+1,yi  )%x&
      +  psfImagePixelCorners(xi  ,yi+1)%x&
      +  psfImagePixelCorners(xi+1,yi+1)%x) / 4.0
    pixelGridCentreY&
      = (psfImagePixelCorners(xi,  yi  )%y&
      +  psfImagePixelCorners(xi+1,yi  )%y&
      +  psfImagePixelCorners(xi,  yi+1)%y&
      +  psfImagePixelCorners(xi+1,yi+1)%y) / 4.0
  else
    yi = 1 + size(psfCube,3) / 2.0
    pixelGridCentreX&
      = (psfImagePixelCorners(xi,  yi)%x&
      +  psfImagePixelCorners(xi+1,yi)%x) / 2.0
    pixelGridCentreY&
      = (psfImagePixelCorners(xi,  yi)%y&
      +  psfImagePixelCorners(xi+1,yi)%y) / 2.0
  end if
else
  xi = 1 + size(psfCube,2) / 2.0
  if (isOdd(size(psfCube,3))) then
    yi = (size(psfCube,3) + 1) / 2.0
    pixelGridCentreX&
      = (psfImagePixelCorners(xi,yi  )%x&
      +  psfImagePixelCorners(xi,yi+1)%x) / 2.0
    pixelGridCentreY&
      = (psfImagePixelCorners(xi,yi  )%y&
      +  psfImagePixelCorners(xi,yi+1)%y) / 2.0
  else
    yi = 1 + size(psfCube,3) / 2.0
    pixelGridCentreX = psfImagePixelCorners(xi,yi)%y
    pixelGridCentreY = psfImagePixelCorners(xi,yi)%y
  end if
end if
! isOdd is in ssclib/ssc_misc.

psfImagePixelCorners%x = psfImagePixelCorners%x - pixelGridCentreX + srcXima
psfImagePixelCorners%y = psfImagePixelCorners%y - pixelGridCentreY + srcYima

    startXi = max(1,     nint(minval(psfImagePixelCorners%x)))
    finisXi = min(xiSize, nint(maxval(psfImagePixelCorners%x), int16))
    startYi = max(1,     nint(minval(psfImagePixelCorners%y)))
    finisYi = min(yiSize, nint(maxval(psfImagePixelCorners%y), int16))
    allocate(patchCube(numBands, startXi:finisXi,startYi:finisYi))

    psfImagePixelCorners%x = psfImagePixelCorners%x - real(startXi - 1)
    psfImagePixelCorners%y = psfImagePixelCorners%y - real(startYi - 1)

    call regridCartesian(psfImagePixelCorners, psfCube&
      , patchCube, status=foreignStatus) ! in ssclib/regridders.

if (foreignStatus < 0) call error('badRegridCartesianStatus','')
    if (foreignStatus > 0) then
      print *, 'regrid exit status:', foreignStatus
    end if

if (testflag) call dumpImageToFits(patchCube(1,:,:),'psf_patch_esm.ds')
if (testflag) print *, '  zzz getCalPsfPatchPixels'

  end subroutine getCalPsfPatchPixels

  !---------------------------------------------------------------------
  subroutine vignetPatchArray(vigSkirtFraction, patch)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)    :: vigSkirtFraction
    real(single), intent(inout) :: patch(:,:)

    real(single), allocatable :: tempCube(:,:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempCube(1,size(patch,1),size(patch,2)))
    tempCube(1,:,:) = patch
    call vignetPatchCube(vigSkirtFraction, tempCube)
    patch = tempCube(1,:,:)

  end subroutine vignetPatchArray

  !---------------------------------------------------------------------
  subroutine vignetPatchCube(vigSkirtFraction, patch)
  !---------------------------------------------------------------------
  ! This is meant to take an array containing a source PSF image and
  ! multiply it by a vignetting function, so as to avoid the otherwise
  ! sharp drop-off at the edge of the array.
  !
  ! This has now been a bit modified to take account of another problem.
  ! emldetect assumes of course that the psf of the source it is measuring
  ! is of normal shape - at least out as far as the energy-fraction
  ! cutoff specified via the emldetect parameter --ecut. Hence it is no
  ! good, when making these simulated images, to multiply the whole PSF
  ! by some vignetting function, because that will distort its shape. So
  ! what I do now is retain the pristine PSF out to the ecut edge, but
  ! multiply by a vignetting or feathering function outboard of this.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)    :: vigSkirtFraction
    real(single), intent(inout) :: patch(:,:,:)

    integer(int16) :: xiSize, yiSize, i, xiPristineHalfSize, yiPristineHalfSize
!    integer(int16), allocatable :: indices(:)
    real(single) :: xHalfSize, yHalfSize
    real(single), allocatable ::&
      xVector(:), yVector(:), xVigFunction(:), yVigFunction(:)

    !-------------------------------------------------------------------
    ! Start

    xiSize = size(patch, 2)
    yiSize = size(patch, 3)
    xHalfSize = xiSize / 2.0 - 0.5
    yHalfSize = yiSize / 2.0 - 0.5
    xiPristineHalfSize = xHalfSize / (1.0 + vigSkirtFraction)
    yiPristineHalfSize = yHalfSize / (1.0 + vigSkirtFraction)

    allocate(xVector(xiSize), yVector(yiSize))
    allocate(xVigFunction(xiSize), yVigFunction(yiSize))

    xVector = (/(abs(i - (xHalfSize + 1.0)), i=1, xiSize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.

    where (xVector > xiPristineHalfSize)
      xVigFunction = (xVector - xiPristineHalfSize)&
      / ((xHalfSize + 1.0) - xiPristineHalfSize)
      xVigFunction = 1.0 - xVigFunction * xVigFunction
    elsewhere
      xVigFunction = 1.0
    end where

    yVector = (/(abs(i - (yHalfSize + 1.0)), i=1, yiSize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.

    where (yVector > yiPristineHalfSize)
      yVigFunction = (yVector - yiPristineHalfSize)&
      / ((yHalfSize + 1.0) - yiPristineHalfSize)
      yVigFunction = 1.0 - yVigFunction * yVigFunction
    elsewhere
      yVigFunction = 1.0
    end where

    do i = 1, size(patch, 1)
      patch(i,:,:) = patch(i,:,:) * spread(xVigFunction, 2, yiSize)&
                                  * spread(yVigFunction, 1, xiSize)
    end do

  end subroutine vignetPatchCube

  !---------------------------------------------------------------------
  subroutine vignetPatchCube_howItShouldBe(vigSkirtFraction, patch)
  !---------------------------------------------------------------------
!******* should change the indices of the cubes in the rest of the program (and in eimchip2sky too!?) so that the x and y axes are the first two. For the moment I will put the present routine on hold and go with the flow. IMS 2006-11-19.

  ! This is meant to take an array containing a source PSF image and
  ! multiply it by a vignetting function, so as to avoid the otherwise
  ! sharp drop-off at the edge of the array.
  !
  ! This has now been a bit modified to take account of another problem.
  ! emldetect assumes of course that the psf of the source it is measuring
  ! is of normal shape - at least out as far as the energy-fraction
  ! cutoff specified via the emldetect parameter --ecut. Hence it is no
  ! good, when making these simulated images, to multiply the whole PSF
  ! by some vignetting function, because that will distort its shape. So
  ! what I do now is retain the pristine PSF out to the ecut edge, but
  ! multiply by a vignetting or feathering function outboard of this.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)    :: vigSkirtFraction
    real(single), intent(inout) :: patch(:,:,:)

    integer(int16) :: xiSize, yiSize, i, xiPristineHalfSize, yiPristineHalfSize
!    integer(int16), allocatable :: indices(:)
    real(single) :: xHalfSize, yHalfSize
    real(single), allocatable ::&
      xVector(:), yVector(:), xVigFunction(:), yVigFunction(:)

    !-------------------------------------------------------------------
    ! Start

    xiSize = size(patch, 1)
    yiSize = size(patch, 2)
    xHalfSize = xiSize / 2.0 - 0.5
    yHalfSize = yiSize / 2.0 - 0.5
    xiPristineHalfSize = xHalfSize / (1.0 + vigSkirtFraction)
    yiPristineHalfSize = yHalfSize / (1.0 + vigSkirtFraction)

    allocate(xVector(xiSize), yVector(yiSize))
    allocate(xVigFunction(xiSize), yVigFunction(yiSize))

    xVector = (/(abs(i - (xHalfSize + 1.0)), i=1, xiSize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.

    where (xVector > xiPristineHalfSize)
      xVigFunction = (xVector - xiPristineHalfSize)&
      / ((xHalfSize + 1.0) - xiPristineHalfSize)
      xVigFunction = 1.0 - xVigFunction * xVigFunction
    elsewhere
      xVigFunction = 1.0
    end where

    yVector = (/(abs(i - (yHalfSize + 1.0)), i=1, yiSize)/)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.

    where (yVector > yiPristineHalfSize)
      yVigFunction = (yVector - yiPristineHalfSize)&
      / ((yHalfSize + 1.0) - yiPristineHalfSize)
      yVigFunction = 1.0 - yVigFunction * yVigFunction
    elsewhere
      yVigFunction = 1.0
    end where

    do i = 1, size(patch, 1)
      patch(i,:,:) = patch(i,:,:) * spread(xVigFunction, 2, yiSize)&
                                  * spread(yVigFunction, 1, xiSize)
    end do

  end subroutine vignetPatchCube_howItShouldBe

  !---------------------------------------------------------------------
  subroutine vignetPatchArray_20061115(patch)
  !---------------------------------------------------------------------
  ! This is meant to take an array containing a source PSF image and
  ! multiply it by a vignetting function, so as to avoid the otherwise
  ! sharp drop-off at the edge of the array.

    implicit none

    real(single), intent(inout) :: patch(:,:)

    real(single), allocatable :: tempCube(:,:,:)

    !-------------------------------------------------------------------
    ! Start

    allocate(tempCube(1,size(patch,1),size(patch,2)))
    tempCube(1,:,:) = patch
    call vignetPatchCube_20061115(tempCube)
    patch = tempCube(1,:,:)

  end subroutine vignetPatchArray_20061115

  !---------------------------------------------------------------------
  subroutine vignetPatchCube_20061115(patch)
  !---------------------------------------------------------------------
  ! This is meant to take an array containing a source PSF image and
  ! multiply it by a vignetting function, so as to avoid the otherwise
  ! sharp drop-off at the edge of the array.

    implicit none

    real(single), intent(inout) :: patch(:,:,:)

    integer(int16) :: xiSize, yiSize, i
    integer(int16), allocatable :: indices(:)
    real(single) ::&
      xHalfSize, yHalfSize,&
      xVector(size(patch, 2)), yVector(size(patch, 3))

    !-------------------------------------------------------------------
    ! Start

    xiSize = size(patch, 2)
    xHalfSize = xiSize / 2.0 - 0.5
    yiSize = size(patch, 3)
    yHalfSize = yiSize / 2.0 - 0.5

    allocate(indices(xiSize))
    indices = (/(i,i=1,xiSize)/)
    xVector = (indices - 1.0 - xHalfSize) / (xHalfSize + 1.0)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.
    xVector = 1.0 - xVector * xVector
    deallocate(indices)

    allocate(indices(yiSize))
    indices = (/(i,i=1,yiSize)/)
    yVector = (indices - 1.0 - yHalfSize) / (yHalfSize + 1.0)
    ! The +1 prevents patch being 0 all around the rim, which would be
    ! a waste of pixels.
    yVector = 1.0 - yVector * yVector
    deallocate(indices)

    do i = 1, size(patch, 1)
      patch(i,:,:) = patch(i,:,:) * spread(xVector, 2, yiSize)&
        * spread(yVector, 1, xiSize)
    end do

  end subroutine vignetPatchCube_20061115

  !---------------------------------------------------------------------
  subroutine addPatchArray(image, patch, rate)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: image(:,:)
    real(single), pointer :: patch(:,:)
    real(single), intent(in) :: rate

    integer :: xStart, xFinis, yStart, yFinis, xiSize, yiSize

    !-------------------------------------------------------------------
    ! Start

    xStart = lbound(patch, 1)
    xFinis = ubound(patch, 1)
    yStart = lbound(patch, 2)
    yFinis = ubound(patch, 2)
    xiSize = size(image, 1)
    yiSize = size(image, 2)

    if (xStart > xiSize .or. xFinis < 1 .or. yStart > yiSize .or. yFinis < 1) then
  return
    end if

    xStart = max(1, xStart)
    xFinis = min(xiSize, xFinis)
    yStart = max(1, yStart)
    yFinis = min(yiSize, yFinis)

    image(xStart:xFinis, yStart:yFinis) = image(xStart:xFinis, yStart:yFinis)&
      + patch(xStart:xFinis, yStart:yFinis) * rate

  end subroutine addPatchArray

  !---------------------------------------------------------------------
  subroutine addPatchCubeScal(image, patch, rate)
  !---------------------------------------------------------------------
  ! With this routine it is assumed that the output images will be scaled
  ! to the correct rates. It is only appropriate really when used with a
  ! list of simulated sources, each of which has the same spectrum.

    implicit none

    real(single), intent(inout) :: image(:,:,:)
    real(single), pointer :: patch(:,:,:)
    real(single), intent(in) :: rate

    integer :: xStart, xFinis, yStart, yFinis, xiSize, yiSize, bandNum, numBands

    !-------------------------------------------------------------------
    ! Start

    numBands = size(image,1)
if (numBands .ne. size(patch,1)) call error('badNumPatchLayers',"")
!if (numBands .ne. size(rates)) call error('badNumRates',"")
    xStart = lbound(patch, 2)
    xFinis = ubound(patch, 2)
    yStart = lbound(patch, 3)
    yFinis = ubound(patch, 3)
    xiSize = size(image, 2)
    yiSize = size(image, 3)

    if (xStart > xiSize .or. xFinis < 1 .or. yStart > yiSize .or. yFinis < 1) then
  return
    end if

    xStart = max(1, xStart)
    xFinis = min(xiSize, xFinis)
    yStart = max(1, yStart)
    yFinis = min(yiSize, yFinis)

    do bandNum = 1, numBands
      image(bandNum, xStart:xFinis, yStart:yFinis)&
        = image(bandNum, xStart:xFinis, yStart:yFinis)&
        + patch(bandNum, xStart:xFinis, yStart:yFinis) * rate
    end do

  end subroutine addPatchCubeScal

  !---------------------------------------------------------------------
  subroutine addPatchCubeVec(image, patch, rates)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(inout) :: image(:,:,:)
    real(single), pointer :: patch(:,:,:)
    real(single), intent(in) :: rates(:)

    integer :: xStart, xFinis, yStart, yFinis, xiSize, yiSize, bandNum, numBands

    !-------------------------------------------------------------------
    ! Start

    numBands = size(image,1)
if (numBands .ne. size(patch,1)) call error('badNumPatchLayers',"")
if (numBands .ne. size(rates)) call error('badNumRates',"")
    xStart = lbound(patch, 2)
    xFinis = ubound(patch, 2)
    yStart = lbound(patch, 3)
    yFinis = ubound(patch, 3)
    xiSize = size(image, 2)
    yiSize = size(image, 3)

    if (xStart > xiSize .or. xFinis < 1 .or. yStart > yiSize .or. yFinis < 1) then
  return
    end if

    xStart = max(1, xStart)
    xFinis = min(xiSize, xFinis)
    yStart = max(1, yStart)
    yFinis = min(yiSize, yFinis)

    do bandNum = 1, numBands
      image(bandNum, xStart:xFinis, yStart:yFinis)&
        = image(bandNum, xStart:xFinis, yStart:yFinis)&
        + patch(bandNum, xStart:xFinis, yStart:yFinis) * rates(bandNum)
    end do

  end subroutine addPatchCubeVec

  !---------------------------------------------------------------------
  integer function numSeparateValues(vector)
  !---------------------------------------------------------------------

    implicit none

    integer(int32), intent(in) :: vector(:)

    integer :: numValues, i
    integer(int32), allocatable :: tempVector(:)

    !-------------------------------------------------------------------
    ! Start

    numSeparateValues = 0
    numValues = size(vector)
  if (numValues <= 0) return
    allocate(tempVector(numValues))
    tempVector = vector
    call quickSort(tempVector) ! in ssclib/sort_mod.
    numSeparateValues = 1
    do i = 2, numValues
      if (tempVector(i) .ne. tempVector(i-1))&
        numSeparateValues = numSeparateValues + 1
    end do

  end function numSeparateValues

!  !---------------------------------------------------------------------
!  function findUniqueValues(vector) result(uniqueValues)
!  !---------------------------------------------------------------------
!!**** present in ssc_misc in ssclib>3.21.
!
!    implicit none
!
!    integer(int32), intent(in) :: vector(:)
!    integer(int32), pointer    :: uniqueValues(:)
!
!    integer :: numValues, i, numUniqueValues
!    integer(int32), allocatable :: tempVector(:), tempUniqueVals(:)
!
!    !-------------------------------------------------------------------
!    ! Start
!
!    numValues = size(vector)
!    if (numValues <= 0) then
!      allocate(uniqueValues(0))
!return
!    end if
!
!    allocate(tempVector(numValues), tempUniqueVals(numValues))
!    tempVector = vector
!    call quickSort(tempVector) ! in ssclib/sort_mod.
!    numUniqueValues = 1
!    tempUniqueVals(numUniqueValues) = tempVector(1)
!    do i = 2, numValues
!      if (tempVector(i) .ne. tempVector(i-1)) then
!        numUniqueValues = numUniqueValues + 1
!        tempUniqueVals(numUniqueValues) = tempVector(i)
!      end if
!    end do
!    allocate(uniqueValues(numUniqueValues))
!    uniqueValues = tempUniqueVals(:numUniqueValues)
!
!  end function findUniqueValues

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module esrcmap_mod
